﻿// <copyright file="ChannelInputStream.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> ChannelInputStream.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Channel
{
    /// <summary> ChannelInputStream.
    /// Implements a stream over a ssh channel for input/reading only.
    /// </summary>
    public class ChannelInputStream : Stream
    {
        private SSHChannel c;
        private bool extendedFlag;
        
        /// <summary>
        /// Initializes a new instance of the ChannelInputStream class.
        /// </summary>
        /// <param name="c">The channel that the stream is part of.</param>
        /// <param name="isExtended">Indicates whether the channel is carrying extended data (usually stderr).</param>
        internal ChannelInputStream(SSHChannel c, bool isExtended)
        {
            this.c = c;
            this.extendedFlag = isExtended;
        }
        
        /// <summary>
        /// Gets a value indicating whether the stream can be read from (always true).
        /// </summary>
        public override bool CanRead
        {
            get { return true; }
        }
        
        /// <summary>
        /// Gets a value indicating whether you can seek on the stream (always false).
        /// </summary>
        public override bool CanSeek
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets a value indictaing whether the stream can timeout or not (always false).
        /// </summary>
        public override bool CanTimeout
        {
            get { return base.CanTimeout; }
        }
        
        /// <summary>
        /// Gets a value indictaing whether the stream can be written to (always false).
        /// </summary>
        public override bool CanWrite
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets or sets the poition within the curent stream (not implemented)
        /// .</summary>
        public override long Position
        {
            get { return 0L; }
            set { }
        }
        
        /// <summary>
        /// Gets the length in bytes of the stream (not implemented).
        /// </summary>
        public override long Length
        {
            get { return 0L; }
        }
        
        /// <summary>
        /// Closes the current stream.
        /// </summary>
        public override void Close()
        {
        }
        
        /// <summary>
        /// Reads a sequence of bytes from the current stream.
        /// </summary>
        /// <param name="buffer">
        /// An array of bytes. When this method returns, the buffer contains the specified byte array
        /// with the values between offset and (offset + count - 1) replaced by the bytes read from the current source.
        /// </param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin storing the data read from the current stream.</param>
        /// <param name="count">The maximum number of bytes to be read from the current stream.</param>
        /// <returns>
        /// The total number of bytes read into the buffer. This can be less than the number of bytes requested if that
        /// many bytes are not currently available, or zero (0) if the end of the stream has been reached.
        /// </returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            if (buffer == null)
            {
                throw new ArgumentException("The buffer to read from is null.");
            }
            
            if ((offset < 0) || (count < 0) || ((offset + count) > buffer.Length) || ((offset + count) < 0) || (offset > buffer.Length))
            {
                throw new ArgumentException("The offset or count parameters are invalid");
            }
            
            if (count == 0)
            {
                return 0;
            }
            
            int ret = this.c.CM.GetChannelData(this.c, this.extendedFlag, buffer, offset, count);
            
            return ret;
        }
        
        /// <summary>
        /// Reads a sequence of bytes from the current stream.
        /// </summary>
        /// <param name="buffer">
        /// An array of bytes. When this method returns, the buffer contains the specified byte array
        /// with the data replaced by the bytes read from the current source.
        /// </param>
        /// <returns>
        /// The total number of bytes read into the buffer. This can be less than the number of bytes requested if that
        /// many bytes are not currently available, or zero (0) if the end of the stream has been reached.
        /// </returns>
        public int Read(byte[] buffer)
        {
            return this.Read(buffer, 0, buffer.Length);
        }
        
        /// <summary>
        /// Writes a sequence of bytes to the current stream. (not implemented).
        /// </summary>
        /// <param name="buffer">An array of bytes. This method copies count bytes from buffer to the current stream.</param>
        /// <param name="offset">The offset into the byte array where the data is written from.</param>
        /// <param name="count">The number of bytes to be written to the current stream.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Set the length of the stream (not implemented).
        /// </summary>
        /// <param name="value">The desired length of the current stream in bytes.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Seek to a place in the stream (not implemented).
        /// </summary>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin copying bytes to the current stream.</param>
        /// <param name="origin">A value of type SeekOrigin indicating the reference point used to obtain the new position.</param>
        /// <returns>The new position within the current stream. </returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Clears all buffers for this stream and causes any buffered data to be written, (not implemented).
        /// </summary>
        public override void Flush()
        {
            throw new NotImplementedException();
        }        
    }
}
