﻿// <copyright file="ChannelOutputStream.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> ChannelOutputStream.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Channel
{
    /// <summary> ChannelOutputStream.
    /// Implements a stream over a ssh channel for output/writing only.
    /// </summary>
    public class ChannelOutputStream : Stream
    {
        private SSHChannel c;
        private bool isClosed = false;
        
        /// <summary>
        /// Initializes a new instance of the ChannelOutputStream class.
        /// </summary>
        /// <param name="c">The channel that the stream is part of.</param>
        internal ChannelOutputStream(SSHChannel c)
        {
            this.c = c;
        }
        
        /// <summary>Gets a value indicating whether the stream can be read from (always false).</summary>
        public override bool CanRead
        {
            get { return false; }
        }
        
        /// <summary>Gets a value indicating whether you can seek on the stream (always false).</summary>
        public override bool CanSeek
        {
            get { return false; }
        }

        /// <summary>Gets a value indictaing whether the stream can timeout or not (always false).</summary>
        public override bool CanTimeout
        {
            get { return false; }
        }

        /// <summary>Gets a value indictaing whether the stream can be written to (always false).</summary>
        public override bool CanWrite
        {
            get { return true; }
        }
        
        /// <summary>Gets or sets the poition within the curent stream (not implemented).</summary>
        public override long Position
        {
            get { return 0L; }
            set { }
        }
        
        /// <summary>Gets the length in bytes of the stream (not implemented).</summary>
        public override long Length
        {
            get { return 0L; }
        }
        
        /// <summary>Closes the current stream.</summary>
        public override void Close()
        {
            if (this.isClosed == false)
            {
                this.isClosed = true;
                this.c.CM.SendEOF(this.c);
            }
        }
        
        /// <summary>Clears all buffers for this stream and causes any buffered data to be written.</summary>
        public override void Flush()
        {
            // This is a no-op, since this stream is unbuffered
            if (this.isClosed)
            {
                throw new IOException("This OutputStream is closed.");
            }
        }
        
        /// <summary>Writes a sequence of bytes to the current stream.</summary>
        /// <param name="buffer">An array of bytes. This method copies count bytes from buffer to the current stream.</param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin copying bytes to the current stream.</param>
        /// <param name="count">The number of bytes to be written to the current stream.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            if (this.isClosed)
            {
                throw new IOException("This OutputStream is closed.");
            }
            
            if (buffer == null)
            {
                throw new ArgumentException("Buffer to write from is null");
            }
            
            if ((offset < 0) || (count < 0) || ((offset + count) > buffer.Length) || ((offset + count) < 0) || (offset > buffer.Length))
            {
                throw new ArgumentException("Offset or count parameters are invalid.");
            }
            
            if (count == 0)
            {
                return;
            }
            
            this.c.CM.SendData(this.c, buffer, offset, count);
        }
        
        /// <summary>Writes a sequence of bytes to the current stream.</summary>
        /// <param name="buffer">An array of bytes. This method copies count bytes from buffer to the current stream.</param>
        public void Write(byte[] buffer)
        {
            this.Write(buffer, 0, buffer.Length);
        }
        
        /// <summary>Reads a sequence of bytes from the current stream, (not implemented).</summary>
        /// <param name="buffer">
        /// An array of bytes. When this method returns, the buffer contains the specified byte array
        /// with the data replaced by the bytes read from the current source.
        /// </param>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin storing the data read from the current stream.</param>
        /// <param name="count">The maximum number of bytes to be read from the current stream.</param>
        /// <returns>
        /// The total number of bytes read into the buffer. This can be less than the number of bytes requested if that
        /// many bytes are not currently available, or zero (0) if the end of the stream has been reached.
        /// </returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>Set the length of the stream (not implemented).</summary>
        /// <param name="value">The desired length of the current stream in bytes.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>Seek to a place in the stream (not implemented).</summary>
        /// <param name="offset">The zero-based byte offset in buffer at which to begin copying bytes to the current stream.</param>
        /// <param name="origin">A value of type SeekOrigin indicating the reference point used to obtain the new position.</param>
        /// <returns>The new position within the current stream. </returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }
    }
}
