﻿// <copyright file="LocalAcceptThread.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> LocalAcceptThread.
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version>  0.80, 11/26/2010
// </version>

using System;
using System.IO;
using System.Net;
using System.Net.Sockets;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// Class implements a TcpListener that listens on it's own thead for an incoming connection.
    /// </summary>
    public class LocalAcceptThread : JavaThread, IChannelWorkerThread
    {
        private ChannelManager cm;
        private string hostToConnect;
        private int portToConnect;
        
        private TcpListener ss;
        
        /// <summary>Initializes a new instance of the LocalAcceptThread class.</summary>
        /// <param name="cm">The ChannelManager that this LocalAcceptThead is associated with.</param>
        /// <param name="localPort">The local port that we will listen on.</param>
        /// <param name="hostToConnect">The remote host to connect to.</param>
        /// <param name="portToConnect">The remote port to connect to.</param>
        public LocalAcceptThread(ChannelManager cm, int localPort, string hostToConnect, int portToConnect)
        {
            this.cm = cm;
            this.hostToConnect = hostToConnect;
            this.portToConnect = portToConnect;
            
            TcpListener listen = new TcpListener(IPAddress.Any, localPort);
            listen.Start();
            this.ss = listen;
        }
        
        /// <summary>Initializes a new instance of the LocalAcceptThread class.</summary>
        /// <param name="cm">The ChannelManager that this LocalAcceptThead is associated with.</param>
        /// <param name="localAddress">The local IPEndPoint that we will listen on.</param>
        /// <param name="hostToConnect">The remote host to connect to.</param>
        /// <param name="portToConnect">The remote port to connect to.</param>
        public LocalAcceptThread(ChannelManager cm, IPEndPoint localAddress, string hostToConnect, int portToConnect)
        {
            this.cm = cm;
            this.hostToConnect = hostToConnect;
            this.portToConnect = portToConnect;
            
            TcpListener listen = new TcpListener(localAddress);
            listen.Start();
            this.ss = listen;
        }
        
        /// <summary>
        /// Runs the thread that creates the local port forward.
        /// </summary>
        public override void Run()
        {
            try
            {
                this.cm.RegisterThread(this);
            }
            catch (IOException)
            {
                this.StopWorking();
                return;
            }
            
            while (true)
            {
                Socket s = null;
                
                try
                {
                    s = this.ss.AcceptSocket();
                }
                catch (SocketException)
                {
                    this.StopWorking();
                    return;
                }
                
                SSHChannel cn = null;
                StreamForwarder r2l = null;
                StreamForwarder l2r = null;
                
                try
                {
                    // This may fail, e.g., if the remote port is closed (in optimistic terms: not open yet)
                    IPEndPoint remote = (IPEndPoint)s.RemoteEndPoint;
                    cn = this.cm.OpenDirectTCPIPChannel(this.hostToConnect, this.portToConnect, remote.Address.ToString(), remote.Port);
                }
                catch (IOException)
                {
                    // Simply close the local socket and wait for the next incoming connection
                    try
                    {
                        s.Close();
                    }
                    catch (IOException)
                    {
                    }
                    
                    continue;
                }
                
                try
                {
                    NetworkStream ns = new NetworkStream(s);
                    r2l = new StreamForwarder(cn, null, null, cn.StandardOutputStream, ns, "RemoteToLocal");
                    l2r = new StreamForwarder(cn, r2l, s, ns, cn.StandardInputStream, "LocalToRemote");
                }
                catch (IOException e)
                {
                    try
                    {
                        /* This message is only visible during debugging, since we discard the channel immediatelly */
                        cn.CM.CloseChannel(cn, "Weird error during creation of StreamForwarder (" + e.ToString() + ")", true);
                    }
                    catch (IOException)
                    {
                    }
                    
                    continue;
                }

                r2l.IsBackground = true;
                l2r.IsBackground = true;
                r2l.Start();
                l2r.Start();
            }
        }
        
        /// <summary>
        /// This method shuts down the local accept thread by shutting down the TcpListener.
        /// </summary>
        public void StopWorking()
        {
            try
            {
                // This will lead to an InvalidOperationException in the ss.AcceptSocket() call
                this.ss.Stop();
            }
            catch (IOException)
            {
            }
        }
    }
}
