﻿// <copyright file="SSHChannel.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Channel
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

namespace Eunomia.SSH2.Channel
{
    /// <summary>
    /// Channel represents a multiplexed channel over a SSH connection.
    /// This is the same as ganymed's Channel.  It is renamed to prevent
    /// a name conflict with the 'Channel' in the Eunomia.SSH2 namespace.
    /// </summary>
    public class SSHChannel : IDisposable
    {
        // OK. Here is an important part of the JVM Specification:
        // (http://java.sun.com/docs/books/vmspec/2nd-edition/html/Threads.doc.html#22214)
        // 
        // Any association between locks and variables is purely conventional.
        // Locking any lock conceptually flushes all variables from a thread's
        // working memory, and unlocking any lock forces the writing out to main
        // memory of all variables that the thread has assigned. That a lock may be
        // associated with a particular object or a class is purely a convention.
        // (...)
        // 
        // If a thread uses a particular shared variable only after locking a
        // particular lock and before the corresponding unlocking of that same lock,
        // then the thread will read the shared value of that variable from main
        // memory after the lock operation, if necessary, and will copy back to main
        // memory the value most recently assigned to that variable before the
        // unlock operation.
        // 
        // This, in conjunction with the mutual exclusion rules for locks, suffices
        // to guarantee that values are correctly transmitted from one thread to
        // another through shared variables.
        // 
        // ====> Always keep that in mind when modifying the Channel/ChannelManger
        // code.
        internal const int STATE_OPENING = 1;
        internal const int STATE_OPEN = 2;
        internal const int STATE_CLOSED = 4;
        
        internal const int CHANNEL_BUFFER_SIZE = 30000;
        
        private ChannelManager cm;
        private ChannelOutputStream stdinStream;
        private ChannelInputStream stdoutStream;
        private ChannelInputStream stderrStream;
        
        private int localID = -1;
        private int remoteID = -1;
        
        private object channelSendLock;
        private bool closeMessageSent = false;

        private byte[] msgWindowAdjust;
        
        private int state;

        private bool closeMessageRecv = false;
        
        private int successCounter = 0;
        private int failedCounter = 0;

        private int localWindow = 0;
        private long remoteWindow = 0;
        private int localMaxPacketSize = -1;
        private int remoteMaxPacketSize = -1;

        private byte[] stdoutBuffer;
        private byte[] stderrBuffer;
        private int stdoutReadpos = 0;
        private int stdoutWritepos = 0;
        private int stderrReadpos = 0;
        private int stderrWritepos = 0;
        private bool eof = false;
        private int exit_status;

        private string exit_signal;
        
        private string hexX11FakeCookie;

        // reasonClosed is special, since we sometimes need to access it
        // while holding the channelSendLock.
        // We protect it with a private short term lock.
        private object reasonClosedLock;
        private string reasonClosed = null;

        /// <summary>Initializes a new instance of the SSHChannel class.</summary>
        /// <param name="cm">This is the ChannelManager that 'owns' the Channel.</param>
        public SSHChannel(ChannelManager cm)
        {
            this.stderrBuffer = new byte[CHANNEL_BUFFER_SIZE];
            this.stdoutBuffer = new byte[CHANNEL_BUFFER_SIZE];
            this.msgWindowAdjust = new byte[9];
            this.channelSendLock = new object();
            this.state = STATE_OPENING;
            this.reasonClosedLock = new object();
            
            this.cm = cm;
            
            this.localWindow = CHANNEL_BUFFER_SIZE;
            this.localMaxPacketSize = 35000 - 1024; // leave enough slack
            
            this.stdinStream = new ChannelOutputStream(this);
            this.stdoutStream = new ChannelInputStream(this, false);
            this.stderrStream = new ChannelInputStream(this, true);
        }
        
        // To achieve correctness, the following rules have to be respected when
        // accessing this object:

        /// <summary>
        /// Gets a reference to the Channel Manager that 'owns' this channel
        /// and can always be read.
        /// </summary>
        public ChannelManager CM
        {
            get { return this.cm; }
        }
        
        /// <summary>Gets the Channel's input stream and can always be read.</summary>
        public ChannelOutputStream StandardInputStream
        {
            get { return this.stdinStream; }
        }
        
        /// <summary>Gets the Channel's output stream and can always be read.</summary>
        public ChannelInputStream StandardOutputStream
        {
            get { return this.stdoutStream; }
        }
        
        /// <summary>Gets Channel's error stream and can always be read.</summary>
        public ChannelInputStream StandardErrorStream
        {
            get { return this.stderrStream; }
        }

        /// <summary>Gets or sets the channel's local id.</summary>
        /// <remarks>
        /// This field will only be written while the channel is in state STATE_OPENING.
        /// The code makes sure that the two fields are written out when the state is
        /// changing to STATE_OPEN.
        /// Therefore, if you know that the Channel is in state STATE_OPEN, then you
        /// can read this field without synchronizing on the Channel. However, make
        /// sure that you get the latest values (e.g., flush caches by synchronizing on any
        /// object). However, to be on the safe side, you can lock the channel.
        /// </remarks>
        public int LocalId
        {
            get { return this.localID; }
            set { this.localID = value; }
        }
        
        /// <summary>Gets or sets the channel's remote id.</summary>
        /// <remarks>
        /// This field will only be written while the channel is in state STATE_OPENING.
        /// The code makes sure that the two fields are written out when the state is
        /// changing to STATE_OPEN.
        /// Therefore, if you know that the Channel is in state STATE_OPEN, then you
        /// can read this field without synchronizing on the Channel. However, make
        /// sure that you get the latest values (e.g., flush caches by synchronizing on any
        /// object). However, to be on the safe side, you can lock the channel.
        /// </remarks>
        public int RemoteId
        {
            get { return this.remoteID; }
            set { this.remoteID = value; }
        }
        
        /// <summary>Gets or sets the channel's closed message sent lock.</summary>
        /// <remarks>
        /// This field is used to lock when we send data on this channel and when we
        /// want to read/write the ClosedMessageSent bool.
        /// Make sure that we never send a data/EOF/WindowChange msg after a CLOSE msg.
        /// This is a little bit complicated, but we have to do it in that way, since
        /// we cannot keep a lock on the Channel during the send operation (this
        /// would block sometimes the receiver thread, and, in extreme cases, can
        /// lead to a deadlock on both sides of the connection (senders are blocked
        /// since the receive buffers on the other side are full, and receiver
        /// threads wait for the senders to finish). It all depends on the
        /// implementation on the other side. But we cannot make any assumptions, we
        /// have to assume the worst case. Confused? Just believe me.
        /// If you send a message on a channel, then you have to aquire the
        /// "channelSendLock" and check the "closeMessageSent" flag (this variable
        /// may only be accessed while holding the "channelSendLock" !!!
        /// BTW: NEVER EVER SEND MESSAGES FROM THE RECEIVE THREAD - see explanation
        /// above.
        /// </remarks>
        public object ChannelSendLock
        {
            get { return this.channelSendLock; }
            set { this.channelSendLock = value; }
        }

        /// <summary>Gets or sets a value indicating whether a close message has been sent.</summary>
        /// <remarks>
        /// This field is true when the channel has sent a closed message.
        /// Make sure that we never send a data/EOF/WindowChange msg after a CLOSE msg.
        /// This is a little bit complicated, but we have to do it in that way, since
        /// we cannot keep a lock on the Channel during the send operation (this
        /// would block sometimes the receiver thread, and, in extreme cases, can
        /// lead to a deadlock on both sides of the connection (senders are blocked
        /// since the receive buffers on the other side are full, and receiver
        /// threads wait for the senders to finish). It all depends on the
        /// implementation on the other side. But we cannot make any assumptions, we
        /// have to assume the worst case. Confused? Just believe me.
        /// If you send a message on a channel, then you have to aquire the
        /// "channelSendLock" and check the "closeMessageSent" flag (this variable
        /// may only be accessed while holding the "channelSendLock" !!!
        /// BTW: NEVER EVER SEND MESSAGES FROM THE RECEIVE THREAD - see explanation
        /// above.
        /// </remarks>
        public bool CloseMessageSent
        {
            get { return this.closeMessageSent; }
            set { this.closeMessageSent = value; }
        }
        
        /// <summary>Gets the channel's msgWindowAdjust byte buffer.</summary>
        /// <remarks>Stop memory fragmentation by allocating this often used buffer.
        /// May only be used while holding the channelSendLock.</remarks>
        public byte[] MessageWindowAdjust
        {
            get { return this.msgWindowAdjust; }
        }

        /// <summary>
        /// Gets or sets the channel's State.  You must lock the channel before reading or
        /// writing this field.
        /// </summary>
        public int State
        {
            get { return this.state; }
            set { this.state = value; }
        }
        
        /// <summary>
        /// Gets or sets a value indicating whether the channel has received a close message.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public bool CloseMessageReceived
        {
            get { return this.closeMessageRecv; }
            set { this.closeMessageRecv = value; }
        }

        /// <summary>
        /// Gets or sets a the number of times we have received a SSH_MSG_CHANNEL_SUCCESS message.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        /// <remarks>This is a stupid implementation. At the moment we can only wait
        /// for one pending request per channel.</remarks>
        public int SuccessCounter
        {
            get { return this.successCounter; }
            set { this.successCounter = value; }
        }
        
        /// <summary>
        /// Gets or sets a the number of times we have received a SSH_MSG_CHANNEL_FAILURE message.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int FailedCounter
        {
            get { return this.failedCounter; }
            set { this.failedCounter = value; }
        }
        
        /// <summary>
        /// Gets or sets the localWindow which is how much data we can accept per message.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        /// <remarks>locally, we use a small window, &lt; 2^31.</remarks>
        public int LocalWindow
        {
            get { return this.localWindow; }
            set { this.localWindow = value; }
        }
        
        /// <summary>
        /// Gets or sets the remoteWindow which is how much data the remote can accept per message.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        /// <remarks>long for readable  2^32 - 1 window support.</remarks>
        public long RemoteWindow
        {
            get { return this.remoteWindow; }
            set { this.remoteWindow = value; }
        }
        
        /// <summary>
        /// Gets or sets the local maximum packet size.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int LocalMaxPacketSize
        {
            get { return this.localMaxPacketSize; }
            set { this.localMaxPacketSize = value; }
        }
        
        /// <summary>
        /// Gets or sets the remote maximum packet size.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int RemoteMaxPacketSize
        {
            get { return this.remoteMaxPacketSize; }
            set { this.remoteMaxPacketSize = value; }
        }
        
        /// <summary>
        /// Gets or sets the Standard output buffer.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public byte[] StandardOutputBuffer
        {
            get { return this.stdoutBuffer; }
            set { this.stdoutBuffer = value; }
        }

        /// <summary>
        /// Gets or sets the standard error buffer.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public byte[] StandardErrorBuffer
        {
            get { return this.stderrBuffer; }
            set { this.stderrBuffer = value; }
        }

        /// <summary>
        /// Gets or sets the standard output buffer read position.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int StandardOutputReadPosition
        {
            get { return this.stdoutReadpos; }
            set { this.stdoutReadpos = value; }
        }

        /// <summary>
        /// Gets or sets the standard output buffer write position.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int StandardOutputWritePosition
        {
            get { return this.stdoutWritepos; }
            set { this.stdoutWritepos = value; }
        }
        
        /// <summary>
        /// Gets or sets the standard error buffer read position.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int StandardErrorReadPosition
        {
            get { return this.stderrReadpos; }
            set { this.stderrReadpos = value; }
        }

        /// <summary>
        /// Gets or sets the standard error buffer write postion.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public int StandardErrorWritePosition
        {
            get { return this.stderrWritepos; }
            set { this.stderrWritepos = value; }
        }

        /// <summary>
        /// Gets or sets a value indicating whether the channel has been closed or not.
        /// You must lock the channel before reading or writing this field.
        /// </summary>
        public bool EOF
        {
            get { return this.eof; }
            set { this.eof = value; }
        }

        /// <summary>
        /// Gets or sets the channel's exit signal.
        /// You must lock the channel before writing to this field.
        /// </summary>
        public string ExitSignal
        {
            get
            {
                lock (this)
                {
                    return this.exit_signal;
                }
            }
            
            set
            {
                this.exit_signal = value;
            }
        }

        /// <summary>
        /// Gets or sets the channel's exit status.
        /// You must lock the channel before writing to this field.
        /// </summary>
        public int ExitStatus
        {
            get
            {
                lock (this)
                {
                    return this.exit_status;
                }
            }
            
            set
            {
                this.exit_status = value;
            }
        }

        /// <summary>
        /// Gets or sets the X11 cookie.
        /// </summary>
        /// <remarks>We keep the x11 cookie so that this channel can be closed when this
        /// specific x11 forwarding gets stopped.</remarks>
        public string HexX11FakeCookie
        {
            get { return this.hexX11FakeCookie; }
            set { this.hexX11FakeCookie = value; }
        }

        /// <summary>
        /// Gets or sets the channel's reason closed.
        /// </summary>.
        public string ReasonClosed
        {
            get
            {
                lock (this.reasonClosedLock)
                {
                    return this.reasonClosed;
                }
            }
            
            set
            {
                lock (this.reasonClosedLock)
                {
                    if ((object)this.reasonClosed == null)
                    {
                        this.reasonClosed = value;
                    }
                }
            }
        }
                
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.stdinStream.Close();
                this.stdoutStream.Close();
                this.stderrStream.Close();
            }
        }
    }
}
