﻿// <copyright file="ChannelCondition.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> ChannelCondition
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;

namespace Eunomia.SSH2
{
    /// <summary>
    /// Contains constants that can be used to specify what conditions to wait for on
    /// a SSH-2 channel (e.g., represented by a <see cref="Session"></see>.
    /// </summary>
    /// <see cref="Session.WaitForCondition(int, long)"/>
    [SuppressMessage("Microsoft.Performance", "CA1815:OverrideEqualsAndOperatorEqualsOnValueTypes",
                    Justification = "We will never compare or sort instances of this struct.")]
    public struct ChannelCondition
    {
        /// <summary>
        /// A timeout has occurred, none of your requested conditions is fulfilled.
        /// However, other conditions may be true - therefore, NEVER use the "=="
        /// operator to test for this (or any other) condition. Always use
        /// something like: <c>((cond &amp; ChannelCondition.CLOSED) != 0)</c>.
        /// </summary>
        public const int Timeout = 1;
        
        /// <summary>
        /// The underlying SSH-2 channel, however not necessarily the whole connection,
        /// has been closed. This implies <c>EOF</c>. Note that there may still
        /// be unread stdout or stderr data in the local window, i.e, <c>STDOUT_DATA</c>
        /// or/and <c>STDERR_DATA</c> may be set at the same time.
        /// </summary>
        public const int Closed = 2;
        
        /// <summary>
        /// There is stdout data available that is ready to be consumed.
        /// </summary>
        public const int StandardOutData = 4;
        
        /// <summary>
        /// There is stderr data available that is ready to be consumed.
        /// </summary>
        public const int StandardErrorData = 8;
        
        /// <summary>
        /// EOF on has been reached, no more _new_ stdout or stderr data will arrive
        /// from the remote server. However, there may be unread stdout or stderr
        /// data, i.e, <c>STDOUT_DATA</c> or/and <c>STDERR_DATA</c>
        /// may be set at the same time.
        /// </summary>
        public const int EOF = 16;
        
        /// <summary>
        /// The exit status of the remote process is available.
        /// Some servers never send the exist status, or occasionally "forget" to do so.
        /// </summary>
        public const int ExitStatus = 32;
        
        /// <summary>
        /// The exit signal of the remote process is available.
        /// </summary>
        public const int ExitSignal = 64;
    }
}
