// <copyright file="BlowFish.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> BlowFish
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

using Org.BouncyCastle.Crypto;
using Org.BouncyCastle.Crypto.Engines;
using Org.BouncyCastle.Crypto.Parameters;

namespace Eunomia.SSH2.Crypto.Cipher
{
    /// <summary>
    /// Class used for Blowfish ecryption.  Just a simple wrapper around BouncyCastle's Blowfish.
    /// </summary>
    /// <example>
    /// Blowfish blow = new Blowfish();
    /// </example>
    public class Blowfish : IBlockCipher
    {
        /// <summary>
        /// AlgorithmName is always "Blowfish".
        /// </summary>
        public const string AlgorithmName = "Blowfish";
        private const int BLOCK_SIZE = 8;
        private BlowfishEngine blw;

        /// <summary>
        /// Initializes a new instance of the Blowfish class.
        /// </summary>
        public Blowfish()
        {
            this.blw = new BlowfishEngine();
        }

        /// <summary> Gets the Block Size. </summary>
        /// <returns> 8 bytes for Blowfish. </returns>
        public int BlockSize
        {
            get { return BLOCK_SIZE; }
        }

        /// <summary>
        /// Initializes the BlowfishEngine.  You must call this before actually encrypting
        /// or decrypting data via calls to TransformBlock.
        /// </summary>
        /// <param name="forEncryption"> True for encryption, false for decryption. </param>
        /// <param name="key"> The key that we want to use. </param>
        public void Init(bool forEncryption, byte[] key)
        {
            ICipherParameters kparam = new KeyParameter(key);
            this.blw.Init(forEncryption, kparam);
        }

        /// <summary>
        /// Transforms a block of byte data and encrypts, or decrypts it depending on
        /// how the Init method was called.
        /// </summary>
        /// <param name="source"> Byte array of data to be processed.</param>
        /// <param name="sourceOffset"> Offset of where we begin processing, (normally 0).</param>
        /// <param name="destination"> Resulting byte array after encrypting/decrypting.</param>
        /// <param name="destinationOffset"> Offset to where the processed data is placed, (normally 0).</param>
        public void TransformBlock(byte[] source, int sourceOffset, byte[] destination, int destinationOffset)
        {
            if ((sourceOffset + BLOCK_SIZE) > source.Length)
            {
                throw new ArgumentException("input buffer too short");
            }
            
            if ((destinationOffset + BLOCK_SIZE) > destination.Length)
            {
                throw new ArgumentException("output buffer too short");
            }

            this.blw.ProcessBlock(source, sourceOffset, destination, destinationOffset);
        }
    }
}
