// <copyright file="CBCMode.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> CBCMode
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

namespace Eunomia.SSH2.Crypto.Cipher
{
    /// <summary>
    /// Class used for CBCMode ecryption.
    /// </summary>
    public class CBCMode : IBlockCipher
    {
        private IBlockCipher tc;
        private int blockSize;
        private bool doEncrypt;

        private byte[] cbc_vector;
        private byte[] tmp_vector;

        /// <summary>
        /// Initializes a new instance of the CBCMode class.
        /// </summary>
        /// <param name="underCipher">The underlying cipher to use.</param>
        /// <param name="iv">The IV to use.</param>
        /// <param name="doEncrypt">True for encryption, false for decryption.</param>
        public CBCMode(IBlockCipher underCipher, byte[] iv, bool doEncrypt)
        {
            this.tc = underCipher;
            this.blockSize = underCipher.BlockSize;
            this.doEncrypt = doEncrypt;

            if (this.blockSize != iv.Length)
            {
                throw new ArgumentException("IV must be " + this.blockSize
                                            + " bytes long! (currently " + iv.Length + ")");
            }
            
            this.cbc_vector = new byte[this.blockSize];
            this.tmp_vector = new byte[this.blockSize];
            Array.Copy(iv, 0, this.cbc_vector, 0, this.blockSize);
        }

        /// <summary>
        /// Gets the Block Size.
        /// </summary>
        /// <returns>The block size of the underlying cipher.</returns>
        public int BlockSize
        {
            get { return this.blockSize; }
        }
        
        /// <summary>
        /// Initializes the CBCMode cipher.  For CBC and CTR this does nothing.
        /// </summary>
        /// <param name="forEncryption">True for encryption, false for decryption.</param>
        /// <param name="key">The key that we want to use. </param>
        public void Init(bool forEncryption, byte[] key)
        {
        }

        /// <summary>
        /// Transforms a block of byte data and encrypts, or decrypts it depending on
        /// how the CBCMode instance was created.
        /// </summary>
        /// <param name="source"> Byte array of data to be processed.</param>
        /// <param name="sourceOffset"> Offset of where we begin processing, (normally 0).</param>
        /// <param name="destination"> Resulting byte array after encrypting/decrypting.</param>
        /// <param name="destinationOffset"> Offset to where the processed data is placed, (normally 0).</param>
        public void TransformBlock(byte[] source, int sourceOffset, byte[] destination, int destinationOffset)
        {
            if (this.doEncrypt)
            {
                this.EncryptBlock(source, sourceOffset, destination, destinationOffset);
            }
            else
            {
                this.DecryptBlock(source, sourceOffset, destination, destinationOffset);
            }
        }

        /// <summary>
        /// Encrypts a block of data.
        /// </summary>
        /// <param name="src">Byte array of data to be encrypted.</param>
        /// <param name="srcoff">Offset of where we begin encrypted, (normally 0).</param>
        /// <param name="dest">Resulting byte array after encrypting.</param>
        /// <param name="destoff">Offset to where the encrypted data is placed, (normally 0).</param>
        private void EncryptBlock(byte[] src, int srcoff, byte[] dest, int destoff)
        {
            for (int i = 0; i < this.blockSize; i++)
            {
                this.cbc_vector[i] ^= src[srcoff + i];
            }

            this.tc.TransformBlock(this.cbc_vector, 0, dest, destoff);

            Array.Copy(dest, destoff, this.cbc_vector, 0, this.blockSize);
        }

        /// <summary>
        /// Decrypts a block of data.
        /// </summary>
        /// <param name="src">Byte array of data to be decrypted.</param>
        /// <param name="srcoff">Offset of where we begin decrypting, (normally 0).</param>
        /// <param name="dest">Resulting byte array after decrypting.</param>
        /// <param name="destoff">Offset to where the decrypted data is placed, (normally 0).</param>
        private void DecryptBlock(byte[] src, int srcoff, byte[] dest, int destoff)
        {
            // Assume the worst, src and dst are overlapping...
            Array.Copy(src, srcoff, this.tmp_vector, 0, this.blockSize);

            this.tc.TransformBlock(src, srcoff, dest, destoff);

            for (int i = 0; i < this.blockSize; i++)
            {
                dest[destoff + i] ^= this.cbc_vector[i];
            }

            // ...that is why we need a tmp buffer.
            byte[] swap = this.cbc_vector;
            this.cbc_vector = this.tmp_vector;
            this.tmp_vector = swap;
        }
    }
}
