// <copyright file="CipherOutputStream.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> CipherOutputStream
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Crypto.Cipher
{
    /// <summary>
    /// A class that implements an encrypted Stream (output only).
    /// </summary>
    public class CipherOutputStream : Stream
    {
        private const int BUFF_SIZE = 2048;

        private IBlockCipher currentCipher;
        private Stream outStream;
        private byte[] buffer;
        private byte[] enc;
        private int blockSize;
        private int pos;
        private byte[] out_buffer = new byte[BUFF_SIZE];
        private int out_buffer_pos = 0;

        /// <summary>
        /// Initializes a new instance of the CipherOutputStream class.
        /// </summary>
        /// <param name="blockCipher">The block cipher to use for encryption.</param>
        /// <param name="outStream">The underlying Stream to use.</param>
        public CipherOutputStream(IBlockCipher blockCipher, Stream outStream)
        {
            this.outStream = outStream;
            this.ChangeCipher(blockCipher);
        }

        /// <summary>
        /// Gets a bool indicating that we can write to this stream.
        /// </summary>
        public override bool CanWrite
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we can never read from this stream.
        /// </summary>
        public override bool CanRead
        {
            get { return true; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we cannot seek on this stream.
        /// </summary>
        public override bool CanSeek
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets a bool indicating that we cannot timeout on this stream.
        /// </summary>
        public override bool CanTimeout
        {
            get { return false; }
        }
        
        /// <summary>
        /// Gets the length of the stream (just returns the buffer length).
        /// </summary>
        public override long Length
        {
            get
            {
                return BUFF_SIZE;
            }
        }
        
        /// <summary>
        /// Gets or sets...nothing (not implemented).
        /// </summary>
        public override long Position
        {
            get { return 0L; }
            set { }
        }

        /// <summary>
        /// Seeks on the stream (not implemented).
        /// </summary>
        /// <param name="offset">The offset we want to seek to.</param>
        /// <param name="origin">The origin that we will seek from.</param>
        /// <returns>The new position in the stream.</returns>
        public override long Seek(long offset, SeekOrigin origin)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Flush the stream.
        /// </summary>
        public override void Flush()
        {
            if (this.pos != 0)
            {
                throw new IOException("FATAL: cannot flush since crypto buffer is not aligned.");
            }

            if (this.out_buffer_pos > 0)
            {
                this.outStream.Write(this.out_buffer, 0, this.out_buffer_pos);
                this.out_buffer_pos = 0;
            }

            this.outStream.Flush();
        }
       
        /// <summary>
        /// Set the length of the stream (not implemented).
        /// </summary>
        /// <param name="value">The length we want.</param>
        public override void SetLength(long value)
        {
            throw new NotImplementedException();
        }

        /// <summary>
        /// This allows us to change the underlying cipher.
        /// </summary>
        /// <param name="bc">The block cipher that we now want to use.</param>
        public void ChangeCipher(IBlockCipher bc)
        {
            this.currentCipher = bc;
            this.blockSize = bc.BlockSize;
            this.buffer = new byte[this.blockSize];
            this.enc = new byte[this.blockSize];
            this.pos = 0;
        }

        /// <summary>
        /// Reads from the stream, (not implemented).
        /// </summary>
        /// <param name="buffer">The buffer to write the data into.</param>
        /// <param name="offset">The offset in the buffer to begin writing at.</param>
        /// <param name="count">The number of bytes to read.</param>
        /// <returns>The number of bytes that where read.</returns>
        public override int Read(byte[] buffer, int offset, int count)
        {
            throw new NotImplementedException();
        }
        
        /// <summary>
        /// Write to the stream.
        /// </summary>
        /// <param name="buffer">The buffer to read from.</param>
        /// <param name="offset">The offset into the buffer to read from.</param>
        /// <param name="count">The number of bytes to write.</param>
        public override void Write(byte[] buffer, int offset, int count)
        {
            while (count > 0)
            {
                int avail = this.blockSize - this.pos;
                int copy = Math.Min(avail, count);

                Array.Copy(buffer, offset, this.buffer, this.pos, copy);
                this.pos += copy;
                offset += copy;
                count -= copy;

                if (this.pos >= this.blockSize)
                {
                    this.WriteBlock();
                }
            }
        }

        /// <summary>
        /// Write a single byte (passed as an int).
        /// </summary>
        /// <param name="writeValue">The byte to write.</param>
        public void Write(int writeValue)
        {
            this.buffer[this.pos++] = (byte)writeValue;
            if (this.pos >= this.blockSize)
            {
                this.WriteBlock();
            }
        }

        /// <summary>
        /// Write a single byte without encryption.
        /// </summary>
        /// <param name="writeValue">The byte to write.</param>
        public void WritePlain(int writeValue)
        {
            if (this.pos != 0)
            {
                throw new IOException("Cannot write plain since crypto buffer is not aligned.");
            }

            this.InternalWrite(writeValue);
        }

        /// <summary>
        /// Write out a byte array without encryption.
        /// </summary>
        /// <param name="buffer">The buffer to read from.</param>
        /// <param name="off">The offset into the buffer to read from.</param>
        /// <param name="len">The number of bytes to write.</param>
        public void WritePlain(byte[] buffer, int off, int len)
        {
            if (this.pos != 0)
            {
                throw new IOException("Cannot write plain since crypto buffer is not aligned.");
            }

            this.InternalWrite(buffer, off, len);
        }

        /// <summary>
        /// Write out data from the internal buffer, with encryption to the underlying stream.
        /// </summary>
        private void WriteBlock()
        {
            try
            {
                this.currentCipher.TransformBlock(this.buffer, 0, this.enc, 0);
            }
            catch (Exception e)
            {
                throw new IOException("Error while decrypting block. " + e.Message);
            }

            this.InternalWrite(this.enc, 0, this.blockSize);
            this.pos = 0;
        }

        /// <summary>
        /// Adds data to the internal buffer and if full,
        /// writes that to the underlying stream.
        /// </summary>
        /// <param name="src">The buffer to read from.</param>
        /// <param name="off">The offset into the buffer to read from.</param>
        /// <param name="len">The number of bytes to write.</param>
        private void InternalWrite(byte[] src, int off, int len)
        {
            while (len > 0)
            {
                int space = BUFF_SIZE - this.out_buffer_pos;
                int copy = (len > space) ? space : len;

                Array.Copy(src, off, this.out_buffer, this.out_buffer_pos, copy);

                off += copy;
                this.out_buffer_pos += copy;
                len -= copy;

                if (this.out_buffer_pos >= BUFF_SIZE)
                {
                    this.outStream.Write(this.out_buffer, 0, BUFF_SIZE);
                    this.out_buffer_pos = 0;
                }
            }
        }

        /// <summary>
        /// Adds a single byte to the internal buffer and if full, writes the buffer to the
        /// underlying stream.
        /// </summary>
        /// <param name="b">The byte to write.</param>
        private void InternalWrite(int b)
        {
            this.out_buffer[this.out_buffer_pos++] = (byte)b;
            if (this.out_buffer_pos >= BUFF_SIZE)
            {
                this.outStream.Write(this.out_buffer, 0, BUFF_SIZE);
                this.out_buffer_pos = 0;
            }
        }
    }
}
