// <copyright file="DhGroupExchange.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> DhGroupExchange
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using Eunomia.SSH2.Crypto.Digest;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.DH
{
    /// <summary>
    /// DhGroupExchange class is all about Diffie-Hellman GEX related data and methods.
    /// </summary>
    public class DHGroupExchange
    {
        // Given by the standard
        private BigInteger p;
        private BigInteger g;

        // Client public and private
        private BigInteger e;
        private BigInteger x;

        // Server public
        private BigInteger f;

        // Shared secret
        private BigInteger k;

        /// <summary>
        /// Initializes a new instance of the DHGroupExchange class.
        /// </summary>
        /// <param name="p">This is a 'big safe' prime.</param>
        /// <param name="g">This is the 'generator'.</param>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "p", Justification = "Consistancy to ganymed project.")]
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "g", Justification = "Consistancy to ganymed project.")]
        public DHGroupExchange(BigInteger p, BigInteger g)
        {
            this.p = p;
            this.g = g;
        }

        /// <summary>
        /// Gets the 'public' version of our randomly generated key.
        /// </summary>
        /// <exception cref="InvalidOperationException">If still null.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "E", Justification = "Consistancy to ganymed project.")]
        public BigInteger E
        {
            get
            {
                if (this.e == null)
                {
                    throw new InvalidOperationException("Not initialized!");
                }

                return this.e;
            }
        }

        /// <summary>
        /// Gets the shared secret if it exists.
        /// </summary>
        /// <exception cref="InvalidOperationException">If still null.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "K", Justification = "Consistancy to ganymed project.")]
        public BigInteger K
        {
            get
            {
                if (this.k == null)
                {
                    throw new InvalidOperationException("Shared secret not yet known, need f first!");
                }

                return this.k;
            }
        }

        /// <summary>
        /// Gets or sets the server 'public' key and calculates the shared secret (k).
        /// </summary>
        /// <exception cref="InvalidOperationException">If our 'public key is not set yet.</exception>
        /// <exception cref="ArgumentException">If the passed value is negative.</exception>
        [SuppressMessage("Microsoft.Naming", "CA1704:IdentifiersShouldBeSpelledCorrectly", MessageId = "F", Justification = "Consistancy to ganymed project.")]
        public BigInteger F
        {
            get
            {
                return this.f;
            }
            
            set
            {
                if (this.e == null)
                {
                    throw new InvalidOperationException("Not initialized!");
                }

                BigInteger zero = BigInteger.Zero;

                if (zero.CompareTo(value) >= 0 || this.p.CompareTo(value) <= 0)
                {
                    throw new InvalidOperationException("Invalid f specified!");
                }

                this.f = value;
                this.k = value.ModPow(this.x, this.p);
            }
        }

        /// <summary>
        /// Initializes the DhGroupExchange data.
        /// </summary>
        /// <param name="random">A SecureRandom instance.</param>
        [SuppressMessage("Microsoft.Design", "CA1011:ConsiderPassingBaseTypesAsParameters", Justification = "SecureRandom is required.")]
        public void Init(SecureRandom random)
        {
            this.k = null;

            this.x = new BigInteger(this.p.BitLength - 1, random);
            this.e = this.g.ModPow(this.x, this.p);
        }

        /// <summary>
        /// Calulates the hash used for verification of the DH data.
        /// </summary>
        /// <param name="clientVersion">The client version in a byte 'string'.</param>
        /// <param name="serverVersion">The server version in a byte 'string'.</param>
        /// <param name="clientKexPayload">The client key exchange payload.</param>
        /// <param name="serverKexPayload">The server key exchange payload.</param>
        /// <param name="hostkey">The host key.</param>
        /// <param name="para">The DHGexParameters.</param>
        /// <returns>The calculated hash in a byte array.</returns>
        public byte[] CalculateH(
            byte[] clientVersion,
            byte[] serverVersion,
            byte[] clientKexPayload,
            byte[] serverKexPayload,
            byte[] hostkey,
            DHGexParameters para)
        {
            HashForSSH2Types hash = new HashForSSH2Types("SHA1");

            hash.UpdateByteString(clientVersion);
            hash.UpdateByteString(serverVersion);
            hash.UpdateByteString(clientKexPayload);
            hash.UpdateByteString(serverKexPayload);
            hash.UpdateByteString(hostkey);
            if (para.MinGroupLength > 0)
            {
                hash.UpdateUINT32(para.MinGroupLength);
            }

            hash.UpdateUINT32(para.PreferredGroupLength);
            if (para.MaxGroupLength > 0)
            {
                hash.UpdateUINT32(para.MaxGroupLength);
            }

            hash.UpdateBigInt(this.p);
            hash.UpdateBigInt(this.g);
            hash.UpdateBigInt(this.e);
            hash.UpdateBigInt(this.f);
            hash.UpdateBigInt(this.k);

            return hash.GetDigest();
        }
    }
}
