// <copyright file="MAC.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> MAC
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

namespace Eunomia.SSH2.Crypto.Digest
{
    /// <summary>
    /// MAC...Message Authentication Code.
    /// Convenience class to set up a MAC from the many posibilities.
    /// </summary>
    public sealed class MAC
    {
        private IDigest mac;
        private int size;

        /// <summary>
        /// Initializes a new instance of the MAC class.
        /// </summary>
        /// <param name="type">The MAC type.</param>
        /// <param name="key">THe MAC key.</param>
        public MAC(string type, byte[] key)
        {
            if (type == "hmac-sha1")
            {
                this.mac = new HMAC(new SHA1(), key, 20);
            }
            else if (type == "hmac-sha1-96")
            {
                this.mac = new HMAC(new SHA1(), key, 12);
            }
            else if (type == "hmac-md5")
            {
                this.mac = new HMAC(new MD5(), key, 16);
            }
            else if (type == "hmac-md5-96")
            {
                this.mac = new HMAC(new MD5(), key, 12);
            }
            else
            {
                throw new ArgumentException("Unkown algorithm " + type);
            }

            this.size = this.mac.DigestLength;
        }

        /// <summary>
        /// Gets or sets the digest size for this MAC.
        /// </summary>
        public int Size
        {
            get { return this.size; }
            set { this.size = value; }
        }

        /// <summary>
        /// Static method that goes through a string array of MACS and verifies that they are valid.
        /// </summary>
        /// <param name="macs">A string array of MAC posibilities.</param>
        /// <exception cref="ArgumentException">If the MAC name is unknown.</exception>
        public static void CheckMacList(string[] macs)
        {
            for (int i = 0; i < macs.Length; i++)
            {
                GetKeyLen(macs[i]);
            }
        }

        /// <summary>
        /// Static method that gets the key length for a specific MAC.
        /// </summary>
        /// <param name="type">A string with the MAC name to check.</param>
        /// <returns>The byte length of the MAC.</returns>
        public static int GetKeyLen(string type)
        {
            if (type == "hmac-sha1")
            {
                return 20;
            }

            if (type == "hmac-sha1-96")
            {
                return 20;
            }

            if (type == "hmac-md5")
            {
                return 16;
            }

            if (type == "hmac-md5-96")
            {
                return 16;
            }

            throw new ArgumentException("Unkown algorithm " + type);
        }

        /// <summary>
        /// Gets a list of supported MAC types (static method).
        /// </summary>
        /// <returns>A string array of supported MACs.</returns>
        public static string[] GetMacList()
        {
            // Higher Priority First
            return new string[] { "hmac-sha1-96", "hmac-sha1", "hmac-md5-96", "hmac-md5" };
        }

        /// <summary>
        /// Initializes the MAC.
        /// </summary>
        /// <param name="seq">Initial byte array to update the MAC.</param>
        public void InitMac(int seq)
        {
            this.mac.Reset();
            this.mac.Update(unchecked((byte)(seq >> 24)));
            this.mac.Update(unchecked((byte)(seq >> 16)));
            this.mac.Update(unchecked((byte)(seq >> 8)));
            this.mac.Update(unchecked((byte)seq));
        }

        /// <summary>
        /// Updates the MAC with a byte array.
        /// </summary>
        /// <param name="packetData">The byte array of data.</param>
        /// <param name="off">The starting offset in the data.</param>
        /// <param name="len">The number of bytes to process.</param>
        public void Update(byte[] packetData, int off, int len)
        {
            this.mac.Update(packetData, off, len);
        }

        /// <summary>
        /// Gets the MAC (digest) into a supplied byte array.
        /// </summary>
        /// <param name="barr">The byte array to return the MAC in.</param>
        /// <param name="off">The offset into the byte array where the MAC should be stored.</param>
        public void GetMac(byte[] barr, int off)
        {
            this.mac.Digest(barr, off);
        }
    }
}
