﻿// <copyright file="PEMDecoder.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PEMDecoder
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;
using System.Text;

using Eunomia.SSH2.Crypto.Cipher;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Signature;
using Eunomia.SSH2.Util;
using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Crypto
{
    /// <summary>
    /// PEM support class.
    /// </summary>
    public static class PEMDecoder
    {
        private const int PEM_RSA_PRIVATE_KEY = 1;
        private const int PEM_DSA_PRIVATE_KEY = 2;

        /// <summary>
        /// Gets a bool that indicates whether the PEM is encrypted or not.
        /// </summary>
        /// <param name="pemStructure">The PEMStructure to check.</param>
        /// <returns>True if the PEM is encrypted.</returns>
        public static bool IsPEMEncrypted(PEMStructure pemStructure)
        {
            if (pemStructure.ProcType == null)
            {
                return false;
            }

            if (pemStructure.ProcType.Length != 2)
            {
                throw new IOException("Unknown Proc-Type field.");
            }

            if ("4".Equals(pemStructure.ProcType[0]) == false)
            {
                throw new IOException("Unknown Proc-Type field (" + pemStructure.ProcType[0] + ")");
            }

            if ("ENCRYPTED".Equals(pemStructure.ProcType[1]))
            {
                return true;
            }

            return false;
        }

        /// <summary>
        /// Decrypt the PEM certificate, get private key.
        /// </summary>
        /// <param name="pem">The PEM certificate.</param>
        /// <param name="password">The password to use.</param>
        /// <returns>The private key (RSA or DSA).</returns>
        public static object Decode(char[] pem, string password)
        {
            PEMStructure ps = ParsePEM(pem);

            if (IsPEMEncrypted(ps))
            {
                if (password == null)
                {
                    throw new IOException("PEM is encrypted, but no password was specified");
                }

                DecryptPEM(ps, StringEncoder.GetBytes(password));
            }

            if (ps.PEMType == PEM_DSA_PRIVATE_KEY)
            {
                SimpleDERReader dr = new SimpleDERReader(ps.Data);

                byte[] seq = dr.ReadSequenceAsByteArray();

                if (dr.Available() != 0)
                {
                    throw new IOException("Padding in DSA PRIVATE KEY DER stream.");
                }

                dr.ResetInput(seq);

                BigInteger version = dr.ReadInt();

                if (version.CompareTo(BigInteger.Zero) != 0)
                {
                    throw new IOException("Wrong version (" + version + ") in DSA PRIVATE KEY DER stream.");
                }

                BigInteger p = dr.ReadInt();
                BigInteger q = dr.ReadInt();
                BigInteger g = dr.ReadInt();
                BigInteger y = dr.ReadInt();
                BigInteger x = dr.ReadInt();

                if (dr.Available() != 0)
                {
                    throw new IOException("Padding in DSA PRIVATE KEY DER stream.");
                }

                return new DSAPrivateKey(p, q, g, y, x);
            }

            if (ps.PEMType == PEM_RSA_PRIVATE_KEY)
            {
                SimpleDERReader dr = new SimpleDERReader(ps.Data);

                byte[] seq = dr.ReadSequenceAsByteArray();

                if (dr.Available() != 0)
                {
                    throw new IOException("Padding in RSA PRIVATE KEY DER stream.");
                }

                dr.ResetInput(seq);

                BigInteger version = dr.ReadInt();

                if ((version.CompareTo(BigInteger.Zero) != 0) && (version.CompareTo(BigInteger.One) != 0))
                {
                    throw new IOException("Wrong version (" + version + ") in RSA PRIVATE KEY DER stream.");
                }

                BigInteger n = dr.ReadInt();
                BigInteger e = dr.ReadInt();
                BigInteger d = dr.ReadInt();

                return new RSAPrivateKey(d, e, n);
            }

            throw new IOException("PEM problem: it is of unknown type");
        }
        
        /// <summary>
        /// Convert a hex string to a byte array.
        /// </summary>
        /// <param name="hex">The hex string to convert.</param>
        /// <returns>The resulting byte array.</returns>
        private static byte[] HexToByteArray(string hex)
        {
            if (hex == null)
            {
                throw new ArgumentException("null argument");
            }

            if ((hex.Length % 2) != 0)
            {
                throw new ArgumentException("Uneven string length in hex encoding.");
            }

            byte[] decoded = new byte[hex.Length / 2];

            for (int i = 0; i < decoded.Length; i++)
            {
                int hi = HexToInt(hex[i * 2]);
                int lo = HexToInt(hex[(i * 2) + 1]);

                decoded[i] = (byte)((hi * 16) + lo);
            }

            return decoded;
        }

        /// <summary>
        /// Utility method to convert a hex char to an integer.
        /// </summary>
        /// <param name="c">The char to convert.</param>
        /// <returns>The converteed integer.</returns>
        private static int HexToInt(char c)
        {
            if ((c >= 'a') && (c <= 'f'))
            {
                return (c - 'a') + 10;
            }

            if ((c >= 'A') && (c <= 'F'))
            {
                return (c - 'A') + 10;
            }

            if ((c >= '0') && (c <= '9'))
            {
                return c - '0';
            }

            throw new ArgumentException("Need hex char");
        }

        /// <summary>
        /// Generate a key from a password and salt, with MD5.
        /// </summary>
        /// <param name="password">The password to use.</param>
        /// <param name="salt">The salt to use.</param>
        /// <param name="keyLen">The key length.</param>
        /// <returns>A byte array with the key.</returns>
        private static byte[] GenerateKeyFromPasswordSaltWithMD5(byte[] password, byte[] salt, int keyLen)
        {
            if (salt.Length < 8)
            {
                throw new ArgumentException("Salt needs to be at least 8 bytes for key generation.");
            }

            MD5 md5 = new MD5();

            byte[] key = new byte[keyLen];
            byte[] tmp = new byte[md5.DigestLength];

            while (true)
            {
                md5.Update(password, 0, password.Length);
                md5.Update(salt, 0, 8); // ARGH we only use the first 8 bytes of the salt in this step.
                // This took me two hours until I got AES-xxx running.
                int copy = (keyLen < tmp.Length) ? keyLen : tmp.Length;

                md5.Digest(tmp, 0);

                Array.Copy(tmp, 0, key, key.Length - keyLen, copy);

                keyLen -= copy;

                if (keyLen == 0)
                {
                    return key;
                }

                md5.Update(tmp, 0, tmp.Length);
            }
        }

        /// <summary>
        /// Remove RFC 1423/PKCS #7 padding.
        /// </summary>
        /// <param name="buff">The byte buffer to work on.</param>
        /// <param name="blockSize">The block size.</param>
        /// <returns>The buffer stripped of the padding.</returns>
        private static byte[] RemovePadding(byte[] buff, int blockSize)
        {
            // Removes RFC 1423/PKCS #7 padding
            int rfc_1423_padding = buff[buff.Length - 1] & 0xff;

            if ((rfc_1423_padding < 1) || (rfc_1423_padding > blockSize))
            {
                throw new IOException("Decrypted PEM has wrong padding, did you specify the correct password?");
            }

            for (int i = 2; i <= rfc_1423_padding; i++)
            {
                if (buff[buff.Length - i] != rfc_1423_padding)
                {
                    throw new IOException("Decrypted PEM has wrong padding, did you specify the correct password?");
                }
            }

            byte[] tmp = new byte[buff.Length - rfc_1423_padding];
            Array.Copy(buff, 0, tmp, 0, buff.Length - rfc_1423_padding);
            return tmp;
        }

        /// <summary>
        /// Parses the PEM style certificate.
        /// </summary>
        /// <param name="pem">The byte array containing the PEM.</param>
        /// <returns>A PEMStructure with the PEM information in it.</returns>
        private static PEMStructure ParsePEM(char[] pem)
        {
            PEMStructure ps = new PEMStructure();

            string line = null;

            StringReader br = new StringReader(pem.ToString());

            string endLine = null;

            while (true)
            {
                line = br.ReadLine();

                if (line == null)
                {
                    throw new IOException("Invalid PEM structure, '-----BEGIN...' missing");
                }

                line = line.Trim();

                if (line.StartsWith("-----BEGIN DSA PRIVATE KEY-----", StringComparison.Ordinal))
                {
                    endLine = "-----END DSA PRIVATE KEY-----";
                    ps.PEMType = PEM_DSA_PRIVATE_KEY;
                    break;
                }

                if (line.StartsWith("-----BEGIN RSA PRIVATE KEY-----", StringComparison.Ordinal))
                {
                    endLine = "-----END RSA PRIVATE KEY-----";
                    ps.PEMType = PEM_RSA_PRIVATE_KEY;
                    break;
                }
            }

            while (true)
            {
                line = br.ReadLine();

                if (line == null)
                {
                    throw new IOException("Invalid PEM structure, " + endLine + " missing");
                }

                line = line.Trim();

                int sem_idx = line.IndexOf(':');

                if (sem_idx == -1)
                {
                    break;
                }

                string name = line.Substring(0, sem_idx + 1);
                string value = line.Substring(sem_idx + 1);

                string[] values = value.Split(',');

                for (int i = 0; i < values.Length; i++)
                {
                    values[i] = values[i].Trim();
                }

                // Proc-Type: 4,ENCRYPTED
                // DEK-Info: DES-EDE3-CBC,579B6BE3E5C60483
                if ("Proc-Type:".Equals(name))
                {
                    ps.ProcType = values;
                    continue;
                }

                if ("DEK-Info:".Equals(name))
                {
                    ps.DekInfo = values;
                    continue;
                }

                // Ignore line
            }

            StringBuilder keyData = new StringBuilder();

            while (true)
            {
                if (line == null)
                {
                    throw new IOException("Invalid PEM structure, " + endLine + " missing");
                }

                line = line.Trim();

                if (line.StartsWith(endLine, StringComparison.Ordinal))
                {
                    break;
                }

                keyData.Append(line);

                line = br.ReadLine();
            }

            char[] pem_chars = new char[keyData.Length];
            keyData.CopyTo(0, pem_chars, 0, pem_chars.Length);

            // ps.data = Base64.decode(pem_chars);
            ps.Data = Convert.FromBase64CharArray(pem_chars, 0, pem_chars.Length);

            if (ps.Data.Length == 0)
            {
                throw new IOException("Invalid PEM structure, no data available");
            }

            return ps;
        }

        /// <summary>
        /// Decrypt the PEM.
        /// </summary>
        /// <param name="ps">The PEMStructure to decrypt.</param>
        /// <param name="pw">The password in byte array form.</param>
        private static void DecryptPEM(PEMStructure ps, byte[] pw)
        {
            if (ps.DekInfo == null)
            {
                throw new IOException("Broken PEM, no mode and salt given, but encryption enabled");
            }

            if (ps.DekInfo.Length != 2)
            {
                throw new IOException("Broken PEM, DEK-Info is incomplete!");
            }
            
            string algo = ps.DekInfo[0];
            byte[] salt = HexToByteArray(ps.DekInfo[1]);

            IBlockCipher bc = null;

            if (algo.Equals("DES-EDE3-CBC"))
            {
                DESEDE des3 = new DESEDE();
                des3.Init(false, GenerateKeyFromPasswordSaltWithMD5(pw, salt, 24));
                bc = new CBCMode(des3, salt, false);
            }
            else if (algo.Equals("DES-CBC"))
            {
                DES des = new DES();
                des.Init(false, GenerateKeyFromPasswordSaltWithMD5(pw, salt, 8));
                bc = new CBCMode(des, salt, false);
            }
            else if (algo.Equals("AES-128-CBC"))
            {
                AES aes = new AES();
                aes.Init(false, GenerateKeyFromPasswordSaltWithMD5(pw, salt, 16));
                bc = new CBCMode(aes, salt, false);
            }
            else if (algo.Equals("AES-192-CBC"))
            {
                AES aes = new AES();
                aes.Init(false, GenerateKeyFromPasswordSaltWithMD5(pw, salt, 24));
                bc = new CBCMode(aes, salt, false);
            }
            else if (algo.Equals("AES-256-CBC"))
            {
                AES aes = new AES();
                aes.Init(false, GenerateKeyFromPasswordSaltWithMD5(pw, salt, 32));
                bc = new CBCMode(aes, salt, false);
            }
            else
            {
                throw new IOException("Cannot decrypt PEM structure, unknown cipher " + algo);
            }

            if ((ps.Data.Length % bc.BlockSize) != 0)
            {
                throw new IOException("Invalid PEM structure, size of encrypted block is not a multiple of "
                                      + bc.BlockSize);
            }
            
            // Now decrypt the content
            byte[] dz = new byte[ps.Data.Length];

            for (int i = 0; i < ps.Data.Length / bc.BlockSize; i++)
            {
                bc.TransformBlock(ps.Data, i * bc.BlockSize, dz, i * bc.BlockSize);
            }

            // Now check and remove RFC 1423/PKCS #7 padding
            dz = RemovePadding(dz, bc.BlockSize);

            ps.Data = dz;
            ps.DekInfo = null;
            ps.ProcType = null;
        }
    }
}
