﻿// <copyright file="HTTPProxyData.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> HTTPProxyData
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>HTTPProxyData</c> object is used to specify the needed connection data
    /// to connect through a HTTP proxy.
    /// </summary>
    /// <seealso cref="Connection.ProxyData"></seealso>
    public class HTTPProxyData : ProxyData
    {
        private string proxyHost;
        private int proxyPort;
        private string proxyUser;
        private string proxyPass;
        private string[] requestHeaderLines;
        
        /// <summary>
        /// Initializes a new instance of the HTTPProxyData class. 
        /// Same as calling <c>HTTPProxyData(proxyHost, proxyPort, null, null).</c>
        /// </summary>
        /// <param name="proxyHost">The proxy host name or ip address.</param>
        /// <param name="proxyPort">The proxy port.</param>
        public HTTPProxyData(string proxyHost, int proxyPort) : this(proxyHost, proxyPort, null, null)
        {
        }
        
        /// <summary>
        /// Initializes a new instance of the HTTPProxyData class. 
        /// Same as calling <c>HTTPProxyData(proxyHost, proxyPort, null, null, null}.</c>
        /// </summary>
        /// <param name="proxyHost">The proxy host name or ip address.</param>
        /// <param name="proxyPort">The proxy port.</param>
        /// <param name="proxyUser">Username for basic authentication (<c>null</c> if no authentication is needed).</param>
        /// <param name="proxyPass">Password for basic authentication (<c>null</c> if no authentication is needed).</param>
        public HTTPProxyData(string proxyHost, int proxyPort, string proxyUser, string proxyPass) :
            this(proxyHost, proxyPort, proxyUser, proxyPass, null)
        {
        }

        /// <summary>
        /// Initializes a new instance of the HTTPProxyData class. 
        /// Connection data for a HTTP proxy. It is possible to specify a username and password
        /// if the proxy requires basic authentication. Also, additional request header lines can
        /// be specified (e.g., "User-Agent: CERN-LineMode/2.15 libwww/2.17b3").
        /// <para>
        /// Please note: if you want to use basic authentication, then both <c>proxyUser</c>
        /// and <c>proxyPass</c> must be non-null.
        /// </para>
        /// </summary>
        /// <example>
        /// new HTTPProxyData("192.168.1.1", "3128", "proxyuser", "secret", new string[] {"User-Agent: EunomiaBasedClient/1.0", "X-My-Proxy-Option: something"});
        /// </example>
        /// <param name="proxyHost">The proxy host name or ip address.</param>
        /// <param name="proxyPort">The proxy port.</param>
        /// <param name="proxyUser">Username for basic authentication (<c>null</c> if no authentication is needed).</param>
        /// <param name="proxyPass">Password for basic authentication (<c>null</c> if no authentication is needed).</param>
        /// <param name="requestHeaderLines">An array with additional request header lines (without end-of-line markers)
        /// that have to be sent to the server. May be <c>null</c>.</param>
        public HTTPProxyData(
            string proxyHost,
            int proxyPort,
            string proxyUser,
            string proxyPass,
            string[] requestHeaderLines)
        {
            if (proxyHost == null)
            {
                throw new ArgumentException("proxyHost must be non-null");
            }
            
            if (proxyPort < 0)
            {
                throw new ArgumentException("proxyPort must be non-negative");
            }
            
            this.proxyHost = proxyHost;
            this.proxyPort = proxyPort;
            this.proxyUser = proxyUser;
            this.proxyPass = proxyPass;
            this.requestHeaderLines = requestHeaderLines;
        }
        
        /// <summary>
        /// Gets the proxy hostName or ip address.
        /// </summary>
        public string ProxyHost
        {
            get { return this.proxyHost; }
        }
        
        /// <summary>
        /// Gets the proxy port.
        /// </summary>
        public int ProxyPort
        {
            get { return this.proxyPort; }
        }
        
        /// <summary>
        /// Gets the proxy user name (for basic authentication).
        /// </summary>
        public string ProxyUser
        {
            get { return this.proxyUser; }
        }
        
        /// <summary>
        /// Gets the proxy password (for basic authentication).
        /// </summary>
        public string ProxyPass
        {
            get { return this.proxyPass; }
        }
        
        /// <summary>
        /// Gets the additional request header lines (without end-of-line markers), that have to be sent to the server.
        /// </summary>
        public string[] RequestHeaderLines
        {
            get { return this.requestHeaderLines; }
        }
    }
}
