﻿// <copyright file="InteractiveCallback.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> InteractiveCallback
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;

namespace Eunomia.SSH2
{
    /// <summary>
    /// An <c>InteractiveCallback</c> is used to respond to challenges sent
    /// by the server if authentication mode "keyboard-interactive" is selected.
    /// </summary>
    public interface IInteractiveCallback
    {
        /// <summary>
        /// This callback interface is used during a "keyboard-interactive"
        /// authentication. Every time the server sends a set of challenges (however,
        /// most often just one challenge at a time), this callback function will be
        /// called to give your application a chance to talk to the user and to
        /// determine the response(s).
        /// <para>
        /// Some copy-paste information from the standard: a command line interface
        /// (CLI) client SHOULD print the name and instruction (if non-empty), adding
        /// newlines. Then for each prompt in turn, the client SHOULD display the
        /// prompt and read the user input. The name and instruction fields MAY be
        /// empty strings, the client MUST be prepared to handle this correctly. The
        /// prompt field(s) MUST NOT be empty strings.
        /// </para>
        /// <para>
        /// Please refer to draft-ietf-secsh-auth-kbdinteract-XX.txt for the details.
        /// </para>
        /// <para>
        /// Note: clients SHOULD use control character filtering as discussed in
        /// RFC4251 to avoid attacks by including
        /// terminal control characters in the fields to be displayed.
        /// </para>
        /// </summary>
        /// <param name="name">The name string sent by the server.</param>
        /// <param name="instruction">The instruction string sent by the server.</param>
        /// <param name="numberOfPrompts">
        /// The number of prompts - may be zero (in this case, you should just
        /// return a string array of length zero).
        /// </param>
        /// <param name="prompt">An array (length numPrompts), of strings.</param>
        /// <param name="echo">
        /// An array (length numberOfPrompts of booleans. For
        /// each prompt, the corresponding echo field indicates whether or
        /// not the user input should be echoed as characters are typed.
        /// </param>
        /// <returns>
        /// An array of reponses - the array size must match the parameter
        /// numberOfPrompts.
        /// </returns>
        string[] ReplyToChallenge(string name, string instruction, int numberOfPrompts, string[] prompt, bool[] echo);
    }
}
