// <copyright file="KnownHosts.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> KnownHosts
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Collections;
using System.Collections.Generic;
using System.Diagnostics.CodeAnalysis;
using System.Globalization;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;

using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Signature;
using Eunomia.SSH2.Util;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2
{
    /// <summary>
    /// The <c>KnownHosts</c> class is a handy tool to verify received server hostkeys
    /// <para>
    /// It offers basically an in-memory database for known_hosts entries, as well as some
    /// helper functions. Entries from a <c>known_hosts</c> file can be loaded at construction time.
    /// It is also possible to add more keys later (e.g., one can parse different
    /// known_hosts files).
    /// </para><para>
    /// It is a thread safe implementation, therefore, you need only to instantiate one
    /// <c>KnownHosts</c> for your whole application.
    /// </para>
    /// </summary>
    public class KnownHosts
    {
        private const int HOSTKEY_IS_OK = 0;
        private const int HOSTKEY_IS_NEW = 1;
        private const int HOSTKEY_HAS_CHANGED = 2;

        private LinkedList<KnownHostsEntry> publicKeys = new LinkedList<KnownHostsEntry>();

        /// <summary>
        /// Initializes a new instance of the KnownHosts class.
        /// </summary>
        public KnownHosts()
        {
        }

        /// <summary>
        /// Initializes a new instance of the KnownHosts class.
        /// Reads the initial hosts data from a character array.
        /// </summary>
        /// <param name="knownHostsData">The initial hosts data in a character array.</param>
        public KnownHosts(char[] knownHostsData)
        {
            this.Initialize(knownHostsData);
        }

        /// <summary>
        /// Initializes a new instance of the KnownHosts class.
        /// Reads the initial hosts data from a file (FileStream).
        /// </summary>
        /// <param name="knownHosts">A FileStream to a file containing the initial hosts data.</param>
        public KnownHosts(FileStream knownHosts)
        {
            this.Initialize(knownHosts);
        }

        /// <summary>
        /// Convert a ssh2 key-blob into a human readable hex fingerprint.
        /// Generated fingerprints are identical to those generated by OpenSSH.
        /// <para>
        /// Example fingerprint: d0:cb:76:19:99:5a:03:fc:73:10:70:93:f2:44:63:47.
        /// </para>
        /// </summary>
        /// <param name="keyType">Either "ssh-rsa" or "ssh-dss".</param>
        /// <param name="publicKey">The pubic key "blob".</param>
        /// <returns>The hex fingerprint.</returns>
        public static string CreateHexFingerprint(string keyType, byte[] publicKey)
        {
            byte[] raw = RawFingerPrint("md5", keyType, publicKey);
            return RawToHexFingerprint(raw);
        }
        
        /// <summary>
        /// Convert a ssh2 key-blob into a human readable bubblebabble fingerprint.
        /// that are easier to remember for humans.
        /// <para>
        /// Example fingerprint: xofoc-bubuz-cazin-zufyl-pivuk-biduk-tacib-pybur-gonar-hotat-lyxux.
        /// </para>
        /// </summary>
        /// <param name="keyType">Either "ssh-rsa" or "ssh-dss".</param>
        /// <param name="publicKey">The pubic key "blob".</param>
        /// <returns>The BubbleBabble fingerprint.</returns>
        public static string CreateBubbleBabbleFingerprint(string keyType, byte[] publicKey)
        {
            byte[] raw = RawFingerPrint("sha1", keyType, publicKey);
            return RawToBubbleBabbleFingerprint(raw);
        }

        /// <summary>
        /// Generate the hashed representation of the given hostName. Useful for adding entries
        /// with hashed hostNames to a knownHosts file. (see -H option of OpenSSH key-gen).
        /// </summary>
        /// <param name="hostName">The host name or ip address of a host.</param>
        /// <returns>The hashed representation, "|1|cDhrv7zwEUV3k71CEPHnhHZezhA=|Xo+2y6rUXo2OIWRAYhBOIijbJMA="
        /// for example.</returns>
        public static string CreateHashedHostName(string hostName)
        {
            SHA1 sha1 = new SHA1();

            byte[] salt = new byte[sha1.DigestLength];

            new SecureRandom().NextBytes(salt);

            byte[] hash = HmacSha1Hash(salt, hostName);

            string base64_salt = Convert.ToBase64String(salt);

            string base64_hash = Convert.ToBase64String(hash);

            return "|1|" + base64_salt + "|" + base64_hash;
        }

        /// <summary>
        /// Adds a single public key entry to the a known_hosts file.
        /// This method is designed to be used in a <c>ServerHostKeyVerifier</c>.
        /// </summary>
        /// <param name="knownHosts">The file (FileInfo) where the publickey entry will be appended.</param>
        /// <param name="hostNames">
        /// A list of hostName patterns - at least one most be specified. Check out the
        /// OpenSSH sshd man page for a description of the pattern matching algorithm.
        /// </param>
        /// <param name="serverHostkeyAlgorithm">The host key alogorithm passed to the <c>ServerHostKeyVerifier</c>.</param>
        /// <param name="serverHostkey">The host key passed to the <c>ServerHostKeyVerifier</c>.</param>
        public static void AddHostkeyToFile(
            FileSystemInfo knownHosts,
            string[] hostNames,
            string serverHostkeyAlgorithm,
            byte[] serverHostkey)
        {
            if ((hostNames == null) || (hostNames.Length == 0))
            {
                throw new ArgumentException("Need at least one hostName specification");
            }

            if ((serverHostkeyAlgorithm == null) || (serverHostkey == null))
            {
                throw new ArgumentException("AddHostkeyToFile at least one string array is null.");
            }

            StreamWriter writer = new StreamWriter(new MemoryStream());

            for (int i = 0; i < hostNames.Length; i++)
            {
                if (i != 0)
                {
                    writer.Write(',');
                }
                
                writer.Write(hostNames[i]);
            }

            writer.Write(' ');
            writer.Write(serverHostkeyAlgorithm);
            writer.Write(' ');

            // writer.Write(Base64.encode(serverHostkey));
            writer.Write(Convert.ToBase64String(serverHostkey));
            writer.Write("\n");

            char[] entry = writer.ToString().ToCharArray();

            FileStream raf = new FileStream(knownHosts.FullName, FileMode.Open, FileAccess.ReadWrite);

            long len = raf.Length;

            if (len > 0)
            {
                raf.Seek(len, SeekOrigin.Begin);
                int last = raf.ReadByte();
                if (last != '\n')
                {
                    raf.WriteByte((byte)'\n');
                }
            }

            raf.Write(StringEncoder.GetBytes(new string(entry)), 0, entry.Length);
            raf.Close();
        }
        
        /// <summary>
        /// Parses the given known_hosts data and adds entries to the database.
        /// </summary>
        /// <param name="knownHostsData">A character array containing at least one entry for a known host.</param>
        /// <exception cref="IOException">If the data is not readable.</exception>
        public void AddHostKeys(char[] knownHostsData)
        {
            this.Initialize(knownHostsData);
        }

        /// <summary>
        /// Parses the given knownHosts file and adds entries to the database.
        /// </summary>
        /// <param name="knownHosts">A FleStream pointing to a file with known host data in it.</param>
        /// <exception cref="IOException">If there is a problem with the file or the data is not readable.</exception>
        public void AddHostKeys(FileStream knownHosts)
        {
            this.Initialize(knownHosts);
        }

        /// <summary>
        /// Adds a single public key entry to the database. Note: this will NOT add the public key
        /// to any physical file (e.g., "~/.ssh/known_hosts") - use <c>AddHostkeyToFile()</c> for that purpose.
        /// This method is designed to be used in a <c>ServerHostKeyVerifier</c>.
        /// </summary>
        /// <param name="hostNames">
        /// A list of hostName patterns - at least one most be specified. Check out the
        /// OpenSSH sshd man page for a description of the pattern matching algorithm.
        /// </param>
        /// <param name="serverHostkeyAlgorithm">The host key alogorithm passed to the <c>ServerHostKeyVerifier</c>.</param>
        /// <param name="serverHostkey">The host key passed to the <c>ServerHostKeyVerifier</c>.</param>
        /// <exception cref="IOException">If the host key type is unknown.</exception>
        public void AddHostkey(string[] hostNames, string serverHostkeyAlgorithm, byte[] serverHostkey)
        {
            if ("ssh-rsa".Equals(serverHostkeyAlgorithm))
            {
                RSAPublicKey rpk = RSASHA1Verify.DecodeSSHRSAPublicKey(serverHostkey);

                lock (this.publicKeys)
                {
                    this.publicKeys.AddLast(new KnownHostsEntry(hostNames, rpk));
                }
            }
            else if ("ssh-dss".Equals(serverHostkeyAlgorithm))
            {
                DSAPublicKey dpk = DSASHA1Verify.DecodeSSHDSAPublicKey(serverHostkey);

                lock (this.publicKeys)
                {
                    this.publicKeys.AddLast(new KnownHostsEntry(hostNames, dpk));
                }
            }
            else
            {
                throw new IOException("Unknwon host key type (" + serverHostkeyAlgorithm + ")");
            }
        }

        /// <summary>
        /// Try to find the preferred order of hostkey algorithms for the given hostName.
        /// Based on the type of hostkey that is present in the internal database
        /// (i.e., either <c>ssh-rsa</c> or <c>ssh-dss</c>)
        /// an ordered list of hostkey algorithms is returned which can be passed
        /// to <c>Connection.ServerHostkeyAlgorithms</c>.
        /// </summary>
        /// <param name="hostName">The host name.</param>
        /// <returns>
        /// If no key for the given hostName is present or there are keys of multiple types
        /// present for the given hostName it will return <c>null</c>. Otherwise,
        /// an array with hostkey algorithms is returned (i.e., an array of length 2).
        /// </returns>
        public string[] GetPreferredServerHostkeyAlgorithmOrder(string hostName)
        {
            string[] algos = this.RecommendHostkeyAlgorithms(hostName);

            if (algos != null)
            {
                return algos;
            }

            IPAddress ipAddress = null;

            try
            {
                ipAddress = IPAddress.Parse(hostName);
            }
            catch (FormatException)
            {
                return null;
            }

            algos = this.RecommendHostkeyAlgorithms(ipAddress.ToString());
            if (algos != null)
            {
                return algos;
            }

            return null;
        }

        /// <summary>
        /// Checks the internal hostkey database for the given hostkey.
        /// If no matching key can be found, then the hostName is resolved to an IP address
        /// and the search is repeated using that IP address.
        /// </summary>
        /// <param name="hostName">The server's hostName, will be matched with all hostName patterns.</param>
        /// <param name="serverHostkeyAlgorithm">The host key alogorithm passed to the <c>ServerHostKeyVerifier</c>.</param>
        /// <param name="serverHostkey">The host key passed to the <c>ServerHostKeyVerifier</c>.</param>
        /// <returns>
        /// <list type="bullet">
        /// <item><c>HOSTKEY_IS_OK</c>: the given hostkey matches an entry for the given hostName.</item>
        /// <item><c>HOSTKEY_HAS_CHANGED</c>: hostName is known, but with another key of the same type
        /// (man-in-the-middle attack?).</item>
        /// </list>
        /// </returns>
        /// <exception cref="ArgumentException">If the host key type is unknown.</exception>
        public int VerifyHostkey(string hostName, string serverHostkeyAlgorithm, byte[] serverHostkey)
        {
            object remoteKey = null;

            if ("ssh-rsa".Equals(serverHostkeyAlgorithm))
            {
                remoteKey = RSASHA1Verify.DecodeSSHRSAPublicKey(serverHostkey);
            }
            else if ("ssh-dss".Equals(serverHostkeyAlgorithm))
            {
                remoteKey = DSASHA1Verify.DecodeSSHDSAPublicKey(serverHostkey);
            }
            else
            {
                throw new ArgumentException("Unknown hostkey type " + serverHostkeyAlgorithm);
            }

            int result = this.CheckKey(hostName, remoteKey);

            if (result == HOSTKEY_IS_OK)
            {
                return result;
            }

            IPAddress[] ipAdresses = null;

            try
            {
                ipAdresses = Dns.GetHostEntry(hostName).AddressList;
            }
            catch (SocketException)
            {
                return result;
            }

            for (int i = 0; i < ipAdresses.Length; i++)
            {
                int newresult = this.CheckKey(Dns.GetHostEntry(Dns.GetHostName()).AddressList[0].ToString(), remoteKey);

                if (newresult == HOSTKEY_IS_OK)
                {
                    return newresult;
                }

                if (newresult == HOSTKEY_HAS_CHANGED)
                {
                    result = HOSTKEY_HAS_CHANGED;
                }
            }

            return result;
        }

        /// <summary>
        /// Generates a "raw" fingerprint of a hostkey.
        /// </summary>
        /// <param name="type">Either "md5" or "sha1".</param>
        /// <param name="keyType">Either "ssh-rsa" or "ssh-dss".</param>
        /// <param name="hostkey">The hostkey.</param>
        /// <returns>The raw fingerprint.</returns>
        /// <exception cref="ArgumentException">If there is a problem with one of the passed arguments.</exception>
        private static byte[] RawFingerPrint(string type, string keyType, byte[] hostkey)
        {
            IDigest dig = null;

            if ("md5".Equals(type))
            {
                dig = new MD5();
            }
            else if ("sha1".Equals(type))
            {
                dig = new SHA1();
            }
            else
            {
                throw new ArgumentException("Unknown hash type " + type);
            }

            if ("ssh-rsa".Equals(keyType))
            {
            }
            else if ("ssh-dss".Equals(keyType))
            {
            }
            else
            {
                throw new ArgumentException("Unknown key type " + keyType);
            }

            if (hostkey == null)
            {
                throw new ArgumentException("hostkey is null");
            }

            dig.Update(hostkey);
            byte[] res = new byte[dig.DigestLength];
            dig.Digest(res);
            return res;
        }

        /// <summary>
        /// Convert a raw fingerprint to hex representation (XX:YY:ZZ...).
        /// </summary>
        /// <param name="fingerprint">The raw fingerprint.</param>
        /// <returns>The hex representation.</returns>
        private static string RawToHexFingerprint(byte[] fingerprint)
        {
            char[] alpha = "0123456789abcdef".ToCharArray();

            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < fingerprint.Length; i++)
            {
                if (i != 0)
                {
                    sb.Append(':');
                }
                
                int b = fingerprint[i] & 0xff;
                sb.Append(alpha[unchecked(b >> 4)]);
                sb.Append(alpha[b & 15]);
            }

            return sb.ToString();
        }

        /// <summary>
        /// Convert a raw fingerprint to bubblebabble representation.
        /// </summary>
        /// <param name="raw">The raw fingerprint.</param>
        /// <returns>The bubblebabble representation.</returns>
        private static string RawToBubbleBabbleFingerprint(byte[] raw)
        {
            char[] v = "aeiouy".ToCharArray();
            char[] c = "bcdfghklmnprstvzx".ToCharArray();

            StringBuilder sb = new StringBuilder();

            int seed = 1;

            int rounds = (raw.Length / 2) + 1;

            sb.Append('x');

            for (int i = 0; i < rounds; i++)
            {
                if (((i + 1) < rounds) || (raw.Length % 2 != 0))
                {
                    sb.Append(v[(((raw[2 * i] >> 6) & 3) + seed) % 6]);
                    sb.Append(c[(raw[2 * i] >> 2) & 15]);
                    sb.Append(v[((raw[2 * i] & 3) + (seed / 6)) % 6]);

                    if ((i + 1) < rounds)
                    {
                        sb.Append(c[(raw[(2 * i) + 1] >> 4) & 15]);
                        sb.Append('-');
                        sb.Append(c[raw[(2 * i) + 1] & 15]);

                        // As long as seed >= 0, seed will be >= 0 afterwards
                        seed = ((seed * 5) + (((raw[2 * i] & 0xff) * 7) + (raw[(2 * i) + 1] & 0xff))) % 36;
                    }
                }
                else
                {
                    sb.Append(v[seed % 6]); // seed >= 0, therefore index positive
                    sb.Append('x');
                    sb.Append(v[seed / 6]);
                }
            }

            sb.Append('x');

            return sb.ToString();
        }

        /// <summary>
        /// Hashs a host name using HMAC/SHA1 and returns the hash in a byte array.
        /// </summary>
        /// <param name="salt">The 'salt' to use.</param>
        /// <param name="hostName">The host name.</param>
        /// <returns>The resulting hash.</returns>
        /// <exception cref="ArgumentException">If the salt is not the right length.</exception>
        private static byte[] HmacSha1Hash(byte[] salt, string hostName)
        {
            SHA1 sha1 = new SHA1();

            if (salt.Length != sha1.DigestLength)
            {
                throw new ArgumentException("Salt has wrong length (" + salt.Length + ")");
            }

            HMAC hmac = new HMAC(sha1, salt, salt.Length);

            hmac.Update(StringEncoder.GetBytes(hostName));

            byte[] dig = new byte[hmac.DigestLength];

            hmac.Digest(dig);

            return dig;
        }

        /// <summary>
        /// Check an entry and make sure that the hash is correct. 
        /// </summary>
        /// <param name="entry">The known hosts entry.</param>
        /// <param name="hostName">The host name.</param>
        /// <returns>True if the hash matches what we calculate.</returns>
        private static bool CheckHashed(string entry, string hostName)
        {
            if (entry.StartsWith("|1|", StringComparison.Ordinal) == false)
            {
                return false;
            }

            int delim_idx = entry.IndexOf('|', 3);

            if (delim_idx == -1)
            {
                return false;
            }

            string salt_base64 = entry.Substring(3, delim_idx);
            string hash_base64 = entry.Substring(delim_idx + 1);

            byte[] salt = null;
            byte[] hash = null;

            try
            {
                salt = Convert.FromBase64String(salt_base64);
                hash = Convert.FromBase64String(hash_base64);
            }
            catch (IOException)
            {
                return false;
            }

            SHA1 sha1 = new SHA1();

            if (salt.Length != sha1.DigestLength)
            {
                return false;
            }

            byte[] dig = HmacSha1Hash(salt, hostName);

            for (int i = 0; i < dig.Length; i++)
            {
                if (dig[i] != hash[i])
                {
                    return false;
                }
            }

            return true;
        }

        /// <summary>
        /// Function to compare two keys.
        /// </summary>
        /// <param name="key1">First key to compare.</param>
        /// <param name="key2">Second key to compare.</param>
        /// <returns>True if the two keys are the same.</returns>
        [SuppressMessage("Microsoft.Performance", "CA1800:DoNotCastUnnecessarily",
                         Justification = "FxCop is confused.  The casting is appropriate.")]
        private static bool MatchKeys(object key1, object key2)
        {
            if ((key1 is RSAPublicKey) && (key2 is RSAPublicKey))
            {
                RSAPublicKey savedRSAKey = (RSAPublicKey)key1;
                RSAPublicKey remoteRSAKey = (RSAPublicKey)key2;

                if (savedRSAKey.E.Equals(remoteRSAKey.E) == false)
                {
                    return false;
                }

                if (savedRSAKey.N.Equals(remoteRSAKey.N) == false)
                {
                    return false;
                }

                return true;
            }

            if ((key1 is DSAPublicKey) && (key2 is DSAPublicKey))
            {
                DSAPublicKey savedDSAKey = (DSAPublicKey)key1;
                DSAPublicKey remoteDSAKey = (DSAPublicKey)key2;

                if (savedDSAKey.G.Equals(remoteDSAKey.G) == false)
                {
                    return false;
                }

                if (savedDSAKey.P.Equals(remoteDSAKey.P) == false)
                {
                    return false;
                }

                if (savedDSAKey.Q.Equals(remoteDSAKey.Q) == false)
                {
                    return false;
                }

                if (savedDSAKey.Y.Equals(remoteDSAKey.Y) == false)
                {
                    return false;
                }

                return true;
            }

            return false;
        }

        /// <summary>
        /// Checks the remote host key to see if it is the same or altered.
        /// </summary>
        /// <param name="remoteHostname">The remote host name.</param>
        /// <param name="remoteKey">The remote key.</param>
        /// <returns>
        /// An integer set to one of three constants:
        /// <list type="bullet">
        /// <item><c>HOSTKEY_IS_OK</c>           0</item>
        /// <item><c>HOSTKEY_IS_NEW</c>          1</item>
        /// <item><c>HOSTKEY_HAS_CHANGED</c>     2</item>
        /// </list>
        /// Maybe this should be an enumeration.
        /// </returns>
        private int CheckKey(string remoteHostname, object remoteKey)
        {
            int result = HOSTKEY_IS_NEW;

            lock (this.publicKeys)
            {
                foreach (KnownHostsEntry ke in this.publicKeys)
                {
                    if (this.HostnameMatches(ke.Patterns, remoteHostname) == false)
                    {
                        continue;
                    }

                    bool res = KnownHosts.MatchKeys(ke.Key, remoteKey);

                    if (res == true)
                    {
                        return HOSTKEY_IS_OK;
                    }

                    result = HOSTKEY_HAS_CHANGED;
                }
            }
            
            return result;
        }

        /// <summary>
        /// Returns all the keys for a particular host name.
        /// </summary>
        /// <param name="hostName">The host name to look up.</param>
        /// <returns>The keys for the host name.</returns>
        private ArrayList GetAllKeys(string hostName)
        {
            ArrayList keys = new ArrayList();

            lock (this.publicKeys)
            {
                foreach (KnownHostsEntry ke in this.publicKeys)
                {
                    if (this.HostnameMatches(ke.Patterns, hostName) == false)
                    {
                        continue;
                    }

                    keys.Add(ke.Key);
                }
            }

            return keys;
        }

        /// <summary>
        /// See if a host name matches a set of host name 'patterns'.
        /// </summary>
        /// <param name="hostpatterns">The host name patterns to check.</param>
        /// <param name="hostName">The host name.</param>
        /// <returns>True if the host name matches one of the patterns.</returns>
        [SuppressMessage("Microsoft.Globalization", "CA1304:SpecifyCultureInfo", MessageId = "System.String.ToLower",
                        Justification = "Should not be an issue for host names.")]
        private bool HostnameMatches(string[] hostpatterns, string hostName)
        {
            bool isMatch = false;
            bool negate = false;

            hostName = hostName.ToLower();

            for (int k = 0; k < hostpatterns.Length; k++)
            {
                if (hostpatterns[k] == null)
                {
                    continue;
                }

                string pattern = null;

                // In contrast to OpenSSH we also allow negated hash entries (as well as hashed
                // entries in lines with multiple entries).
                if ((hostpatterns[k].Length > 0) && (hostpatterns[k][0] == '!'))
                {
                    pattern = hostpatterns[k].Substring(1);
                    negate = true;
                }
                else
                {
                    pattern = hostpatterns[k];
                    negate = false;
                }

                // Optimize, no need to check this entry
                if (isMatch && (negate == false))
                {
                    continue;
                }

                // Now compare
                if (pattern[0] == '|')
                {
                    if (KnownHosts.CheckHashed(pattern, hostName))
                    {
                        if (negate)
                        {
                            return false;
                        }
                        
                        isMatch = true;
                    }
                }
                else
                {
                    pattern = pattern.ToLower();

                    if ((pattern.IndexOf('?') != -1) || (pattern.IndexOf('*') != -1))
                    {
                        if (this.PseudoRegex(pattern.ToCharArray(), 0, hostName.ToCharArray(), 0))
                        {
                            if (negate)
                            {
                                return false;
                            }
                            
                            isMatch = true;
                        }
                    }
                    else if (string.Compare(pattern, hostName, StringComparison.Ordinal) == 0)
                    {
                        if (negate)
                        {
                            return false;
                        }
                        
                        isMatch = true;
                    }
                }
            }

            return isMatch;
        }

        /// <summary>
        /// Initializes a character array to hold known hosts.
        /// </summary>
        /// <param name="knownHostsData">The initial data to parse into the character array.</param>
        private void Initialize(char[] knownHostsData)
        {
            string khd = new string(knownHostsData);
            StringReader br = new StringReader(khd);

            while (true)
            {
                string line = br.ReadLine();

                if (line == null)
                {
                    break;
                }

                line = line.Trim();

                if (line.StartsWith("#", StringComparison.Ordinal))
                {
                    continue;
                }

                string[] arr = line.Split(' ');

                if (arr.Length >= 3)
                {
                    if ((string.Compare(arr[1], "ssh-rsa", StringComparison.Ordinal) == 0)
                        || (string.Compare(arr[1], "ssh-dss", StringComparison.Ordinal) == 0))
                    {
                        string[] hostNames = arr[0].Split(',');

                        byte[] msg = Convert.FromBase64String(arr[2]);

                        this.AddHostkey(hostNames, arr[1], msg);
                    }
                }
            }
        }

        /// <summary>
        /// Initializes a file (FileStream) to hold known hosts.
        /// </summary>
        /// <param name="knownHosts">The open FileStream.</param>
        private void Initialize(FileStream knownHosts)
        {
            byte[] buff = new byte[512];

            StringWriter cw = new StringWriter(CultureInfo.InvariantCulture);
            System.Text.ASCIIEncoding enc = new System.Text.ASCIIEncoding();

            while (true)
            {
                int len = knownHosts.Read(buff, 0, 512);
                if (len < 0)
                {
                    break;
                }
                
                cw.Write(enc.GetString(buff), 0, len);
            }

            knownHosts.Close();

            this.Initialize(cw.ToString().ToCharArray());
        }

        /// <summary>
        /// Apparently, a regex-alike function.
        /// </summary>
        /// <param name="pattern">Pattern to match against.</param>
        /// <param name="i">Index for the pattern.</param>
        /// <param name="match">The character array to match (or not).</param>
        /// <param name="j">Index for the character array to match.</param>
        /// <returns>True if there is a match.</returns>
        private bool PseudoRegex(char[] pattern, int i, char[] match, int j)
        {
            // This matching logic is equivalent to the one present in OpenSSH 4.1
            while (true)
            {
                // Are we at the end of the pattern?
                if (pattern.Length == i)
                {
                    return match.Length == j;
                }

                if (pattern[i] == '*')
                {
                    i++;

                    if (pattern.Length == i)
                    {
                        return true;
                    }

                    if ((pattern[i] != '*') && (pattern[i] != '?'))
                    {
                        while (true)
                        {
                            if ((pattern[i] == match[j]) && this.PseudoRegex(pattern, i + 1, match, j + 1))
                            {
                                return true;
                            }
                            
                            j++;
                            if (match.Length == j)
                            {
                                return false;
                            }
                        }
                    }

                    while (true)
                    {
                        if (this.PseudoRegex(pattern, i, match, j))
                        {
                            return true;
                        }
                        
                        j++;
                        if (match.Length == j)
                        {
                            return false;
                        }
                    }
                }

                if (match.Length == j)
                {
                    return false;
                }

                if ((pattern[i] != '?') && (pattern[i] != match[j]))
                {
                    return false;
                }

                i++;
                j++;
            }
        }

        /// <summary>
        /// Returns the recommended host key algorithms based on the key for the host.
        /// </summary>
        /// <param name="hostName">The host name that we are getting the key algorithms for.</param>
        /// <returns>A string array with the recommended host key algorithms.</returns>
        private string[] RecommendHostkeyAlgorithms(string hostName)
        {
            string preferredAlgo = null;

            ArrayList keys = this.GetAllKeys(hostName);

            for (int i = 0; i < keys.Count; i++)
            {
                string thisAlgo = null;

                if (keys[i] is RSAPublicKey)
                {
                    thisAlgo = "ssh-rsa";
                }
                else if (keys[i] is DSAPublicKey)
                {
                    thisAlgo = "ssh-dss";
                }
                else
                {
                    continue;
                }

                if (preferredAlgo != null)
                {
                    // If we find different key types, then return null
                    if (string.Compare(preferredAlgo, thisAlgo, StringComparison.Ordinal) != 0)
                    {
                        return null;
                    }

                    // OK, we found the same algo again, optimize
                    continue;
                }
            }

            // If we did not find anything that we know of, return null
            if (preferredAlgo == null)
            {
                return null;
            }

            // Now put the preferred algo to the start of the array.
            // You may ask yourself why we do it that way - basically, we could just
            // return only the preferred algorithm: since we have a saved key of that
            // type (sent earlier from the remote host), then that should work out.
            // However, imagine that the server is (for whatever reasons) not offering
            // that type of hostkey anymore (e.g., "ssh-rsa" was disabled and
            // now "ssh-dss" is being used). If we then do not let the server send us
            // a fresh key of the new type, then we shoot ourself in the foot:
            // the connection cannot be established and hence the user cannot decide
            // if he/she wants to accept the new key.
            if (preferredAlgo.Equals("ssh-rsa"))
            {
                return new string[] { "ssh-rsa", "ssh-dss" };
            }

            return new string[] { "ssh-dss", "ssh-rsa" };
        }

        /// <summary>
        /// A class to contain the data for a known host.
        /// </summary>
        private class KnownHostsEntry
        {
            private string[] patterns;
            private object key;
            
            /// <summary>
            /// Initializes a new instance of the KnownHostsEntry class.
            /// </summary>
            /// <param name="patterns">A string array of the patterns that match this host.</param>
            /// <param name="key">The host key.</param>
            public KnownHostsEntry(string[] patterns, object key)
            {
                this.patterns = patterns;
                this.key = key;
            }
            
            /// <summary>
            /// Gets the string array of patterns for this host entry.
            /// </summary>
            public string[] Patterns
            {
                get { return this.patterns; }
            }
            
            /// <summary>
            /// Gets the host key for this entry.
            /// </summary>
            public object Key
            {
                get { return this.key; }
            }
        }
    }
}
