// <copyright file="LocalStreamForwarder.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> LocalStreamForwarder
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;

using Eunomia.SSH2.Channel;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>LocalStreamForwarder</c> forwards an Input and OutputStream
    /// pair via the secure tunnel to another host (which may or may not be identical
    /// to the remote SSH-2 server).
    /// </summary>
    public class LocalStreamForwarder
    {
        private ChannelManager cm;
        private Eunomia.SSH2.Channel.SSHChannel cn;

        /// <summary>
        /// Initializes a new instance of the LocalStreamForwarder class.
        /// </summary>
        /// <param name="cm">The ChannelManager that manages this forward.</param>
        /// <param name="hostToConnect">The host to forward to.</param>
        /// <param name="portToConnect">The port to forward to.</param>
        internal LocalStreamForwarder(ChannelManager cm, string hostToConnect, int portToConnect)
        {
            this.cm = cm;

            this.cn = cm.OpenDirectTCPIPChannel(hostToConnect, portToConnect, "127.0.0.1", 0);
        }

        /// <summary>
        /// Gets the underlying InputStream for this LocalStreamForwarder.
        /// </summary>
        /// <exception cref="IOException"></exception>
        public Stream InputStream
        {
            get { return this.cn.StandardOutputStream; }
        }
        
        /// <summary>
        /// Gets the OutputStream. Please be aware that the implementation MAY use an
        /// internal buffer. To make sure that the buffered data is sent over the
        /// tunnel, you have to call the <c>Flush</c> method of the
        /// <c>OutputStream</c>. To signal EOF, please use the
        /// <c>Close</c> method of the <c>OutputStream</c>.
        /// </summary>
        /// <returns>An <c>OutputStream</c> object.</returns>
        /// <exception cref="IOException"></exception>
        public Stream OutputStream
        {
            get { return this.cn.StandardInputStream; }
        }

        /// <summary>
        /// Close the underlying SSH forwarding channel and free up resources.
        /// You can also use this method to force the shutdown of the underlying
        /// forwarding channel. Pending output (OutputStream not flushed) will NOT
        /// be sent. Pending input (InputStream) can still be read. If the shutdown
        /// operation is already in progress (initiated from either side), then this
        /// call is a no-op.
        /// </summary>
        /// <exception cref="IOException"></exception>
        public void Close()
        {
            this.cm.CloseChannel(this.cn, "Closed due to user request.", true);
        }
    }
}
