﻿// <copyright file="PacketChannelOpenConfirmation.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketChannelOpenConfirmation
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketChannelOpenConfirmation packet.
    /// </summary>
    public class PacketChannelOpenConfirmation
    {
        private byte[] payload;

        private int recipientChannelID;
        private int senderChannelID;
        private int initialWindowSize;
        private int maxPacketSize;
        
        /// <summary>
        /// Initializes a new instance of the PacketChannelOpenConfirmation class.
        /// </summary>
        /// <param name="recipientChannelId">The receipient channel id.</param>
        /// <param name="senderChannelId">The sender channel id.</param>
        /// <param name="initialWindowSize">The initial window size.</param>
        /// <param name="maxPacketSize">The maximum packet size.</param>
        public PacketChannelOpenConfirmation(
            int recipientChannelId,
            int senderChannelId,
            int initialWindowSize,
            int maxPacketSize)
        {
            this.recipientChannelID = recipientChannelId;
            this.senderChannelID = senderChannelId;
            this.initialWindowSize = initialWindowSize;
            this.maxPacketSize = maxPacketSize;
        }

        /// <summary>
        /// Initializes a new instance of the PacketChannelOpenConfirmation class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketChannelOpenConfirmation(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageChannelOpenConfirmation)
            {
                throw new IOException(
                    "This is not a SSH_MSG_CHANNEL_OPEN_CONFIRMATION! ("
                    + packet_type + ")");
            }

            this.recipientChannelID = tr.ReadUINT32();
            this.senderChannelID = tr.ReadUINT32();
            this.initialWindowSize = tr.ReadUINT32();
            this.maxPacketSize = tr.ReadUINT32();
            
            if (tr.Remain != 0)
            {
                throw new IOException("Padding in SSH_MSG_CHANNEL_OPEN_CONFIRMATION packet!");
            }
        }

        /// <summary>Gets or sets the recipient channel id.</summary>
        public int RecipientChannelId
        {
            get { return this.recipientChannelID; }
            set { this.recipientChannelID = value; }
        }

        /// <summary>Gets or sets the sender channel id.</summary>
        public int SenderChannelId
        {
            get { return this.senderChannelID; }
            set { this.senderChannelID = value; }
        }
        
        /// <summary>Gets or sets the initial window size.</summary>
        public int InitialWindowSize
        {
            get { return this.initialWindowSize; }
            set { this.initialWindowSize = value; }
        }
        
        /// <summary>Gets or sets the maximum packet size.</summary>
        public int MaxPacketSize
        {
            get { return this.maxPacketSize; }
            set { this.maxPacketSize = value; }
        }
        
        /// <summary>
        /// Gets the constructed payload for the packet.
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageChannelOpenConfirmation);
                    tw.WriteUINT32((uint)this.recipientChannelID);
                    tw.WriteUINT32((uint)this.senderChannelID);
                    tw.WriteUINT32((uint)this.initialWindowSize);
                    tw.WriteUINT32((uint)this.maxPacketSize);
                    this.payload = tw.GetBytes();
                }
                
                return this.payload;
            }
        }
    }
}
