﻿// <copyright file="PacketChannelOpenFailure.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketChannelOpenFailure
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketChannelOpenFailure packet.
    /// </summary>
    public class PacketChannelOpenFailure
    {
        private byte[] payload;

        private int recipientChannelID;
        private int reasonCode;
        private string description;
        private string languageTag;
        
        /// <summary>
        /// Initializes a new instance of the PacketChannelOpenFailure class.
        /// </summary>
        /// <param name="recipientChannelId">The receipient channel id.</param>
        /// <param name="reasonCode">The failure reason code.</param>
        /// <param name="description">The failure description.</param>
        /// <param name="languageTag">The language tag.</param>
        public PacketChannelOpenFailure(
            int recipientChannelId,
            int reasonCode,
            string description,
            string languageTag)
        {
            this.recipientChannelID = recipientChannelId;
            this.reasonCode = reasonCode;
            this.description = description;
            this.languageTag = languageTag;
        }

        /// <summary>
        /// Initializes a new instance of the PacketChannelOpenFailure class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketChannelOpenFailure(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageChannelOpenFailure)
            {
                throw new IOException(
                    "This is not a SSH_MSG_CHANNEL_OPEN_FAILURE! ("
                    + packet_type + ")");
            }
            
            this.recipientChannelID = tr.ReadUINT32();
            this.reasonCode = tr.ReadUINT32();
            this.description = tr.ReadString();
            this.languageTag = tr.ReadString();
            
            if (tr.Remain != 0)
            {
                throw new IOException("Padding in SSH_MSG_CHANNEL_OPEN_FAILURE packet!");
            }
        }
        
        /// <summary>Gets or sets the recipient channel id.</summary>
        public int RecipientChannelId
        {
            get { return this.recipientChannelID; }
            set { this.recipientChannelID = value; }
        }

        /// <summary>Gets or sets the failure reasong code.</summary>
        public int ReasonCode
        {
            get { return this.reasonCode; }
            set { this.reasonCode = value; }
        }
        
        /// <summary>Gets or sets the failure description.</summary>
        public string Description
        {
            get { return this.description; }
            set { this.description = value; }
        }
        
        /// <summary>Gets or sets the language tag.</summary>
        public string LanguageTag
        {
            get { return this.languageTag; }
            set { this.languageTag = value; }
        }

        /// <summary>
        /// Gets the constructed payload for the packet.
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageChannelOpenFailure);
                    tw.WriteUINT32((uint)this.recipientChannelID);
                    tw.WriteUINT32((uint)this.reasonCode);
                    tw.WriteString(this.description);
                    tw.WriteString(this.languageTag);
                    this.payload = tw.GetBytes();
                }
                
                return this.payload;
            }
        }
    }
}
