// <copyright file="PacketKexInit.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketKexInit
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Collections.ObjectModel;
using System.IO;
using Eunomia.SSH2.Crypto;
using Eunomia.SSH2.Transport;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketKexInit packet.
    /// </summary>
    public class PacketKexInit
    {
        private byte[] payload;

        private KexParameters kp = new KexParameters();

        /// <summary>
        /// Initializes a new instance of the PacketKexInit class.
        /// </summary>
        /// <param name="cryptoWishList">A CryptoWishList of what we support.</param>
        /// <param name="random">An instance of the SecureRandom class.</param>
        public PacketKexInit(CryptoWishList cryptoWishList, SecureRandom random)
        {
            this.kp.Cookie = new byte[16];
            random.NextBytes(this.kp.Cookie);

            this.kp.KexAlgorithms = cryptoWishList.KexAlgorithms;
            this.kp.ServerHostkeyAlgorithms = cryptoWishList.ServerHostkeyAlgorithms;
            this.kp.EncryptionAlgorithmsClientToServer = cryptoWishList.C2SEncAlgos;
            this.kp.EncryptionAlgorithmsServerToClient = cryptoWishList.S2CEncAlgos;
            this.kp.MacAlgorithmsClientToServer = cryptoWishList.C2SMacAlgos;
            this.kp.MacAlgorithmsServerToClient = cryptoWishList.S2CMacAlgos;
            this.kp.CompressionAlgorithmsClientToServer = new string[] { "none" };
            this.kp.CompressionAlgorithmsServerToClient = new string[] { "none" };
            this.kp.LanguagesClientToServer = new string[] { };
            this.kp.LanguagesServerToClient = new string[] { };
            this.kp.FirstKexPacketFollows = false;
            this.kp.ReservedField1 = 0;
        }

        /// <summary>
        /// Initializes a new instance of the PacketKexInit class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketKexInit(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageKEXInitialize)
            {
                throw new IOException("This is not a KexInitPacket! (" + packet_type + ")");
            }

            this.kp.Cookie = tr.ReadBytes(16);
            this.kp.KexAlgorithms = tr.ReadNameList();
            this.kp.ServerHostkeyAlgorithms = tr.ReadNameList();
            this.kp.EncryptionAlgorithmsClientToServer = tr.ReadNameList();
            this.kp.EncryptionAlgorithmsServerToClient = tr.ReadNameList();
            this.kp.MacAlgorithmsClientToServer = tr.ReadNameList();
            this.kp.MacAlgorithmsServerToClient = tr.ReadNameList();
            this.kp.CompressionAlgorithmsClientToServer = tr.ReadNameList();
            this.kp.CompressionAlgorithmsServerToClient = tr.ReadNameList();
            this.kp.LanguagesClientToServer = tr.ReadNameList();
            this.kp.LanguagesServerToClient = tr.ReadNameList();
            this.kp.FirstKexPacketFollows = tr.ReadBoolean();
            this.kp.ReservedField1 = tr.ReadUINT32();

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in KexInitPacket!");
            }
        }

        /// <summary>
        /// Gets the payload of the packet (creates if required).
        /// </summary>
        public byte[] Payload
        {
            get
            {
                if (this.payload == null)
                {
                    TypesWriter tw = new TypesWriter();
                    tw.WriteByte(Packet.SSHMessageKEXInitialize);
                    tw.WriteBytes(this.kp.Cookie, 0, 16);
                    tw.WriteNameList(this.kp.KexAlgorithms);
                    tw.WriteNameList(this.kp.ServerHostkeyAlgorithms);
                    tw.WriteNameList(this.kp.EncryptionAlgorithmsClientToServer);
                    tw.WriteNameList(this.kp.EncryptionAlgorithmsServerToClient);
                    tw.WriteNameList(this.kp.MacAlgorithmsClientToServer);
                    tw.WriteNameList(this.kp.MacAlgorithmsServerToClient);
                    tw.WriteNameList(this.kp.CompressionAlgorithmsClientToServer);
                    tw.WriteNameList(this.kp.CompressionAlgorithmsServerToClient);
                    tw.WriteNameList(this.kp.LanguagesClientToServer);
                    tw.WriteNameList(this.kp.LanguagesServerToClient);
                    tw.WriteBoolean(this.kp.FirstKexPacketFollows);
                    tw.WriteUINT32((uint)this.kp.ReservedField1);
                    this.payload = tw.GetBytes();
                }

                return this.payload;
            }
        }

        /// <summary>
        /// Gets the KexParameters.
        /// </summary>
        public KexParameters KexParameters
        {
            get { return this.kp; }
        }

        /// <summary>
        /// Gets the compression algorithms from client to server.
        /// </summary>
        public string[] CompressionAlgorithmsClientToServer
        {
            get { return this.kp.CompressionAlgorithmsClientToServer; }
        }

        /// <summary>
        /// Gets the compression algorithms from server to client.
        /// </summary>
        public string[] CompressionAlgorithmsServerToClient
        {
            get { return this.kp.CompressionAlgorithmsServerToClient; }
        }

        /// <summary>
        /// Gets the associated cookie.
        /// </summary>
        public byte[] Cookie
        {
            get { return this.kp.Cookie; }
        }

        /// <summary>
        /// Gets the encryption algorithms from client to server.
        /// </summary>
        public string[] EncryptionAlgorithmsClientToServer
        {
            get { return this.kp.EncryptionAlgorithmsClientToServer; }
        }

        /// <summary>
        /// Gets the encryption algorithms from server to client.
        /// </summary>
        public string[] EncryptionAlgorithmsServerToClient
        {
            get { return this.kp.EncryptionAlgorithmsServerToClient; }
        }

        /// <summary>
        /// Gets a value indicating whether the first kex packet will follow.
        /// </summary>
        public bool FirstKexPacketFollows
        {
            get { return this.kp.FirstKexPacketFollows; }
        }

        /// <summary>
        /// Gets a string array of all the kex algorithms.
        /// </summary>
        public string[] KexAlgorithms
        {
            get { return this.kp.KexAlgorithms; }
        }

        /// <summary>
        /// Gets a string array of all the available client to server languages.
        /// </summary>
        public string[] LanguagesClientToServer
        {
            get { return this.kp.LanguagesClientToServer; }
        }

        /// <summary>
        /// Gets a string array of all the available server to client languages.
        /// </summary>
        public string[] LanguagesServerToClient
        {
            get { return this.kp.LanguagesServerToClient; }
        }

        /// <summary>
        /// Gets a string array of all the available client to server MAC algorithms.
        /// </summary>
        public string[] MacAlgorithmsClientToServer
        {
            get { return this.kp.MacAlgorithmsClientToServer; }
        }

        /// <summary>
        /// Gets a string array of all the available server to client MAC algorithms.
        /// </summary>
        public string[] MacAlgorithmsServerToClient
        {
            get { return this.kp.MacAlgorithmsServerToClient; }
        }

        /// <summary>
        /// Gets the reserved field 1.
        /// </summary>
        public int ReservedField1
        {
            get { return this.kp.ReservedField1; }
        }

        /// <summary>
        /// Gets the server host key algorithms.
        /// </summary>
        public string[] ServerHostkeyAlgorithms
        {
            get { return this.kp.ServerHostkeyAlgorithms; }
        }
    }
}
