// <copyright file="PacketUserauthInfoRequest.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> PacketUserauthInfoRequest
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// Handles a PacketUserauthInfoRequest packet.
    /// </summary>
    public class PacketUserauthInfoRequest
    {
        private byte[] payload;

        private string name;
        private string instruction;
        private string languageTag;
        private int numPrompts;

        private string[] prompt;
        private bool[] echo;

        /// <summary>
        /// Initializes a new instance of the PacketUserauthInfoRequest class from an existing payload.
        /// </summary>
        /// <param name="payload">The byte array with the packet in it.</param>
        /// <param name="off">The offset to the beginning of the packet.</param>
        /// <param name="len">The length of the packet data.</param>
        public PacketUserauthInfoRequest(byte[] payload, int off, int len)
        {
            this.payload = new byte[len];
            Array.Copy(payload, off, this.payload, 0, len);

            TypesReader tr = new TypesReader(payload, off, len);

            int packet_type = tr.ReadByte();

            if (packet_type != Packet.SSHMessageUserAuthorizationInfoRequest)
            {
                throw new IOException("This is not a SSH_MSG_USERAUTH_INFO_REQUEST! (" + packet_type + ")");
            }

            this.name = tr.ReadString();
            this.instruction = tr.ReadString();
            this.languageTag = tr.ReadString();

            this.numPrompts = tr.ReadUINT32();

            this.prompt = new string[this.numPrompts];
            this.echo = new bool[this.numPrompts];

            for (int i = 0; i < this.numPrompts; i++)
            {
                this.prompt[i] = tr.ReadString();
                this.echo[i] = tr.ReadBoolean();
            }

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in SSH_MSG_USERAUTH_INFO_REQUEST packet!");
            }
        }
        
        /// <summary>
        /// Gets a value (in an array) indicating whether to echo characters or not.
        /// There is an Echo for each Prompt.
        /// </summary>
        public bool[] Echo
        {
            get { return this.echo; }
        }
        
        /// <summary>
        /// Gets the instruction to be displayed.
        /// </summary>
        public string Instruction
        {
            get { return this.instruction; }
        }

        /// <summary>
        /// Gets the language for this exchange (normally blank).
        /// </summary>
        public string LanguageTag
        {
            get { return this.languageTag; }
        }

        /// <summary>
        /// Gets the name to be displayed (a user name).
        /// </summary>
        public string Name
        {
            get { return this.name; }
        }

        /// <summary>
        /// Gets the number of prompts (may be greater than 1).
        /// Each prompt has a Prompt, and an Echo.
        /// </summary>
        public int NumberOfPrompts
        {
            get { return this.numPrompts; }
        }

        /// <summary>
        /// Gets the array of prompts...must not be blank.
        /// </summary>
        public string[] Prompt
        {
            get { return this.prompt; }
        }
    }
}
