﻿// <copyright file="Packets.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Packets
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;

namespace Eunomia.SSH2.Packets
{
    #pragma warning disable 1591
    /// <summary>
    /// This class contains a bunch of packet related constants as well as a
    /// method to translate from the error number to the error message.
    /// </summary>
    public static class Packet
    {
        public const int SSHMessageDisconnect = 1;
        public const int SSHMessageIgnore = 2;
        public const int SSHMessageUnimplemented = 3;
        public const int SSHMessageDebug = 4;
        public const int SSHMessageServiceRequest = 5;
        public const int SSHMessageServiceAccept = 6;

        public const int SSHMessageKEXInitialize = 20;
        public const int SSHMessageNewKeys = 21;

        public const int SSHMessageKEXDHInitialize = 30;
        public const int SSHMessageKEXDHReply = 31;

        public const int SSHMessageKEXDHGEXRequestOld = 30;
        public const int SSHMessageKEXDHGEXRequest = 34;
        public const int SSHMessageKEXDHGEXGroup = 31;
        public const int SSHMessageKEXDHGEXInitialize = 32;
        public const int SSHMessageKEXDHGEXReply = 33;

        public const int SSHMessageUserAuthorizationRequest = 50;
        public const int SSHMessageUserAuthorizationFailure = 51;
        public const int SSHMessageUserAuthorizationSuccess = 52;
        public const int SSHMessageUserAuthorizationBanner = 53;
        public const int SSHMessageUserAuthorizationInfoRequest = 60;
        public const int SSHMessageUserAuthorizationInfoReply = 61;

        public const int SSHMessageGlobalRequest = 80;
        public const int SSHMessageRequestSuccess = 81;
        public const int SSHMessageRequestFailure = 82;

        public const int SSHMessageChannelOpen = 90;
        public const int SSHMessageChannelOpenConfirmation = 91;
        public const int SSHMessageChannelOpenFailure = 92;
        public const int SSHMessageChannelWindowAdjust = 93;
        public const int SSHMessageChannelData = 94;
        public const int SSHMessageChannelExtendedData = 95;
        public const int SSHMessageChannelEOF = 96;
        public const int SSHMessageChannelClose = 97;
        public const int SSHMessageChannelRequest = 98;
        public const int SSHMessageChannelSuccess = 99;
        public const int SSHMessageChannelFailure = 100;

        public const int SSHExtendedDataStandardError = 1;

        public const int SSHDisconnectHostNotAllowedToConnect = 1;
        public const int SSHDisconnectProtocolError = 2;
        public const int SSHDisconnectKeyExchangeFailed = 3;
        public const int SSHDisconnectReserved = 4;
        public const int SSHDisconnectMACError = 5;
        public const int SSHDisconnectCompressionError = 6;
        public const int SSHDisconnectServiceNotAvailable = 7;
        public const int SSHDisconnectProtocolVersionNotSupported = 8;
        public const int SSHDisconnectHostkeyNotVerifiable = 9;
        public const int SSHDisconnectConnectionLost = 10;
        public const int SSHDisconnectByApplication = 11;
        public const int SSHDisconnectToManyConnections = 12;
        public const int SSHDisconnectAuthorizationCanceledByUser = 13;
        public const int SSHDisconnectNoMoreAuthorizationMethodsAvailable = 14;
        public const int SSHDisconnectIllegalUserName = 15;

        public const int SSHOpenAdministrativelyProhibited = 1;
        public const int SSHOpenConnectFailed = 2;
        public const int SSHOpenUnknownChannelType = 3;
        public const int SSHOpenResourceShortage = 4;

        private static string[] reverseNames = new string[101]
        {
            string.Empty,

            // 1
            "SSH_MSG_DISCONNECT",
            "SSH_MSG_IGNORE",
            "SSH_MSG_UNIMPLEMENTED",
            "SSH_MSG_DEBUG",
            "SSH_MSG_SERVICE_REQUEST",
            "SSH_MSG_SERVICE_ACCEPT",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty,

            // 20
            "SSH_MSG_KEXINIT",
            "SSH_MSG_NEWKEYS",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty,

            // 30
            "SSH_MSG_KEXDH_INIT",
            "SSH_MSG_KEXDH_REPLY/SSH_MSG_KEX_DH_GEX_GROUP",
            "SSH_MSG_KEX_DH_GEX_INIT",
            "SSH_MSG_KEX_DH_GEX_REPLY",
            "SSH_MSG_KEX_DH_GEX_REQUEST",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,

            // 50
            "SSH_MSG_USER_AUTH_REQUEST",
            "SSH_MSG_USER_AUTH_FAILURE",
            "SSH_MSG_USER_AUTH_SUCCESS",
            "SSH_MSG_USER_AUTH_BANNER",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,

            // 60
            "SSH_MSG_USER_AUTH_INFO_REQUEST",
            "SSH_MSG_USER_AUTH_INFO_RESPONSE",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,

            // 80
            "SSH_MSG_GLOBAL_REQUEST",
            "SSH_MSG_REQUEST_SUCCESS",
            "SSH_MSG_REQUEST_FAILURE",
            string.Empty, string.Empty, string.Empty, string.Empty, string.Empty, string.Empty,
            string.Empty,

            // 90
            "SSH_MSG_CHANNEL_OPEN",
            "SSH_MSG_CHANNEL_OPEN_CONFIRMATION",
            "SSH_MSG_CHANNEL_OPEN_FAILURE",
            "SSH_MSG_CHANNEL_WINDOW_ADJUST",
            "SSH_MSG_CHANNEL_DATA",
            "SSH_MSG_CHANNEL_EXTENDED_DATA",
            "SSH_MSG_CHANNEL_EOF",
            "SSH_MSG_CHANNEL_CLOSE",
            "SSH_MSG_CHANNEL_REQUEST",
            "SSH_MSG_CHANNEL_SUCCESS",
            "SSH_MSG_CHANNEL_FAILURE",
        };

        /// <summary>
        /// Method to translate from the error number to the error message.
        /// </summary>
        /// <param name="errorNumber">The error number.</param>
        /// <returns>The error message.</returns>
        [SuppressMessage("Microsoft.Performance", "CA1820:TestForEmptyStringsUsingStringLength",
                        Justification = "Speed is not an issue here.")]
        public static string GetMessageName(int errorNumber)
        {
            string res = null;

            if ((errorNumber >= 0) && (errorNumber < reverseNames.Length))
            {
                res = reverseNames[errorNumber];
            }

            return (res == string.Empty) ? ("UNKNOWN MSG " + errorNumber) : res;
        }
    #pragma warning restore 1591
    }
}
