// <copyright file="TypesReader.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> TypesReader
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;
using System.Text;

using Eunomia.SSH2.Util;
using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// TypesReader takes data received in a packet and parses it into various data elements.
    /// </summary>
    public class TypesReader
    {
        private byte[] arr;
        private int pos = 0;
        private int max = 0;

        /// <summary>
        /// Initializes a new instance of the TypesReader class.
        /// </summary>
        /// <param name="buffer">The byte array to read.</param>
        public TypesReader(byte[] buffer) : this(buffer, 0, buffer.Length)
        {
        }

        /// <summary>
        /// Initializes a new instance of the TypesReader class.
        /// </summary>
        /// <param name="buffer">The byte array to read.</param>
        /// <param name="offset">The offset where we will begin reading.</param>
        public TypesReader(byte[] buffer, int offset) : this(buffer, offset, buffer.Length - offset)
        {
        }

        /// <summary>
        /// Initializes a new instance of the TypesReader class.
        /// </summary>
        /// <param name="buffer">The byte array to read.</param>
        /// <param name="offset">The offset where we will begin reading.</param>
        /// <param name="length">The length of the data we are reading.</param>
        public TypesReader(byte[] buffer, int offset, int length)
        {
            this.arr = buffer;
            this.pos = offset;
            this.max = offset + length;

            if ((this.pos < 0) || (this.pos > buffer.Length))
            {
                throw new ArgumentException("Illegal offset.");
            }

            if ((this.max < 0) || (this.max > buffer.Length))
            {
                throw new ArgumentException("Illegal length.");
            }
        }

        /// <summary>
        /// Gets the number of bytes yet to process.
        /// </summary>
        public int Remain
        {
            get { return this.max - this.pos; }
        }

        /// <summary>
        /// Reads a byte from the packet at the current position.
        /// </summary>
        /// <returns>The byte at the current position.</returns>
        public int ReadByte()
        {
            if (this.pos >= this.max)
            {
                throw new IOException("Packet too short.");
            }

            return this.arr[this.pos++] & 0xff;
        }

        /// <summary>
        /// Reads a byte array starting at the current position for len length.
        /// </summary>
        /// <param name="len">The length of the byte array to read.</param>
        /// <returns>The byte array that was read.</returns>
        public byte[] ReadBytes(int len)
        {
            if ((this.pos + len) > this.max)
            {
                throw new IOException("Packet too short.");
            }

            byte[] res = new byte[len];

            Array.Copy(this.arr, this.pos, res, 0, len);
            this.pos += len;

            return res;
        }

        /// <summary>
        /// Reads bytes from the current packet and postion.
        /// Places them into an existing destination byte array.
        /// </summary>
        /// <param name="dst">The byte array that we place the bytes into.</param>
        /// <param name="off">The offset into the destination array.</param>
        /// <param name="len">The number of bytes to read.</param>
        public void ReadBytes(byte[] dst, int off, int len)
        {
            if ((this.pos + len) > this.max)
            {
                throw new IOException("Packet too short.");
            }

            Array.Copy(this.arr, this.pos, dst, off, len);
            this.pos += len;
        }

        /// <summary>
        /// Reads a boolean value from the current packet/position.
        /// </summary>
        /// <returns>The boolean that is read.</returns>
        public bool ReadBoolean()
        {
            if (this.pos >= this.max)
            {
                throw new IOException("Packet too short.");
            }

            return this.arr[this.pos++] != 0;
        }

        /// <summary>
        /// Reads a signed 32 bit integer from the current packet/position.
        /// WARNING: Despite the name of the function we return a signed integer.
        /// </summary>
        /// <returns>A signed integer read from the current packet/position.</returns>
        public int ReadUINT32()
        {
            if ((this.pos + 4) > this.max)
            {
                throw new IOException("Packet too short.");
            }
            
            return ((this.arr[this.pos++] & 0xff) << 24) |
                ((this.arr[this.pos++] & 0xff) << 16) |
                ((this.arr[this.pos++] & 0xff) << 8) |
                (this.arr[this.pos++] & 0xff);
        }

        /// <summary>
        /// Reads a signed 64 bit integer from the current packet/position.
        /// WARNING: Despite the name of the function we return a signed long.
        /// </summary>
        /// <returns>A signed long read from the current packet/position.</returns>
        public long ReadUINT64()
        {
            if ((this.pos + 8) > this.max)
            {
                throw new IOException("Packet too short.");
            }

            long high = ((this.arr[this.pos++] & 0xff) << 24) | ((this.arr[this.pos++] & 0xff) << 16)
                | ((this.arr[this.pos++] & 0xff) << 8) | (this.arr[this.pos++] & 0xff); /* sign extension may take place - will be shifted away =) */

            long low = ((this.arr[this.pos++] & 0xff) << 24) | ((this.arr[this.pos++] & 0xff) << 16) |
                ((this.arr[this.pos++] & 0xff) << 8) | (this.arr[this.pos++] & 0xff); /* sign extension may take place - handle below */

            return (high << 32) | (low & 0xffffffffL); /* see Java language spec (15.22.1, 5.6.2) */
        }

        /// <summary>
        /// Reads a BigInteger from the current packet/position.
        /// </summary>
        /// <returns>A BigInteger read from the current packet/position.</returns>
        public BigInteger ReadMPINT()
        {
            BigInteger b;

            byte[] raw = this.ReadByteString();

            if (raw.Length == 0)
            {
                b = BigInteger.Zero;
            }
            else
            {
                b = new BigInteger(raw);
            }

            return b;
        }

        /// <summary>
        /// Reads a byte string from the current packet/position.
        /// </summary>
        /// <returns>The byte array with the 'string' in it.</returns>
        public byte[] ReadByteString()
        {
            int len = this.ReadUINT32();

            if ((len + this.pos) > this.max)
            {
                throw new IOException("Malformed SSH byte string.");
            }

            byte[] res = new byte[len];
            Array.Copy(this.arr, this.pos, res, 0, len);
            this.pos += len;
            return res;
        }

        /// <summary>
        /// Read a string from the current packet/position.
        /// </summary>
        /// <param name="charsetName">The character set that we need to use.</param>
        /// <returns>The straing that we read.</returns>
        public string ReadString(string charsetName)
        {
            int len = this.ReadUINT32();

            if ((len + this.pos) > this.max)
            {
                throw new IOException("Malformed SSH string.");
            }

            string res;
            Encoding enc;
            if (charsetName == null)
            {
                res = StringEncoder.GetString(this.arr, this.pos, len);
            }
            else
            {
                enc = Encoding.GetEncoding(charsetName);
                res = enc.GetString(this.arr, this.pos, len);
            }
            
            this.pos += len;

            return res;
        }

        /// <summary>
        /// Read a string from the current packet/position using the default character set.
        /// </summary>
        /// <returns>The string that we read.</returns>
        public string ReadString()
        {
            int len = this.ReadUINT32();

            if ((len + this.pos) > this.max)
            {
                throw new IOException("Malformed SSH string.");
            }

            string res = StringEncoder.GetString(this.arr, this.pos, len);
            this.pos += len;

            return res;
        }

        /// <summary>
        /// Gets a string array from the current packet/position.
        /// </summary>
        /// <returns>The string array that we read.</returns>
        public string[] ReadNameList()
        {
            return this.ReadString().Split(',');
        }
    }
}
