﻿// <copyright file="TypesWriter.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> TypesWriter
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Text;
using Eunomia.SSH2.Util;
using Org.BouncyCastle.Math;

namespace Eunomia.SSH2.Packets
{
    /// <summary>
    /// TypesWriter takes data to be sent in a packet and adds it to a byte buffer.
    /// </summary>
    public class TypesWriter
    {
        private byte[] arr;
        private int pos;

        /// <summary>
        /// Initializes a new instance of the TypesWriter class.
        /// </summary>
        public TypesWriter()
        {
            this.arr = new byte[256];
            this.pos = 0;
        }

        /// <summary>
        /// Gets the current length of the data in the arr byte buffer.
        /// </summary>
        public int Length
        {
            get
            {
                return this.pos;
            }
        }

        /// <summary>
        /// Copies the bytes in the arr byte buffer into a new destination byte array and returns it.
        /// </summary>
        /// <returns>A copy of the bytes that are in the arr byte buffer.</returns>
        public byte[] GetBytes()
        {
            byte[] dst = new byte[this.pos];
            Array.Copy(this.arr, 0, dst, 0, this.pos);
            return dst;
        }

        /// <summary>
        /// Copies the bytes in the arr byte buffer into the supplied destination byte array.
        /// </summary>
        /// <param name="dst">The byte array to copy the data to.</param>
        public void GetBytes(byte[] dst)
        {
            Array.Copy(this.arr, 0, dst, 0, this.pos);
        }

        /// <summary>
        /// Adds the data for a bool into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The bool to add to the byte buffer.</param>
        public void WriteBoolean(bool value)
        {
            if ((this.pos + 1) > this.arr.Length)
            {
                this.Resize(this.arr.Length + 32);
            }

            this.arr[this.pos++] = value ? (byte)1 : (byte)0;
        }

        /// <summary>
        /// Adds the data for a byte (passed as an int) into the arr byte buffer at a supplied position.
        /// </summary>
        /// <param name="value">The byte (passed as an int) to add to the byte buffer.</param>
        /// <param name="offset">The location in the byte buffer to write to.</param>
        public void WriteByte(int value, int offset)
        {
            if ((offset + 1) > this.arr.Length)
            {
                this.Resize(offset + 32);
            }

            this.arr[offset] = (byte)value;
        }

        /// <summary>
        /// Adds the data for a byte (passed as an int) into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The byte (passed as an int) to add to the byte buffer.</param>
        public void WriteByte(int value)
        {
            this.WriteByte(value, this.pos);
            this.pos++;
        }

        /// <summary>
        /// Adds the data for a BigInteger into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The BigInteger to add to the byte buffer.</param>
        public void WriteMPInt(BigInteger value)
        {
            byte[] raw = value.ToByteArray();

            if ((raw.Length == 1) && (raw[0] == 0))
            {
                this.WriteUINT32(0); /* string with zero bytes of data */
            }
            else
            {
                this.WriteString(raw, 0, raw.Length);
            }
        }

        /// <summary>
        /// Adds the data for byte array into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="buff">The byte array to add to the byte buffer.</param>
        public void WriteBytes(byte[] buff)
        {
            this.WriteBytes(buff, 0, buff.Length);
        }
        
        /// <summary>
        /// Adds the data for byte array into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="buff">The byte array to add to the arr byte buffer.</param>
        /// <param name="off">The offset in the buff byte array that we will write the bytes from.</param>
        /// <param name="len">The length of the bytes that we are adding to the arr byte buffer.</param>
        public void WriteBytes(byte[] buff, int off, int len)
        {
            if ((this.pos + len) > this.arr.Length)
            {
                this.Resize(this.arr.Length + len + 32);
            }

            Array.Copy(buff, off, this.arr, this.pos, len);
            this.pos += len;
        }

        /// <summary>
        /// Adds the data for a string (passed as a byte array) into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="buff">The byte array (containing a string) to be added to the arr byte buffer.</param>
        /// <param name="off">The offset in the buff byte array that we will write the bytes from.</param>
        /// <param name="len">The length of the bytes that we are adding to the arr byte buffer.</param>
        public void WriteString(byte[] buff, int off, int len)
        {
            this.WriteUINT32((uint)len);
            this.WriteBytes(buff, off, len);
        }

        /// <summary>
        /// Adds the data for a string into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The string to be added to the arr byte buffer.</param>
        public void WriteString(string value)
        {
            byte[] b = StringEncoder.GetBytes(value);

            this.WriteUINT32((uint)b.Length);
            this.WriteBytes(b, 0, b.Length);
        }
        
        /// <summary>
        /// Adds the data for a string into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The string to be added to the arr byte buffer.</param>
        /// <param name="charsetName">The character set used to encode the string.</param>
        public void WriteString(string value, string charsetName)
        {
            byte[] b;
            if (charsetName == null)
            {
                b = StringEncoder.GetBytes(value);
            }
            else
            {
                Encoding enc = Encoding.GetEncoding(charsetName);
                b = enc.GetBytes(value);
            }

            this.WriteUINT32((uint)b.Length);
            this.WriteBytes(b, 0, b.Length);
        }
        
        /// <summary>
        /// Adds the data for a string array into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="value">The string array to be added to the arr byte buffer.</param>
        public void WriteNameList(string[] value)
        {
            StringBuilder sb = new StringBuilder();
            for (int i = 0; i < value.Length; i++)
            {
                if (i > 0)
                {
                    sb.Append(',');
                }
                
                sb.Append(value[i]);
            }
            
            this.WriteString(sb.ToString());
        }

        /// <summary>
        /// Adds the data for an unsigned integer into the arr byte buffer at a supplied position.
        /// </summary>
        /// <param name="val">The unsigned integer to add to the byte buffer.</param>
        /// <param name="off">The location in the byte buffer to write to.</param>
        internal void WriteUINT32(uint val, int off)
        {
            if ((off + 4) > this.arr.Length)
            {
                this.Resize(off + 32);
            }
            
            unchecked
            {
                this.arr[off++] = (byte)(val >> 24);
                this.arr[off++] = (byte)(val >> 16);
                this.arr[off++] = (byte)(val >> 8);
                this.arr[off++] = (byte)val;
            }
        }

        /// <summary>
        /// Adds the data for an unsigned integer into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="val">The unsigned integer to add to the byte buffer.</param>
        internal void WriteUINT32(uint val)
        {
            this.WriteUINT32(val, this.pos);
            this.pos += 4;
        }

        /// <summary>
        /// Adds the data for an unsigned long into the arr byte buffer at the current position.
        /// </summary>
        /// <param name="val">The unsigned long to add to the byte buffer.</param>
        internal void WriteUINT64(ulong val)
        {
            if ((this.pos + 8) > this.arr.Length)
            {
                this.Resize(this.arr.Length + 32);
            }
            
            unchecked
            {
                this.arr[this.pos++] = (byte)(val >> 56);
                this.arr[this.pos++] = (byte)(val >> 48);
                this.arr[this.pos++] = (byte)(val >> 40);
                this.arr[this.pos++] = (byte)(val >> 32);
                this.arr[this.pos++] = (byte)(val >> 24);
                this.arr[this.pos++] = (byte)(val >> 16);
                this.arr[this.pos++] = (byte)(val >> 8);
                this.arr[this.pos++] = (byte)val;
            }
        }
        
        /// <summary>
        /// Resizes the byte array that we are writing data to.
        /// </summary>
        /// <param name="len">The length we need the byte array to be now.</param>
        private void Resize(int len)
        {
            byte[] new_arr = new byte[len];
            Array.Copy(this.arr, 0, new_arr, 0, this.arr.Length);
            this.arr = new_arr;
        }
    }
}
