// <copyright file="SFTPv3Client.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> SFTPv3Client
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Collections;
using System.Globalization;
using System.IO;
using System.Text;

using Eunomia.SSH2.Packets;
using Eunomia.SSH2.SFTP;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>SFTPv3Client</c> represents a SFTP (protocol version 3)
    /// client connection tunnelled over a SSH-2 connection. This is a very simple
    /// (synchronous) implementation.
    /// <para>
    /// Basically, most methods in this class map directly to one of
    /// the packet types described in draft-ietf-secsh-filexfer-02.txt.
    /// </para><para>
    /// Note: this is experimental code.
    /// </para><para>
    /// Error handling: the methods of this class throw IOExceptions. However, unless
    /// there is catastrophic failure, exceptions of the type {@link SFTPv3Client} will
    /// be thrown (a subclass of IOException). Therefore, you can implement more verbose
    /// behavior by checking if a thrown exception if of this type. If yes, then you
    /// can cast the exception and access detailed information about the failure.
    /// </para><para>
    /// Notes about file names, directory names and paths, copy-pasted
    /// from the specs:
    /// <list type="bullet">
    /// <item>SFTP v3 represents file names as strings. File names are
    /// assumed to use the slash ('/') character as a directory separator.</item>
    /// <item>File names starting with a slash are "absolute", and are relative to
    /// the root of the file system.  Names starting with any other character
    /// are relative to the user's default directory (home directory).</item>
    /// <item>Servers SHOULD interpret a path name component ".." as referring to
    /// the parent directory, and "." as referring to the current directory.
    /// If the server implementation limits access to certain parts of the
    /// file system, it must be extra careful in parsing file names when
    /// enforcing such restrictions.  There have been numerous reported
    /// security bugs where a ".." in a path name has allowed access outside
    /// the intended area.</item>
    /// <item>An empty path name is valid, and it refers to the user's default
    /// directory (usually the user's home directory).</item>
    /// </list>
    /// </para><para>
    /// If you are still not tired then please go on and read the comment for
    /// <see cref="Charset"></see>.
    /// </para>
    /// </summary>
    public class SFTPv3Client : IDisposable
    {
        private Session sess;
        private StreamWriter debug;

        private Stream istr;
        private Stream ostr;

        private int protocolVersion = 0;
        private Hashtable serverExtensions;

        private int nextRequest_id = 1000;

        private string charsetName = null;

        /// <summary>
        /// Initializes a new instance of the SFTPv3Client class.
        /// </summary>
        /// <param name="conn">The underlying SSH-2 connection to be used.</param>
        /// <param name="debug">A stream to write debugging information to.</param>
        public SFTPv3Client(Connection conn, StreamWriter debug)
        {
            this.serverExtensions = new Hashtable();
            if (conn == null)
            {
                throw new ArgumentException("Cannot accept null argument!");
            }

            this.debug = debug;

            if (debug != null)
            {
                debug.WriteLine("Opening session and starting SFTP subsystem.");
            }

            this.sess = conn.OpenSession();
            this.sess.StartSubsystem("sftp");

            this.istr = this.sess.StandardOutput;
            this.ostr = new BufferedStream(this.sess.StandardInput, 2048);

            if ((this.istr == null) || (this.ostr == null))
            {
                throw new IOException("There is a problem with the streams of the underlying channel.");
            }

            this.Init();
        }

        /// <summary>
        /// Initializes a new instance of the SFTPv3Client class.
        /// </summary>
        /// <param name="conn">The underlying SSH-2 connection to be used.</param>
        public SFTPv3Client(Connection conn) : this(conn, null)
        {
        }

        /// <summary>
        /// Gets or sets the charset used to convert between Java Unicode Strings and byte encodings
        /// used by the server for paths and file names. Unfortunately, the SFTP v3 draft
        /// says NOTHING about such conversions (well, with the exception of error messages
        /// which have to be in UTF-8). Newer drafts specify to use UTF-8 for file names
        /// (if I remember correctly). However, a quick test using OpenSSH serving a EXT-3
        /// filesystem has shown that UTF-8 seems to be a bad choice for SFTP v3 (tested with
        /// filenames containing german umlauts). "windows-1252" seems to work better for Europe.
        /// Luckily, "windows-1252" is the platform default in my case =).
        /// <para>
        /// If you don't set anything, then the platform default will be used (this is the default
        /// behavior).
        /// </para>
        /// </summary>
        /// <exception cref="IOException">If the character set is unknown.</exception>
        public string Charset
        {
            get
            {
                return this.charsetName;
            }
            
            set
            {
                if (value == null)
                {
                    this.charsetName = value;
                    return;
                }

                try
                {
                    Encoding.GetEncoding(value);
                }
                catch (Exception)
                {
                    throw (IOException)new IOException("This charset (" + value + ") is not supported");
                }

                this.charsetName = value;
            }
        }
        
        /// <summary>
        /// Gets the negotiated SFTP protocol version between the client and the server.
        /// For example, "3".
        /// </summary>
        public int ProtocolVersion
        {
            get { return this.protocolVersion; }
        }

        /// <summary>
        /// Retrieve the file attributes of an open file.
        /// </summary>
        /// <param name="handle">A SFTPv3FileHandle handle.</param>
        /// <returns>A SFTPv3FileAttributes object.</returns>
        /// <exception cref="IOException">If data is received that we don't make sense of.</exception>
        public SFTPv3FileAttributes Fstat(SFTPv3FileHandle handle)
        {
            this.CheckHandleValidAndOpen(handle);

            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(handle.FileHandle, 0, handle.FileHandle.Length);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_FSTAT...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferFstat, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            if (this.debug != null)
            {
                this.debug.WriteLine("Got REPLY.");
                this.debug.Flush();
            }

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferAttributes)
            {
                return this.ReadAttrs(tr);
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            throw new SFTPException(tr.ReadString(), errorCode);
        }

        /// <summary>
        /// Retrieve the file attributes of a file. This method
        /// follows symbolic links on the server.
        /// See <see cref="Lstat"></see>.
        /// </summary>
        /// <param name="path">
        /// See the <see cref="SFTPv3Client"></see> for the class for more details.
        /// </param>
        /// <returns>
        /// A SFTPv3FileAttributes object.
        /// </returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileAttributes Stat(string path)
        {
            return this.StatBoth(path, FTPPacket.SSHFileTransferStat);
        }

        /// <summary>
        /// Retrieve the file attributes of a file. This method
        /// does NOT follow symbolic links on the server.
        /// <see cref="Stat"></see>.
        /// </summary>
        /// <param name="path">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>A SFTPv3FileAttributes object.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileAttributes Lstat(string path)
        {
            return this.StatBoth(path, FTPPacket.SSHFileTransferLstat);
        }

        /// <summary>
        /// Read the target of a symbolic link.
        /// </summary>
        /// <param name="path">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>The target of the link.</returns>
        /// <exception cref="IOException"></exception>
        public string ReadLink(string path)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(path, this.charsetName);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_READLINK...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferReadLink, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            if (this.debug != null)
            {
                this.debug.WriteLine("Got REPLY.");
                this.debug.Flush();
            }

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferName)
            {
                int count = tr.ReadUINT32();

                if (count != 1)
                {
                    throw new IOException("The server sent an invalid SSH_FXP_NAME packet.");
                }

                return tr.ReadString(this.charsetName);
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            throw new SFTPException(tr.ReadString(), errorCode);
        }

        /// <summary>
        /// Modify the attributes of a file. Used for operations such as changing
        /// the ownership, permissions or access times, as well as for truncating a file.
        /// </summary>
        /// <param name="path">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="attr">
        /// A SFTPv3FileAttributes object. Specifies the modifications to be
        /// made to the attributes of the file. Empty fields will be ignored.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Setstat(string path, SFTPv3FileAttributes attr)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(path, this.charsetName);
            tw.WriteBytes(SFTPv3Client.CreateAttrs(attr));

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_SETSTAT...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferSetstat, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Modify the attributes of a file. Used for operations such as changing
        /// the ownership, permissions or access times, as well as for truncating a file.
        /// </summary>
        /// <param name="handle">A SFTPv3FileHandle handle.</param>
        /// <param name="attr">
        /// A SFTPv3FileAttributes object. Specifies the modifications to be
        /// made to the attributes of the file. Empty fields will be ignored.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Fsetstat(SFTPv3FileHandle handle, SFTPv3FileAttributes attr)
        {
            this.CheckHandleValidAndOpen(handle);

            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(handle.FileHandle, 0, handle.FileHandle.Length);
            tw.WriteBytes(SFTPv3Client.CreateAttrs(attr));

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_FSETSTAT...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferFsetstat, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Create a symbolic link on the server. Creates a link "src" that points
        /// to "target".
        /// </summary>
        /// <param name="source">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="target">
        /// Like src, see the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void CreateSymlink(string source, string target)
        {
            int req_id = this.GenerateNextRequestID();

            // Either I am too stupid to understand the SFTP draft
            // or the OpenSSH guys changed the semantics of src and target.
            TypesWriter tw = new TypesWriter();
            tw.WriteString(target, this.charsetName);
            tw.WriteString(source, this.charsetName);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_SYMLINK...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferSymbolicLink, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Have the server canonicalize any given path name to an absolute path.
        /// This is useful for converting path names containing ".." components or
        /// relative pathnames without a leading slash into absolute paths.
        /// </summary>
        /// <param name="path">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>An absolute path.</returns>
        /// <exception cref="IOException"></exception>
        public string CanonicalPath(string path)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(path, this.charsetName);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_REALPATH...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferRealPath, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            if (this.debug != null)
            {
                this.debug.WriteLine("Got REPLY.");
                this.debug.Flush();
            }

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferName)
            {
                int count = tr.ReadUINT32();

                if (count != 1)
                {
                    throw new IOException("The server sent an invalid SSH_FXP_NAME packet.");
                }

                return tr.ReadString(this.charsetName);
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            throw new SFTPException(tr.ReadString(), errorCode);
        }

        /// <summary>
        /// Close this SFTP session. NEVER forget to call this method to free up
        /// resources - even if you got an exception from one of the other methods.
        /// Sometimes these other methods may throw an exception, saying that the
        /// underlying channel is closed (this can happen, e.g., if the other server
        /// sent a close message.) However, as long as you have not called the
        /// <c>Close()</c> method, you are likely wasting resources.
        /// </summary>
        public void Close()
        {
            this.sess.Close();
        }

        /// <summary>
        /// List the contents of a directory.
        /// </summary>
        /// <param name="dirName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>
        /// An ArrayList containing <see cref="SFTPv3DirectoryEntry"></see> objects.
        /// </returns>
        /// <exception cref="IOException"></exception>
        public ArrayList Ls(string dirName)
        {
            byte[] handle = this.OpenDirectory(dirName);
            ArrayList result = this.ScanDirectory(handle);
            this.CloseHandle(handle);
            return result;
        }

        /// <summary>
        /// Create a new directory.
        /// </summary>
        /// <param name="dirName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="posixPermissions">
        /// The permissions for this directory, e.g., "0700" (remember that
        /// this is octal noation). The server will likely apply a umask.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Mkdir(string dirName, int posixPermissions)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(dirName, this.charsetName);
            tw.WriteUINT32(AttribFlags.SSHFileTransferAttributePermissions);
            tw.WriteUINT32((uint)posixPermissions);

            this.SendMessage(FTPPacket.SSHFileTransferMkdir, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Remove a file.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Rm(string fileName)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(fileName, this.charsetName);

            this.SendMessage(FTPPacket.SSHFileTransferRemove, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Remove an empty directory.
        /// </summary>
        /// <param name="dirName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Rmdir(string dirName)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(dirName, this.charsetName);

            this.SendMessage(FTPPacket.SSHFileTransferRmdir, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Move a file or directory.
        /// </summary>
        /// <param name="oldPath">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="newPath">
        /// Like oldPath, see the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <exception cref="IOException"></exception>
        public void Mv(string oldPath, string newPath)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(oldPath, this.charsetName);
            tw.WriteString(newPath, this.charsetName);

            this.SendMessage(FTPPacket.SSHFileTransferRename, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Open a file for reading.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle OpenFileRO(string fileName)
        {
            return this.OpenFile(fileName, 0x00000001, null); // SSH_FXF_READ
        }

        /// <summary>
        /// Open a file for reading and writing.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle OpenFileRW(string fileName)
        {
            return this.OpenFile(fileName, 0x00000003, null); // SSH_FXF_READ | SSH_FXF_WRITE
        }

        /// <summary>
        /// Create a file and open it for reading and writing.
        /// Same as <see cref="CreateFile(string, SFTPv3FileAttributes)"></see> where
        /// the SFTPv3FileAttributes is set to <c>null</c>.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle CreateFile(string fileName)
        {
            return this.CreateFile(fileName, null);
        }

        /// <summary>
        /// Create a file and open it for reading and writing.
        /// You can specify the default attributes of the file (the server may or may
        /// not respect your wishes).
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="attr">
        /// May be <c>null</c> to use server defaults. Probably only
        /// the <c>uid</c>, <c>gid</c> and <c>permissions</c>
        /// (remember the server may apply a umask) entries of the <see cref="SFTPv3FileHandle"></see>
        /// structure make sense. You need only to set those fields where you want
        /// to override the server's defaults.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle CreateFile(string fileName, SFTPv3FileAttributes attr)
        {
            return this.OpenFile(fileName, 0x00000008 | 0x00000003, attr); // SSH_FXF_CREAT | SSH_FXF_READ | SSH_FXF_WRITE
        }

        /// <summary>
        /// Create a file (truncate it if it already exists) and open it for reading and writing.
        /// Same as <see cref="CreateFileTruncate(string, SFTPv3FileAttributes)"></see>,
        /// where the SFTPv3FileAttributes is null.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle CreateFileTruncate(string fileName)
        {
            return this.CreateFileTruncate(fileName, null);
        }

        /// <summary>
        /// Create a file (truncate it if it already exists) and open it for reading and writing.
        /// You can specify the default attributes of the file (the server may or may
        /// not respect your wishes).
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="attr">
        /// May be <c>null</c> to use server defaults. Probably only
        /// the <c>uid</c>, <c>gid</c> and <c>permissions</c>
        /// (remember the server may apply a umask) entries of the <see cref="SFTPv3FileHandle"></see>
        /// structure make sense. You need only to set those fields where you want
        /// to override the server's defaults.
        /// </param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        /// <exception cref="IOException"></exception>
        public SFTPv3FileHandle CreateFileTruncate(string fileName, SFTPv3FileAttributes attr)
        {
            return this.OpenFile(fileName, 0x00000018 | 0x00000003, attr); // SSH_FXF_CREAT | SSH_FXF_TRUNC | SSH_FXF_READ | SSH_FXF_WRITE
        }

        /// <summary>
        /// Read bytes from a file. No more than 32768 bytes may be read at once.
        /// <para>
        /// <list type="bullet">
        /// <item>The server will read as many bytes as it can from the file (up to <c>len</c>),
        /// and return them.</item>
        /// <item>If EOF is encountered before reading any data, <c>-1</c> is returned.</item>
        /// <item>If an error occurs, an exception is thrown</item>.
        /// <item>For normal disk files, it is guaranteed that the server will return the specified
        /// number of bytes, or up to end of file. For, e.g., device files this may return
        /// fewer bytes than requested.</item>
        /// </list>
        /// </para>
        /// </summary>
        /// <param name="handle">A SFTPv3FileHandle handle.</param>
        /// <param name="fileOffset">Offset (in bytes) in the file.</param>
        /// <param name="dst">The destination byte array.</param>
        /// <param name="destinationOffset">Offset in the destination byte array.</param>
        /// <param name="len">How many bytes to read, 0 &lt; len &lt;= 32768 bytes.</param>
        /// <returns>
        /// The number of bytes that could be read, may be less than requested if
        /// the end of the file is reached, -1 is returned in case of <c>EOF</c>.
        /// </returns>
        /// <exception cref="IOException"></exception>
        public int Read(SFTPv3FileHandle handle, long fileOffset, byte[] dst, int destinationOffset, int len)
        {
            this.CheckHandleValidAndOpen(handle);

            if ((len > 32768) || (len <= 0))
            {
                throw new ArgumentException("invalid len argument");
            }

            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(handle.FileHandle, 0, handle.FileHandle.Length);
            tw.WriteUINT64((ulong)fileOffset);
            tw.WriteUINT32((uint)len);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_READ...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferRead, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferData)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("Got SSH_FXP_DATA...");
                    this.debug.Flush();
                }

                int readLen = tr.ReadUINT32();

                if ((readLen < 0) || (readLen > len))
                {
                    throw new IOException("The server sent an invalid length field.");
                }

                tr.ReadBytes(dst, destinationOffset, readLen);

                return readLen;
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            if (errorCode == ErrorCodes.SSHFileTransferEOF)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("Got SSH_FXP_EOF.");
                    this.debug.Flush();
                }

                return -1;
            }

            string errorMessage = tr.ReadString();

            throw new SFTPException(errorMessage, errorCode);
        }

        /// <summary>
        /// Write bytes to a file. If <c>len</c> &gt; 32768, then the write operation will
        /// be split into multiple writes.
        /// </summary>
        /// <param name="handle">A SFTPv3FileHandle handle.</param>
        /// <param name="fileOffset">Offset (in bytes) in the file.</param>
        /// <param name="source">The source byte array.</param>
        /// <param name="sourceOffset">Offset in the source byte array.</param>
        /// <param name="len">How many bytes to write.</param>
        /// <exception cref="IOException"></exception>
        public void Write(SFTPv3FileHandle handle, long fileOffset, byte[] source, int sourceOffset, int len)
        {
            this.CheckHandleValidAndOpen(handle);

            while (len > 0)
            {
                int writeRequestLen = len;

                if (writeRequestLen > 32768)
                {
                    writeRequestLen = 32768;
                }

                int req_id = this.GenerateNextRequestID();

                TypesWriter tw = new TypesWriter();
                tw.WriteString(handle.FileHandle, 0, handle.FileHandle.Length);
                tw.WriteUINT64((ulong)fileOffset);
                tw.WriteString(source, sourceOffset, writeRequestLen);

                if (this.debug != null)
                {
                    this.debug.WriteLine("Sending SSH_FXP_WRITE...");
                    this.debug.Flush();
                }

                this.SendMessage(FTPPacket.SSHFileTransferWrite, req_id, tw.GetBytes());

                fileOffset += writeRequestLen;

                sourceOffset += writeRequestLen;
                len -= writeRequestLen;

                byte[] resp = this.ReceiveMessage(34000);

                TypesReader tr = new TypesReader(resp);

                int t = tr.ReadByte();

                int rep_id = tr.ReadUINT32();
                if (rep_id != req_id)
                {
                    throw new IOException("The server sent an invalid id field.");
                }

                if (t != FTPPacket.SSHFileTransferStatus)
                {
                    throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
                }

                int errorCode = tr.ReadUINT32();

                if (errorCode == ErrorCodes.SSHFileTransferOk)
                {
                    continue;
                }

                string errorMessage = tr.ReadString();

                throw new SFTPException(errorMessage, errorCode);
            }
        }

        /// <summary>
        /// Close a file.
        /// </summary>
        /// <param name="handle">A SFTPv3FileHandle handle.</param>
        /// <exception cref="IOException"></exception>
        public void CloseFile(SFTPv3FileHandle handle)
        {
            try
            {
                if (handle.Closed == false)
                {
                    this.CloseHandle(handle.FileHandle);
                }
            }
            finally
            {
                handle.Closed = true;
            }
        }
        
        /// <summary>
        /// Implement the public IDisposable interface method.
        /// </summary>
        public void Dispose()
        {
            this.Dispose(true);
            GC.SuppressFinalize(this);
        }
        
        /// <summary>
        /// Implement the protected IDisposable interface method.
        /// </summary>
        /// <param name="disposing">True if we are disposing.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                this.ostr.Close();
            }
        }
        
        /// <summary>
        /// Convert file attributes to a byte array that can be sent.
        /// </summary>
        /// <param name="attr">The file attributes to convert.</param>
        /// <returns>The byte array to send.</returns>
        private static byte[] CreateAttrs(SFTPv3FileAttributes attr)
        {
            TypesWriter tw = new TypesWriter();

            uint attrFlags = 0;

            if (attr == null)
            {
                tw.WriteUINT32(0);
            }
            else
            {
                attrFlags = attrFlags | AttribFlags.SSHFileTransferAttributeSize;

                attrFlags = attrFlags | AttribFlags.SSHFileTransferAttributeV3UIDGID;

                attrFlags = attrFlags | AttribFlags.SSHFileTransferAttributePermissions;

                attrFlags = attrFlags | AttribFlags.SSHFileTransferAttributeV3ACModtime;

                tw.WriteUINT32((uint)attrFlags);

                tw.WriteUINT64((ulong)attr.Size);

                tw.WriteUINT32((uint)attr.UID);
                tw.WriteUINT32((uint)attr.GID);

                tw.WriteUINT32((uint)attr.Permissions);

                tw.WriteUINT32((uint)attr.Atime);
                tw.WriteUINT32((uint)attr.Mtime);
            }

            return tw.GetBytes();
        }

        /// <summary>
        /// Convert a byte array into a string.  If the character is outside of the
        /// normal ASCII range it is converted to a hex string.
        /// </summary>
        /// <param name="b">The byte array.</param>
        /// <param name="off">The offset into the byte array where we begin reading.</param>
        /// <param name="len">The number of bytes to read.</param>
        /// <returns>Returns the converted sting.</returns>
        private static string ExpandString(byte[] b, int off, int len)
        {
            StringBuilder sb = new StringBuilder();

            for (int i = 0; i < len; i++)
            {
                int c = b[off + i] & 0xff;

                if ((c >= 32) && (c <= 126))
                {
                    sb.Append((char)c);
                }
                else
                {
                    sb.Append("{0x" + c.ToString("2X", CultureInfo.InvariantCulture) + "}");
                }
            }

            return sb.ToString();
        }

        /// <summary>
        /// Gets a byte array that is a directory handle and process the returned list of files in the directory.
        /// </summary>
        /// <param name="handle">The handle to the directory.</param>
        /// <returns>An array list with all the files.</returns>
        /// <exception cref="IOException"></exception>
        /// <exception cref="SFTPException">If we receive an error code.</exception>
        private ArrayList ScanDirectory(byte[] handle)
        {
            ArrayList files = new ArrayList(10);

            while (true)
            {
                int req_id = this.GenerateNextRequestID();

                TypesWriter tw = new TypesWriter();
                tw.WriteString(handle, 0, handle.Length);

                if (this.debug != null)
                {
                    this.debug.WriteLine("Sending SSH_FXP_READDIR...");
                    this.debug.Flush();
                }

                this.SendMessage(FTPPacket.SSHFileTransferReadDirectory, req_id, tw.GetBytes());

                byte[] resp = this.ReceiveMessage(34000);

                if (this.debug != null)
                {
                    this.debug.WriteLine("Got REPLY.");
                    this.debug.Flush();
                }

                TypesReader tr = new TypesReader(resp);

                int t = tr.ReadByte();

                int rep_id = tr.ReadUINT32();
                if (rep_id != req_id)
                {
                    throw new IOException("The server sent an invalid id field.");
                }

                if (t == FTPPacket.SSHFileTransferName)
                {
                    int count = tr.ReadUINT32();

                    if (this.debug != null)
                    {
                        this.debug.WriteLine("Parsing " + count + " name entries...");
                    }

                    while (count > 0)
                    {
                        SFTPv3DirectoryEntry dirEnt = new SFTPv3DirectoryEntry();

                        dirEnt.FileName = tr.ReadString(this.charsetName);
                        dirEnt.LongEntry = tr.ReadString(this.charsetName);

                        dirEnt.Attributes = this.ReadAttrs(tr);
                        files.Add(dirEnt);

                        if (this.debug != null)
                        {
                            this.debug.WriteLine("File: '" + dirEnt.FileName + "'");
                        }
                        
                        count--;
                    }

                    continue;
                }

                if (t != FTPPacket.SSHFileTransferStatus)
                {
                    throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
                }

                int errorCode = tr.ReadUINT32();

                if (errorCode == ErrorCodes.SSHFileTransferEOF)
                {
                    return files;
                }

                throw new SFTPException(tr.ReadString(), errorCode);
            }
        }

        /// <summary>
        /// Opens up a directory in preparation to list the contents.  Used by "Ls".
        /// </summary>
        /// <param name="path">The path on the remote server to get the directory contents.</param>
        /// <returns>Returns the byte array with the directory handle in it.</returns>
        private byte[] OpenDirectory(string path)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(path, this.charsetName);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_OPENDIR...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferOpenDirectory, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferHandle)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("Got SSH_FXP_HANDLE.");
                    this.debug.Flush();
                }

                byte[] handle = tr.ReadByteString();
                return handle;
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();
            string errorMessage = tr.ReadString();

            throw new SFTPException(errorMessage, errorCode);
        }

        /// <summary>
        /// Called from the constructors.  Gets the SFTP version from the server (needs to be '3').
        /// Also reads and stores any server extensions.
        /// </summary>
        /// <exception cref="IOException"></exception>
        private void Init()
        {
            // Send SSHFileTransferINIT (version 3)
            int client_version = 3;

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_INIT (" + client_version + ")...");
            }

            TypesWriter tw = new TypesWriter();
            tw.WriteUINT32((uint)client_version);
            this.SendMessage(FTPPacket.SSHFileTransferInitialize, 0, tw.GetBytes());

            // Receive SSHFileTransferVERSION
            if (this.debug != null)
            {
                this.debug.WriteLine("Waiting for SSH_FXP_VERSION...");
            }

            TypesReader tr = new TypesReader(this.ReceiveMessage(34000)); // Should be enough for any reasonable server

            int type = tr.ReadByte();

            if (type != FTPPacket.SSHFileTransferVersion)
            {
                throw new IOException("The server did not send a SSH_FXP_VERSION packet (got " + type + ")");
            }

            this.protocolVersion = tr.ReadUINT32();

            if (this.debug != null)
            {
                this.debug.WriteLine("SSH_FXP_VERSION: this.protocol_version = " + this.protocolVersion);
            }

            if (this.protocolVersion != 3)
            {
                throw new IOException("Server version " + this.protocolVersion + " is currently not supported");
            }

            // Read and save extensions (if any) for later use
            while (tr.Remain != 0)
            {
                string name = tr.ReadString();
                byte[] value = tr.ReadByteString();
                this.serverExtensions.Add(name, value);

                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FXP_VERSION: extension: " + name + " = '" + SFTPv3Client.ExpandString(value, 0, value.Length)
                                         + "'");
                }
            }
        }

        /// <summary>
        /// Open a file.  Returns a file handle from a file name, flags and file attributes.
        /// </summary>
        /// <param name="fileName">
        /// See the <see cref="SFTPv3Client"></see> comment for the class for more details.
        /// </param>
        /// <param name="flags">The file open flags.</param>
        /// <param name="attr">The file attributes.</param>
        /// <returns>A SFTPv3FileHandle handle.</returns>
        private SFTPv3FileHandle OpenFile(string fileName, int flags, SFTPv3FileAttributes attr)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(fileName, this.charsetName);
            tw.WriteUINT32((uint)flags);
            tw.WriteBytes(SFTPv3Client.CreateAttrs(attr));

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_OPEN...");
                this.debug.Flush();
            }

            this.SendMessage(FTPPacket.SSHFileTransferOpen, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferHandle)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("Got SSH_FXP_HANDLE.");
                    this.debug.Flush();
                }

                return new SFTPv3FileHandle(this, tr.ReadByteString());
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();
            string errorMessage = tr.ReadString();

            throw new SFTPException(errorMessage, errorCode);
        }

        /// <summary>
        /// Check a file handle to make sure it's 'ours' and not closed.
        /// </summary>
        /// <param name="handle">The SFTPv3FileHandle handle to check.</param>
        private void CheckHandleValidAndOpen(SFTPv3FileHandle handle)
        {
            if (handle.Client != this)
            {
                throw new IOException("The file handle was created with another SFTPv3FileHandle instance.");
            }

            if (handle.Closed == true)
            {
                throw new IOException("The file handle is closed.");
            }
        }

        /// <summary>
        /// Sends a SFTP message.
        /// </summary>
        /// <param name="type">The SFTP packet type.</param>
        /// <param name="requestId">A sequential generated request id.</param>
        /// <param name="msg">The byte array that contains the message.</param>
        /// <param name="off">The starting offset into the message.</param>
        /// <param name="len">The number of bytes to send.</param>
        private void SendMessage(int type, int requestId, byte[] msg, int off, int len)
        {
            int msgLength = len + 1;

            if (type != FTPPacket.SSHFileTransferInitialize)
            {
                msgLength += 4;
            }

            unchecked
            {
                this.ostr.WriteByte((byte)(msgLength >> 24));
                this.ostr.WriteByte((byte)(msgLength >> 16));
                this.ostr.WriteByte((byte)(msgLength >> 8));
                this.ostr.WriteByte((byte)msgLength);
                this.ostr.WriteByte((byte)type);
            }

            if (type != FTPPacket.SSHFileTransferInitialize)
            {
                unchecked
                {
                    this.ostr.WriteByte((byte)(requestId >> 24));
                    this.ostr.WriteByte((byte)(requestId >> 16));
                    this.ostr.WriteByte((byte)(requestId >> 8));
                    this.ostr.WriteByte((byte)requestId);
                }
            }

            this.ostr.Write(msg, off, len);
            this.ostr.Flush();
        }

        /// <summary>
        /// This is what gets called...it just calls the other SendMessage.
        /// </summary>
        /// <param name="type">The SFTP packet type.</param>
        /// <param name="requestId">A sequential generated request id.</param>
        /// <param name="msg">The byte array that contains the message.</param>
        private void SendMessage(int type, int requestId, byte[] msg)
        {
            this.SendMessage(type, requestId, msg, 0, msg.Length);
        }

        /// <summary>
        /// Process read bytes...called from ReceivedMessage method. 
        /// </summary>
        /// <param name="buff">A byte array with the received bytes in it.</param>
        /// <param name="pos">The starting position in the byte array to read from.</param>
        /// <param name="len">The number of bytes to read.</param>
        /// <exception cref="IOException"></exception>
        private void ReadBytes(byte[] buff, int pos, int len)
        {
            while (len > 0)
            {
                int count = this.istr.Read(buff, pos, len);
                if (count < 0)
                {
                    throw new IOException("Unexpected end of sftp stream.");
                }

                if ((count == 0) || (count > len))
                {
                    throw new IOException("Underlying stream implementation is bogus!");
                }

                len -= count;
                pos += count;
            }
        }

        /// <summary>
        /// Read a message and guarantee that the <b>contents</b> is not larger than
        /// <c>maxlen</c> bytes.
        /// <para>
        /// Note: receiveMessage(34000) actually means that the message may be up to 34004
        /// bytes (the length attribute preceeding the contents is 4 bytes).
        /// </para>
        /// </summary>
        /// <param name="maxlen">The maximum length we should read.</param>
        /// <returns>A byte array with the message contents.</returns>
        /// <exception cref="IOException"></exception>
        private byte[] ReceiveMessage(int maxlen)
        {
            byte[] msgLength = new byte[4];

            this.ReadBytes(msgLength, 0, 4);

            int len = ((msgLength[0] & 0xff) << 24) |
                ((msgLength[1] & 0xff) << 16) |
                ((msgLength[2] & 0xff) << 8) |
                (msgLength[3] & 0xff);

            if ((len > maxlen) || (len <= 0))
            {
                throw new IOException("Illegal sftp packet len: " + len);
            }

            byte[] msg = new byte[len];

            this.ReadBytes(msg, 0, len);

            return msg;
        }

        /// <summary>
        /// Generates and returns a sequential id (int).
        /// </summary>
        /// <returns>An int with the new id in it.</returns>
        private int GenerateNextRequestID()
        {
            lock (this)
            {
                return this.nextRequest_id++;
            }
        }

        /// <summary>
        /// Close a file (or directory?) handle.
        /// </summary>
        /// <param name="handle">The handle to close.</param>
        private void CloseHandle(byte[] handle)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(handle, 0, handle.Length);

            this.SendMessage(FTPPacket.SSHFileTransferClose, req_id, tw.GetBytes());

            this.ExpectStatusOKMessage(req_id);
        }

        /// <summary>
        /// Read a SFTPv3FileAttributes from a TypesReader.
        /// </summary>
        /// <param name="tr">A TypesReader instance with file attribute information in it.</param>
        /// <returns>The newly read and created SFTPv3FileAttributes instance.</returns>
        private SFTPv3FileAttributes ReadAttrs(TypesReader tr)
        {
            // uint32   flags
            // uint64   size           present only if flag SSH_FILEXFER_ATTR_SIZE
            // uint32   uid            present only if flag SSH_FILEXFER_ATTR_V3_UIDGID
            // uint32   gid            present only if flag SSH_FILEXFER_ATTR_V3_UIDGID
            // uint32   permissions    present only if flag SSH_FILEXFER_ATTR_PERMISSIONS
            // uint32   atime          present only if flag SSH_FILEXFER_ATTR_V3_ACMODTIME
            // uint32   mtime          present only if flag SSH_FILEXFER_ATTR_V3_ACMODTIME
            // uint32   extended_count present only if flag SSH_FILEXFER_ATTR_EXTENDED
            // string   extended_type
            // string   extended_data
            //          more extended data (extended_type - extended_data pairs),
            //          so that number of pairs equals extended_count
            SFTPv3FileAttributes fa = new SFTPv3FileAttributes();

            int flags = tr.ReadUINT32();

            if ((flags & AttribFlags.SSHFileTransferAttributeSize) != 0)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FILEXFER_ATTR_SIZE");
                }

                fa.Size = tr.ReadUINT64();
            }

            if ((flags & AttribFlags.SSHFileTransferAttributeV3UIDGID) != 0)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FILEXFER_ATTR_V3_UIDGID");
                }

                fa.UID = tr.ReadUINT32();
                fa.GID = tr.ReadUINT32();
            }

            if ((flags & AttribFlags.SSHFileTransferAttributePermissions) != 0)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FILEXFER_ATTR_PERMISSIONS");
                }

                fa.Permissions = tr.ReadUINT32();
            }

            if ((flags & AttribFlags.SSHFileTransferAttributeV3ACModtime) != 0)
            {
                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FILEXFER_ATTR_V3_ACMODTIME");
                }

                fa.Atime = tr.ReadUINT32();
                fa.Mtime = tr.ReadUINT32();
            }

            if ((flags & AttribFlags.SSHFileTransferAttributeExtended) != 0)
            {
                int count = tr.ReadUINT32();

                if (this.debug != null)
                {
                    this.debug.WriteLine("SSH_FILEXFER_ATTR_EXTENDED (" + count + ")");
                }

                // Read it anyway to detect corrupt packets
                while (count > 0)
                {
                    tr.ReadByteString();
                    tr.ReadByteString();
                    count--;
                }
            }

            return fa;
        }

        /// <summary>
        /// Reads a SFTPv3FileAttributes from a path and a statMethod,
        /// (either FTPPacket.SSHFileTransferSTAT or SSHFileTransferLSTAT).
        /// </summary>
        /// <param name="path">The path to the file.</param>
        /// <param name="statMethod">The statMethod...see the method summary.</param>
        /// <returns>The newly read and created SFTPv3FileAttributes instance.</returns>
        /// <exception cref="IOException"></exception>
        private SFTPv3FileAttributes StatBoth(string path, int statMethod)
        {
            int req_id = this.GenerateNextRequestID();

            TypesWriter tw = new TypesWriter();
            tw.WriteString(path, this.charsetName);

            if (this.debug != null)
            {
                this.debug.WriteLine("Sending SSH_FXP_STAT/SSH_FXP_LSTAT...");
                this.debug.Flush();
            }

            this.SendMessage(statMethod, req_id, tw.GetBytes());

            byte[] resp = this.ReceiveMessage(34000);

            if (this.debug != null)
            {
                this.debug.WriteLine("Got REPLY.");
                this.debug.Flush();
            }

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != req_id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t == FTPPacket.SSHFileTransferAttributes)
            {
                return this.ReadAttrs(tr);
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            throw new SFTPException(tr.ReadString(), errorCode);
        }

        /// <summary>
        /// Receive a status or other message.
        /// </summary>
        /// <param name="id">The sequential id.</param>
        /// <exception cref="IOException"></exception>
        /// <exception cref="SFTPException">If we get a error from the server.</exception>
        private void ExpectStatusOKMessage(int id)
        {
            byte[] resp = this.ReceiveMessage(34000);

            if (this.debug != null)
            {
                this.debug.WriteLine("Got REPLY.");
                this.debug.Flush();
            }

            TypesReader tr = new TypesReader(resp);

            int t = tr.ReadByte();

            int rep_id = tr.ReadUINT32();
            if (rep_id != id)
            {
                throw new IOException("The server sent an invalid id field.");
            }

            if (t != FTPPacket.SSHFileTransferStatus)
            {
                throw new IOException("The SFTP server sent an unexpected packet type (" + t + ")");
            }

            int errorCode = tr.ReadUINT32();

            if (errorCode == ErrorCodes.SSHFileTransferOk)
            {
                return;
            }

            throw new SFTPException(tr.ReadString(), errorCode);
        }
    }
}
