// <copyright file="SFTPv3FileAttributes.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> SFTPv3FileAttributes
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Text;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>SFTPv3FileAttributes</c> object represents detail information
    /// about a file on the server. Not all fields may/must be present.
    /// </summary>
    public class SFTPv3FileAttributes
    {
        private long size;
        private int uid;
        private int gid;
        private int permissions;
        private int atime;
        private int mtime;
        
        /// <summary>
        /// Gets or sets the Size attribute.
        /// </summary>
        public long Size
        {
            get { return this.size; }
            set { this.size = value; }
        }

        /// <summary>
        /// Gets or sets the UID attribute.
        /// </summary>
        public int UID
        {
            get { return this.uid; }
            set { this.uid = value; }
        }

        /// <summary>
        /// Gets or sets the GID attribute.
        /// </summary>
        public int GID
        {
            get { return this.gid; }
            set { this.gid = value; }
        }

        /// <summary>
        /// Gets or sets the Posix permissions.
        /// </summary>
        /// <remarks>
        ///  Here is a list:
        /// Note: these numbers are all OCTAL.
        /// --
        /// S_IFMT     0170000   bitmask for the file type bitfields
        /// S_IFSOCK   0140000   socket
        /// S_IFLNK    0120000   symbolic link
        /// S_IFREG    0100000   regular file
        /// S_IFBLK    0060000   block device
        /// S_IFDIR    0040000   directory
        /// S_IFCHR    0020000   character device
        /// S_IFIFO    0010000   fifo
        /// S_ISUID    0004000   set UID bit
        /// S_ISGID    0002000   set GID bit
        /// S_ISVTX    0001000   sticky bit
        /// --
        /// S_IRWXU    00700     mask for file owner permissions
        /// S_IRUSR    00400     owner has read permission
        /// S_IWUSR    00200     owner has write permission
        /// S_IXUSR    00100     owner has execute permission
        /// S_IRWXG    00070     mask for group permissions
        /// S_IRGRP    00040     group has read permission
        /// S_IWGRP    00020     group has write permission
        /// S_IXGRP    00010     group has execute permission
        /// S_IRWXO    00007     mask for permissions for others (not in group)
        /// S_IROTH    00004     others have read permission
        /// S_IWOTH    00002     others have write permisson
        /// S_IXOTH    00001     others have execute permission
        /// </remarks>
        public int Permissions
        {
            get { return this.permissions; }
            set { this.permissions = value; }
        }

        /// <summary>
        /// Gets or sets the ATIME attribute. Represented as seconds from Jan 1, 1970 in UTC.
        /// </summary>
        public int Atime
        {
            get { return this.atime; }
            set { this.atime = value; }
        }

        /// <summary>
        /// Gets or sets the MTIME attribute. Represented as seconds from Jan 1, 1970 in UTC.
        /// </summary>
        public int Mtime
        {
            get { return this.mtime; }
            set { this.mtime = value; }
        }

        /// <summary>
        /// Gets a value indicating whether the entry is a Directory.
        /// </summary>
        public bool Directory
        {
            get
            {
                return (this.permissions & 0040000) != 0;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the entry is a regular file.
        /// </summary>
        public bool RegularFile
        {
            get
            {
                return (this.permissions & 0100000) != 0;
            }
        }

        /// <summary>
        /// Gets a value indicating whether the entry is a Symlink.
        /// </summary>
        public bool Symlink
        {
            get
            {
                return (this.permissions & 0120000) != 0;
            }
        }

        /// <summary>
        /// Gets the POSIX permissions into a 7 digit octal representation.
        /// Note: the returned value is first masked with <c>0177777</c>.
        /// </summary>
        public string OctalPermissions
        {
            get
            {
                string res = Convert.ToString((int)(this.permissions & 0177777), 8);

                StringBuilder sb = new StringBuilder();

                int leadingZeros = 7 - res.Length;

                while (leadingZeros > 0)
                {
                    sb.Append('0');
                    leadingZeros--;
                }

                sb.Append(res);

                return sb.ToString();
            }
        }
    }
}
