// <copyright file="Session.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Session
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Globalization;
using System.IO;
using System.Text;

using Eunomia.SSH2.Channel;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2
{
    /// <summary>
    /// A <c>Session</c> is a remote execution of a program. "Program" means
    /// in this context either a shell, an application or a system command. The
    /// program may or may not have a tty. Only one single program can be started on
    /// a session. However, multiple sessions can be active simultaneously.
    /// </summary>
    public class Session
    {
        private ChannelManager cm;
        private Eunomia.SSH2.Channel.SSHChannel cn;

        private bool flagPtyRequested = false;
        private bool flagX11Requested = false;
        private bool flagExecutionStarted = false;
        private bool flagClosed = false;

        private string x11FakeCookie = null;

        private SecureRandom rnd;

        /// <summary>
        /// Initializes a new instance of the Session class.
        /// </summary>
        /// <param name="cm">The underlying ConnectionManager.</param>
        /// <param name="random">An instance of SecureRandom (used for random bytes).</param>
        public Session(ChannelManager cm, SecureRandom random)
        {
            this.cm = cm;
            this.cn = cm.OpenSessionChannel();
            this.rnd = random;
        }

        /// <summary>
        /// Gets the Standard Output stream for this Session.
        /// </summary>
        public Stream StandardOutput
        {
            get { return this.cn.StandardOutputStream; }
        }

        /// <summary>
        /// Gets the Standard Error stream for this Session.
        /// </summary>
        public Stream StandardError
        {
            get { return this.cn.StandardErrorStream; }
        }

        /// <summary>
        /// Gets the Standard Input stream for this Session.
        /// </summary>
        public Stream StandardInput
        {
            get { return this.cn.StandardInputStream; }
        }

        /// <summary>
        /// Gets the int exit code/status from the remote command - if available. Be
        /// careful - not all server implementations return this value. It is
        /// generally a good idea to call this method only when all data from the
        /// remote side has been consumed (see also the <see cref="WaitForCondition"></see> method).
        /// Returns <c>0</c> if no exit code is (yet) available.
        /// </summary>
        public int ExitStatus
        {
            get { return this.cn.ExitStatus; }
        }

        /// <summary>
        /// Gets the name of the signal by which the process on the remote side was
        /// stopped - if available and applicable. Be careful - not all server
        /// implementations return this value.
        /// <para>
        /// If the process exited normally or the process is still
        /// running (or if the server forgot to send this information),
        /// this will return <c>null</c>.
        /// </para>
        /// </summary>
        public string ExitSignal
        {
            get { return this.cn.ExitSignal; }
        }

        /// <summary>
        /// Basically just a wrapper for lazy people - identical to calling
        /// <c>RequestPTY("dumb", 0, 0, 0, 0, null)</c>.
        /// </summary>
        /// <exception cref="IOException">If there is a problem 'behind the scenes'.</exception>
        public void RequestDumbPTY()
        {
            this.RequestPTY("dumb", 0, 0, 0, 0, null);
        }

        /// <summary>
        /// Basically just another wrapper for lazy people - identical to calling
        /// <c>RequestPTY(term, 0, 0, 0, 0, null)</c>.
        /// </summary>
        /// <param name="term">The terminal type, ("dumb" for example).</param>
        /// <exception cref="IOException">If there is a problem 'behind the scenes'.</exception>
        /// <exception cref="ArgumentException">If term is null.</exception>
        public void RequestPTY(string term)
        {
            this.RequestPTY(term, 0, 0, 0, 0, null);
        }

        /// <summary>
        /// Allocate a pseudo-terminal for this session.
        /// <para>
        /// This method may only be called before a program or shell is started in
        /// this session.
        /// </para>
        /// <para>
        /// <list type="bullet">
        /// <item>The TERM environment variable value (e.g., vt100)</item>
        /// <item>The terminal's dimensions.</item>
        /// <item>The encoded terminal modes.</item>
        /// </list>
        /// </para>
        /// <para>
        /// Zero dimension parameters are ignored. The character/row dimensions
        /// override the pixel dimensions (when nonzero). Pixel dimensions refer to
        /// the drawable area of the window. The dimension parameters are only
        /// informational. The encoding of terminal modes (parameter
        /// <c>terminalModes</c>) is described in RFC4254.
        /// </para>
        /// </summary>
        /// <param name="term">The TERM environment variable value (e.g., vt100).</param>
        /// <param name="termWidthCharacters">Terminal width, characters (e.g., 80).</param>
        /// <param name="termHeightCharacters">Terminal height, rows (e.g., 24).</param>
        /// <param name="termWidthPixels">Terminal width, pixels (e.g., 640).</param>
        /// <param name="termHeightPixels">Terminal height, pixels (e.g., 480).</param>
        /// <param name="terminalModes">Encoded terminalModes (may be <c>null</c>).</param>
        /// <exception cref="ArgumentException">If term is null.</exception>
        /// <exception cref="IOException">
        /// If the connection is closed or the terminalModes data is wrong.
        /// Also if we already requested a pty or we are already doing remote execution.
        /// </exception>
        public void RequestPTY(
            string term,
            int termWidthCharacters,
            int termHeightCharacters,
            int termWidthPixels,
            int termHeightPixels,
            byte[] terminalModes)
        {
            if (term == null)
            {
                throw new ArgumentException("TERM cannot be null.");
            }

            if ((terminalModes != null) && (terminalModes.Length > 0))
            {
                if (terminalModes[terminalModes.Length - 1] != 0)
                {
                    throw new IOException("Illegal terminal modes description, does not end in zero byte");
                }
            }
            else
            {
                terminalModes = new byte[] { 0 };
            }

            lock (this)
            {
                /* The following is just a nicer error, we would catch it anyway later in the channel code */
                if (this.flagClosed)
                {
                    throw new IOException("This session is closed.");
                }

                if (this.flagPtyRequested)
                {
                    throw new IOException("A PTY was already requested.");
                }

                if (this.flagExecutionStarted)
                {
                    throw new IOException(
                        "Cannot request PTY at this stage anymore, a remote execution has already started.");
                }
                
                this.flagPtyRequested = true;
            }

            this.cm.RequestPTY(
                this.cn,
                term,
                termWidthCharacters,
                termHeightCharacters,
                termWidthPixels,
                termHeightPixels,
                terminalModes);
        }

        /// <summary>
        /// Request X11 forwarding for the current session.
        /// <para>
        /// You have to supply the name and port of your X-server.
        /// </para>
        /// <para>
        /// This method may only be called before a program or shell is started in
        /// this session.
        /// </para>
        /// </summary>
        /// <param name="hostName">The hostName of the real (target) X11 server (e.g., 127.0.0.1).</param>
        /// <param name="port">You have to supply the name and port of your X-server.</param>
        /// <param name="cookie">If non-null, then present this cookie to the real X11 server.</param>
        /// <param name="singleConnection">
        /// If true, then the server is instructed to only forward one single
        /// connection, no more connections shall be forwarded after first, or after the session
        /// channel has been closed.
        /// </param>
        /// <exception cref="ArgumentException">If the host name is null.</exception>
        /// <exception cref="IOException">
        /// For a variety of issues: closed session, X11 forward already requested,
        /// command execution in progress already, etc.
        /// </exception>
        public void RequestX11Forwarding(string hostName, int port, byte[] cookie, bool singleConnection)
        {
            if (hostName == null)
            {
                throw new ArgumentException("hostName argument may not be null");
            }

            lock (this)
            {
                // The following is just a nicer error, we would catch it anyway later in the channel code
                if (this.flagClosed)
                {
                    throw new IOException("This session is closed.");
                }

                if (this.flagX11Requested)
                {
                    throw new IOException("X11 forwarding was already requested.");
                }

                if (this.flagExecutionStarted)
                {
                    throw new IOException(
                        "Cannot request X11 forwarding at this stage anymore, a remote execution has already started.");
                }
                
                this.flagX11Requested = true;
            }

            // X11ServerData - used to store data about the target X11 server
            X11ServerData x11data = new X11ServerData();

            x11data.HostName = hostName;
            x11data.Port = port;
            x11data.X11MagicCookie = cookie; // if non-null, then present this cookie to the real X11 server

            // Generate fake cookie - this one is used between remote clients and the Eunomia proxy
            byte[] fakeCookie = new byte[16];
            string hexEncodedFakeCookie;

            // Make sure that this fake cookie is unique for this connection
            while (true)
            {
                this.rnd.NextBytes(fakeCookie);

                // Generate also hex representation of fake cookie
                StringBuilder tmp = new StringBuilder(32);
                for (int i = 0; i < fakeCookie.Length; i++)
                {
                    string digit2 = (fakeCookie[i] & 0xff).ToString("X2", CultureInfo.InvariantCulture);
                    tmp.Append((digit2.Length == 2) ? digit2 : "0" + digit2);
                }

                hexEncodedFakeCookie = tmp.ToString();

                // Well, yes, chances are low, but we want to be on the safe side
                // If the length of the cookie is zero we did not find the same cookie.
                if (this.cm.CheckX11Cookie(hexEncodedFakeCookie) == null)
                {
                    break;
                }
            }

            // Ask for X11 forwarding
            this.cm.RequestX11(this.cn, singleConnection, "MIT-MAGIC-COOKIE-1", hexEncodedFakeCookie, 0);

            // OK, that went fine, get ready to accept X11 connections...
            // ... but only if the user has not called close() in the meantime =)
            lock (this)
            {
                if (this.flagClosed == false)
                {
                    this.x11FakeCookie = hexEncodedFakeCookie;
                    this.cm.RegisterX11Cookie(hexEncodedFakeCookie, x11data);
                }
            }

            // Now it is safe to start remote X11 programs
        }

        /// <summary>
        /// Execute a command on the remote machine.
        /// </summary>
        /// <param name="cmd">The command to execute on the remote host.</param>
        /// <exception cref="ArgumentException">If the cmd is null.</exception>
        /// <exception cref="IOException">For a variety of behind the scenes issues.</exception>
        public void ExecCommand(string cmd)
        {
            if (cmd == null)
            {
                throw new ArgumentException("cmd argument may not be null");
            }

            lock (this)
            {
                // The following is just a nicer error, we would catch it anyway later in the channel code
                if (this.flagClosed)
                {
                    throw new IOException("This session is closed.");
                }

                if (this.flagExecutionStarted)
                {
                    throw new IOException("A remote execution has already started.");
                }

                this.flagExecutionStarted = true;
            }

            this.cm.RequestExecCommand(this.cn, cmd);
        }

        /// <summary>
        /// Start a shell on the remote machine.
        /// </summary>
        /// <exception cref="IOException">For a variety of behind the scenes issues.</exception>
        public void StartShell()
        {
            lock (this)
            {
                /* The following is just a nicer error, we would catch it anyway later in the channel code */
                if (this.flagClosed)
                {
                    throw new IOException("This session is closed.");
                }

                if (this.flagExecutionStarted)
                {
                    throw new IOException("A remote execution has already started.");
                }

                this.flagExecutionStarted = true;
            }

            this.cm.RequestShell(this.cn);
        }

        /// <summary>
        /// Start a subsystem on the remote machine.
        /// Unless you know what you are doing, you will never need this.
        /// </summary>
        /// <param name="name">The name of the subsystem.</param>
        /// <exception cref="ArgumentException">If the name is null.</exception>
        /// <exception cref="IOException">For a variety of behind the scenes issues.</exception>
        public void StartSubsystem(string name)
        {
            if (name == null)
            {
                throw new ArgumentException("name argument may not be null");
            }

            lock (this)
            {
                // The following is just a nicer error, we would catch it anyway later in the channel code
                if (this.flagClosed)
                {
                    throw new IOException("This session is closed.");
                }

                if (this.flagExecutionStarted)
                {
                    throw new IOException("A remote execution has already started.");
                }

                this.flagExecutionStarted = true;
            }

            this.cm.RequestSubsystem(this.cn, name);
        }

        /// <summary>
        /// This method blocks until there is more data available on either the
        /// stdout or stderr InputStream of this <c>Session</c>. Very useful
        /// if you do not want to use two parallel threads for reading from the two
        /// InputStreams. One can also specify a timeout. NOTE: do NOT call this
        /// method if you use concurrent threads that operate on either of the two
        /// InputStreams of this <c>Session</c> (otherwise this method may
        /// block, even though more data is available).
        /// </summary>
        /// <param name="timeout">
        /// The (non-negative) timeout in <c>ms</c>. <c>0</c> means no
        /// timeout, the call may block forever.
        /// </param>
        /// <returns>
        /// <list type="bullet">
        /// <item><c>0</c> if no more data will arrive.</item>
        /// <item><c>1</c> if more data is available.</item>
        /// <item><c>-1</c> if a timeout occurred.</item>
        /// </list>
        /// </returns>
        /// <seealso cref="WaitForCondition"></seealso>
        /// <exception cref="ArgumentException">If the timeout is negative.</exception>
        /// <exception cref="IOException">For a variety of behind the scenes issues.</exception>
        /// <exception cref="InvalidOperationException">If the condition result is unexpected.</exception>
        [ObsoleteAttribute("Please use the WaitForCondition method instead of this.")]
        public int WaitUntilDataAvailable(long timeout)
        {
            if (timeout < 0)
            {
                throw new ArgumentException("timeout must not be negative!");
            }

            int conditions = ChannelManager.WaitForCondition(
                this.cn,
                timeout,
                ChannelCondition.StandardOutData | ChannelCondition.StandardErrorData | ChannelCondition.EOF);

            if ((conditions & ChannelCondition.Timeout) != 0)
            {
                return -1;
            }

            if ((conditions & (ChannelCondition.StandardOutData | ChannelCondition.StandardErrorData)) != 0)
            {
                return 1;
            }

            // Here we do not need to check separately for CLOSED, since CLOSED implies EOF
            if ((conditions & ChannelCondition.EOF) != 0)
            {
                return 0;
            }

            throw new InvalidOperationException("Unexpected condition result (" + conditions + ")");
        }

        /// <summary>
        /// This method blocks until certain conditions hold true on the underlying SSH-2 channel.
        /// <para>
        /// This method returns as soon as one of the following happens:
        /// <list type="bullet">
        /// <item>One of the specified conditions <see cref="ChannelCondition"></see> holds true.</item>
        /// <item>Timeout > 0 and a timeout occured (TIMEOUT will be set in result conditions).</item>
        /// <item>The underlying channel was closed (CLOSED will be set in result conditions).</item>
        /// </list>
        /// </para>
        /// <para>
        /// In any case, the result value contains ALL current conditions, which may be more
        /// than the specified condition set (i.e., never use the "==" operator to test for conditions
        /// in the bitmask, see also comments in <see cref="ChannelCondition"></see>.
        /// </para>
        /// <para>
        /// Note: do NOT call this method if you want to wait for STDOUT_DATA or STDERR_DATA and
        /// there are concurrent threads (e.g., StreamGobblers) that operate on either of the two
        /// InputStreams of this <c>Session</c> (otherwise this method may
        /// block, even though more data is available in the StreamGobblers).
        /// </para>
        /// </summary>
        /// <param name="conditionSet">
        /// A bitmask based on <see cref="ChannelCondition"></see> values.
        /// </param>
        /// <param name="timeout">
        /// Non-negative timeout in ms, <c>0</c> means no timeout.
        /// </param>
        /// <returns>
        /// Bitmask specifying all current conditions that are true.
        /// </returns>
        /// <exception cref="ArgumentException">If the timeout is negative.</exception>
        /// <exception cref="IOException">For a variety of behind the scenes issues.</exception>
        public int WaitForCondition(int conditionSet, long timeout)
        {
            if (timeout < 0)
            {
                throw new ArgumentException("timeout must be non-negative!");
            }

            return ChannelManager.WaitForCondition(this.cn, timeout, conditionSet);
        }

        /// <summary>
        /// Close this session. NEVER forget to call this method to free up resources -
        /// even if you got an exception from one of the other methods (or when
        /// getting an Exception on the Input- or OutputStreams). Sometimes these other
        /// methods may throw an exception, saying that the underlying channel is
        /// closed (this can happen, e.g., if the other server sent a close message.)
        /// However, as long as you have not called the <c>Close()</c>
        /// method, you may be wasting (local) resources.
        /// </summary>
        public void Close()
        {
            lock (this)
            {
                if (this.flagClosed)
                {
                    return;
                }

                this.flagClosed = true;

                if (this.x11FakeCookie != null)
                {
                    this.cm.UnregisterX11Cookie(this.x11FakeCookie, true);
                }

                try
                {
                    this.cm.CloseChannel(this.cn, "Closed due to user request", true);
                }
                catch (IOException)
                {
                }
            }
        }
    }
}
