﻿// <copyright file="OpenFlags.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> OpenFlags
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;

namespace Eunomia.SSH2.SFTP
{
    /// <summary>
    ///  SFTP Open Flags.
    /// </summary>
    /// <remarks>
    /// The following table is provided to assist in mapping POSIX semantics
    /// to equivalent SFTP file open parameters:
    /// O_RDONLY            desired-access = READ_DATA | READ_ATTRIBUTES
    /// O_WRONLY            desired-access = WRITE_DATA | WRITE_ATTRIBUTES
    /// O_RDWR              desired-access = READ_DATA | READ_ATTRIBUTES | WRITE_DATA | WRITE_ATTRIBUTES
    /// O_APPEND            desired-access = WRITE_DATA | WRITE_ATTRIBUTES | APPEND_DATA
    ///                     flags = SSH_FXF_ACCESS_APPEND_DATA and or SSH_FXF_ACCESS_APPEND_DATA_ATOMIC
    /// O_CREAT             flags = SSH_FXF_OPEN_OR_CREATE
    /// O_TRUNC             flags = SSH_FXF_TRUNCATE_EXISTING
    /// O_TRUNC|O_CREATE    flags = SSH_FXF_CREATE_TRUNCATE
    /// </remarks>
    [SuppressMessage("Microsoft.Performance", "CA1815:OverrideEqualsAndOperatorEqualsOnValueTypes",
                    Justification = "We will never compare or sort instances of this struct.")]
    [SuppressMessage("Microsoft.Naming", "CA1726:UsePreferredTerms", MessageId = "Flags",
                    Justification = "To retain name compatibility with ganymed.")]
    public struct OpenFlags
    {
        /// <summary>
        /// Disposition is a 3 bit field that controls how the file is opened.
        /// The server MUST support these bits (possible enumaration values:
        /// SSH_FXF_CREATE_NEW, SSH_FXF_CREATE_TRUNCATE, SSH_FXF_OPEN_EXISTING,
        /// SSH_FXF_OPEN_OR_CREATE or SSH_FXF_TRUNCATE_EXISTING).
        /// </summary>
        public const int SSHFileAccessDisposition = 0x00000007;

        /// <summary>
        /// A new file is created; if the file already exists, the server
        /// MUST return status SSH_FXP_FILE_ALREADY_EXISTS.
        /// </summary>
        public const int SSHFileAccessCreateNew = 0x00000000;

        /// <summary>
        /// A new file is created; if the file already exists, it is opened
        /// and truncated.
        /// </summary>
        public const int SSHFileAccessTruncate = 0x00000001;

        /// <summary>
        /// An existing file is opened.  If the file does not exist, the
        /// server MUST return SSH_FXP_NO_SUCH_FILE. If a directory in the
        /// path does not exist, the server SHOULD return
        /// SSH_FXP_NO_SUCH_PATH. It is also acceptable if the server
        /// returns SSH_FXP_NO_SUCH_FILE in this case.
        /// </summary>
        public const int SSHFileAccessOpenExisting = 0x00000002;

        /// <summary>
        /// If the file exists, it is opened. If the file does not exist,
        /// it is created.
        /// </summary>
        public const int SSHFileAccessOpenOrCreate = 0x00000003;

        /// <summary>
        /// An existing file is opened and truncated. If the file does not
        /// exist, the server MUST return the same error codes as defined
        /// for SSH_FXF_OPEN_EXISTING.
        /// </summary>
        public const int SSHFileAccessTruncateExisting = 0x00000004;

        /// <summary>
        /// Data is always written at the end of the file. The offset field
        /// of the SSH_FXP_WRITE requests are ignored.
        /// <para>
        /// Data is not required to be appended atomically. This means that
        /// if multiple writers attempt to append data simultaneously, data
        /// from the first may be lost. However, data MAY be appended
        /// atomically.
        /// </para>
        /// </summary>
        public const int SSHFileAccessAppendData = 0x00000008;

        /// <summary>
        /// Data is always written at the end of the file. The offset field
        /// of the SSH_FXP_WRITE requests are ignored.
        /// <para>
        /// Data MUST be written atomically so that there is no chance that
        /// multiple appenders can collide and result in data being lost.
        /// </para>
        /// <para>
        /// If both append flags are specified, the server SHOULD use atomic
        /// append if it is available, but SHOULD use non-atomic appends
        /// otherwise. The server SHOULD NOT fail the request in this case.
        /// </para>
        /// </summary>
        public const int SSHFileAccessAppendDataAtomic = 0x00000010;

        /// <summary>
        /// Indicates that the server should treat the file as text and
        /// convert it to the canonical newline convention in use.
        /// (See Determining Server Newline Convention in section 5.3 in the
        /// SFTP standard draft).
        /// <para>
        /// When a file is opened with this flag, the offset field in the read
        /// and write functions is ignored.
        /// </para>
        /// <para>
        /// Servers MUST process multiple, parallel reads and writes correctly
        /// in this mode.  Naturally, it is permissible for them to do this by
        /// serializing the requests.
        /// </para>
        /// <para>
        /// Clients SHOULD use the SSH_FXF_ACCESS_APPEND_DATA flag to append
        /// data to a text file rather then using write with a calculated offset.
        /// </para>
        /// </summary>
        public const int SSHFileAccessTextMode = 0x00000020;

        /// <summary>
        /// The server MUST guarantee that no other handle has been opened
        /// with ACE4_READ_DATA access, and that no other handle will be
        /// opened with ACE4_READ_DATA access until the client closes the
        /// handle. (This MUST apply both to other clients and to other
        /// processes on the server.)
        /// <para>
        /// If there is a conflicting lock the server MUST return
        /// SSH_FXP_LOCK_CONFLICT.  If the server cannot make the locking
        /// guarantee, it MUST return SSH_FXP_OP_UNSUPPORTED.
        /// </para>
        /// <para>
        /// Other handles MAY be opened for ACE4_WRITE_DATA or any other
        /// combination of accesses, as long as ACE4_READ_DATA is not included
        /// in the mask.
        /// </para>
        /// </summary>
        public const int SSHFileAccessBlockRead = 0x00000040;

        /// <summary>
        /// The server MUST guarantee that no other handle has been opened
        /// with ACE4_WRITE_DATA or ACE4_APPEND_DATA access, and that no other
        /// handle will be opened with ACE4_WRITE_DATA or ACE4_APPEND_DATA
        /// access until the client closes the handle. (This MUST apply both
        /// to other clients and to other processes on the server.)
        /// <para>
        /// If there is a conflicting lock the server MUST return
        /// SSH_FXP_LOCK_CONFLICT. If the server cannot make the locking
        /// guarantee, it MUST return SSH_FXP_OP_UNSUPPORTED.
        /// </para>
        /// <para>
        /// Other handles MAY be opened for ACE4_READ_DATA or any other
        /// combination of accesses, as long as neither ACE4_WRITE_DATA nor
        /// ACE4_APPEND_DATA are included in the mask.
        /// </para>
        /// </summary>
        public const int SSHFileAccessBlockWrite = 0x00000080;

        /// <summary>
        /// The server MUST guarantee that no other handle has been opened
        /// with ACE4_DELETE access, opened with the
        /// SSH_FXF_ACCESS_DELETE_ON_CLOSE flag set, and that no other handle
        /// will be opened with ACE4_DELETE access or with the
        /// SSH_FXF_ACCESS_DELETE_ON_CLOSE flag set, and that the file itself
        /// is not deleted in any other way until the client closes the handle.
        /// <para>
        /// If there is a conflicting lock the server MUST return
        /// SSH_FXP_LOCK_CONFLICT.  If the server cannot make the locking
        /// guarantee, it MUST return SSH_FXP_OP_UNSUPPORTED.
        /// </para>
        /// </summary>
        public const int SSHFileAccessBlockDelete = 0x00000100;

        /// <summary>
        /// If this bit is set, the above BLOCK modes are advisory. In advisory
        /// mode, only other accesses that specify a BLOCK mode need be
        /// considered when determining whether the BLOCK can be granted,
        /// and the server need not prevent I/O operations that violate the
        /// block mode.
        /// <para>
        /// The server MAY perform mandatory locking even if the BLOCK_ADVISORY
        /// bit is set.
        /// </para>
        /// </summary>
        public const int SSHFileAccessBlockAdvisory = 0x00000200;

        /// <summary>
        /// If the final component of the path is a symlink, then the open
        /// MUST fail, and the error SSH_FXP_LINK_LOOP MUST be returned.
        /// </summary>
        public const int SSHFileAccessNoFollow = 0x00000400;

        /// <summary>
        /// The file should be deleted when the last handle to it is closed.
        /// (The last handle may not be an sftp-handle.)  This MAY be emulated
        /// by a server if the OS doesn't support it by deleting the file when
        /// this handle is closed.
        /// <para>
        /// It is implementation specific whether the directory entry is
        /// removed immediately or when the handle is closed.
        /// </para>
        /// </summary>
        public const int SSHFileAccessDeleteOnClose = 0x00000800;
    }
}
