﻿// <copyright file="DSASHA1Verify.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> DSASHA1Verify
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Diagnostics.CodeAnalysis;
using System.IO;

using Eunomia.SSH2.Crypto.Digest;
using Eunomia.SSH2.Log;
using Eunomia.SSH2.Packets;
using Org.BouncyCastle.Math;
using Org.BouncyCastle.Security;

namespace Eunomia.SSH2.Signature
{
    /// <summary>
    /// Class full of static methods to decode and encode public keys and
    /// signatures, and verify and generate signatures.
    /// </summary>
    public static class DSASHA1Verify
    {
        private static Logger log = Logger.GetLogger(typeof(DSASHA1Verify));

        /// <summary>
        /// Decode a DSA public key.
        /// </summary>
        /// <param name="key">The key to decode.</param>
        /// <returns>The DSAPublicKey.</returns>
        public static DSAPublicKey DecodeSSHDSAPublicKey(byte[] key)
        {
            TypesReader tr = new TypesReader(key);

            string key_format = tr.ReadString();

            if (key_format.Equals("ssh-dss") == false)
            {
                throw new ArgumentException("This is not a ssh-dss public key!");
            }

            BigInteger p = tr.ReadMPINT();
            BigInteger q = tr.ReadMPINT();
            BigInteger g = tr.ReadMPINT();
            BigInteger y = tr.ReadMPINT();

            if (tr.Remain != 0)
            {
                throw new IOException("Padding in DSA public key!");
            }

            return new DSAPublicKey(p, q, g, y);
        }

        /// <summary>
        /// Encode a DSA public key.
        /// </summary>
        /// <param name="publicKey">The DSAPublicKey to encode.</param>
        /// <returns>The encoded key as a byte array.</returns>
        public static byte[] EncodeSSHDSAPublicKey(DSAPublicKey publicKey)
        {
            TypesWriter tw = new TypesWriter();

            tw.WriteString("ssh-dss");
            tw.WriteMPInt(publicKey.P);
            tw.WriteMPInt(publicKey.Q);
            tw.WriteMPInt(publicKey.G);
            tw.WriteMPInt(publicKey.Y);

            return tw.GetBytes();
        }

        /// <summary>
        /// Encode a DSA signature.
        /// </summary>
        /// <param name="ds">The DSASignature to encode.</param>
        /// <returns>The encoded signature as a byte array.</returns>
        public static byte[] EncodeSSHDSASignature(DSASignature ds)
        {
            TypesWriter tw = new TypesWriter();

            tw.WriteString("ssh-dss");

            byte[] r = ds.R.ToByteArray();
            byte[] s = ds.S.ToByteArray();

            byte[] a40 = new byte[40];

            // Patch (unsigned) r and s into the target array.
            int rcopylen = (r.Length < 20) ? r.Length : 20;
            int scopylen = (s.Length < 20) ? s.Length : 20;

            Array.Copy(r, r.Length - rcopylen, a40, 20 - rcopylen, rcopylen);
            Array.Copy(s, s.Length - scopylen, a40, 40 - scopylen, scopylen);

            tw.WriteString(a40, 0, 40);

            return tw.GetBytes();
        }

        /// <summary>
        /// Decode a DSA signature.
        /// </summary>
        /// <param name="signature">The signature to decoode.</param>
        /// <returns>The DSA signature.</returns>
        public static DSASignature DecodeSSHDSASignature(byte[] signature)
        {
            byte[] rsarray;

            if (signature.Length == 40)
            {
                rsarray = signature;
            }
            else
            {
                TypesReader tr = new TypesReader(signature);

                string sig_format = tr.ReadString();

                if (sig_format.Equals("ssh-dss") == false)
                {
                    throw new IOException("Peer sent wrong signature format");
                }

                rsarray = tr.ReadByteString();

                if (rsarray.Length != 40)
                {
                    throw new IOException("Peer sent corrupt signature");
                }

                if (tr.Remain != 0)
                {
                    throw new IOException("Padding in DSA signature!");
                }
            }

            // Remember, s and r are unsigned ints.
            byte[] tmp = new byte[20];

            Array.Copy(rsarray, 0, tmp, 0, 20);
            BigInteger r = new BigInteger(1, tmp);

            Array.Copy(rsarray, 20, tmp, 0, 20);
            BigInteger s = new BigInteger(1, tmp);

            if (log.Enabled)
            {
                log.Log(
                    30,
                    "decoded ssh-dss signature: first bytes r(" + (rsarray[0] & 0xff) + "), s(" + (rsarray[20] & 0xff) + ")");
            }

            return new DSASignature(r, s);
        }

        /// <summary>
        /// Verify the DSASignature.
        /// </summary>
        /// <param name="message">The message as a byte array.</param>
        /// <param name="ds">The signature.</param>
        /// <param name="publicKey">The public key.</param>
        /// <returns>True if the message verifies.</returns>
        public static bool VerifySignature(byte[] message, DSASignature ds, DSAPublicKey publicKey)
        {
            // Inspired by Bouncycastle's DSASigner class
            SHA1 md = new SHA1();
            md.Update(message);
            byte[] sha_message = new byte[md.DigestLength];
            md.Digest(sha_message);

            BigInteger m = new BigInteger(1, sha_message);

            BigInteger r = ds.R;
            BigInteger s = ds.S;

            BigInteger g = publicKey.G;
            BigInteger p = publicKey.P;
            BigInteger q = publicKey.Q;
            BigInteger y = publicKey.Y;

            BigInteger zero = BigInteger.Zero;

            if (log.Enabled)
            {
                log.Log(60, "ssh-dss signature: m: " + m.ToString(16));
                log.Log(60, "ssh-dss signature: r: " + r.ToString(16));
                log.Log(60, "ssh-dss signature: s: " + s.ToString(16));
                log.Log(60, "ssh-dss signature: g: " + g.ToString(16));
                log.Log(60, "ssh-dss signature: p: " + p.ToString(16));
                log.Log(60, "ssh-dss signature: q: " + q.ToString(16));
                log.Log(60, "ssh-dss signature: y: " + y.ToString(16));
            }

            if (zero.CompareTo(r) >= 0 || q.CompareTo(r) <= 0)
            {
                log.Log(20, "ssh-dss signature: zero.compareTo(r) >= 0 || q.compareTo(r) <= 0");
                return false;
            }

            if (zero.CompareTo(s) >= 0 || q.CompareTo(s) <= 0)
            {
                log.Log(20, "ssh-dss signature: zero.compareTo(s) >= 0 || q.compareTo(s) <= 0");
                return false;
            }

            BigInteger w = s.ModInverse(q);

            BigInteger u1 = m.Multiply(w).Mod(q);
            BigInteger u2 = r.Multiply(w).Mod(q);

            u1 = g.ModPow(u1, p);
            u2 = y.ModPow(u2, p);

            BigInteger v = u1.Multiply(u2).Mod(p).Mod(q);

            return v.Equals(r);
        }

        /// <summary>
        /// Generate a signature for a message and private key.
        /// </summary>
        /// <param name="message">A byte array containing the message.</param>
        /// <param name="publicKey">The private key.</param>
        /// <param name="random">An instance of the SecureRandom class.</param>
        /// <returns>The DSA signature.</returns>
        [SuppressMessage("Microsoft.Design", "CA1011:ConsiderPassingBaseTypesAsParameters", Justification = "SecureRandom is required.")]
        public static DSASignature GenerateSignature(byte[] message, DSAPrivateKey publicKey, SecureRandom random)
        {
            SHA1 md = new SHA1();
            md.Update(message);
            byte[] sha_message = new byte[md.DigestLength];
            md.Digest(sha_message);

            BigInteger m = new BigInteger(1, sha_message);
            BigInteger k;
            int qbitLength = publicKey.Q.BitLength;

            do
            {
                k = new BigInteger(qbitLength, random);
            }
            while (k.CompareTo(publicKey.Q) >= 0);

            BigInteger r = publicKey.G.ModPow(k, publicKey.P).Mod(publicKey.Q);

            k = k.ModInverse(publicKey.Q).Multiply(m.Add(publicKey.X.Multiply(r)));

            BigInteger s = k.Mod(publicKey.Q);

            return new DSASignature(r, s);
        }
    }
}
