// <copyright file="ClientServerHello.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> ClientServerHello
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.IO;
using Eunomia.SSH2.Util;

namespace Eunomia.SSH2.Transport
{
    /// <summary>
    /// Class used for data and methods revolving around the
    /// client, server 'hello' exchange.
    /// </summary>
    public class ClientServerHello
    {
        private string server_line;
        private string client_line;
        
        private string server_versioncomment;

        /// <summary>
        /// Initializes a new instance of the ClientServerHello class.
        /// </summary>
        /// <param name="inStream">The input stream.</param>
        /// <param name="outStream">The output stream.</param>
        public ClientServerHello(Stream inStream, Stream outStream)
        {
            this.client_line = "SSH-2.0-" + Connection.Identification;

            byte[] outbytes = StringEncoder.GetBytes(this.client_line + "\r\n");

            outStream.Write(outbytes, 0, outbytes.Length);
            outStream.Flush();

            byte[] serverVersion = new byte[512];

            for (int i = 0; i < 50; i++)
            {
                int len = ReadLineRN(inStream, serverVersion);

                this.server_line = StringEncoder.GetString(serverVersion, 0, len);

                if (this.server_line.StartsWith("SSH-", StringComparison.Ordinal))
                {
                    break;
                }
            }

            if (this.server_line.StartsWith("SSH-", StringComparison.Ordinal) == false)
            { 
                throw new IOException(
                    "Malformed server identification string. There was no line starting with 'SSH-' amongst the first 50 lines.");
            }
            
            if (this.server_line.StartsWith("SSH-1.99-", StringComparison.Ordinal))
            {
                this.server_versioncomment = this.server_line.Substring(9);
            }
            else if (this.server_line.StartsWith("SSH-2.0-", StringComparison.Ordinal))
            {
                this.server_versioncomment = this.server_line.Substring(8);
            }
            else
            {
                throw new IOException("Server uses incompatible protocol, it is not SSH-2 compatible.");
            }
        }

        /// <summary>
        /// Gets the server version comment.
        /// </summary>
        public string ServerVersionComment
        {
            get { return this.server_versioncomment; }
        }
        
        /// <summary>
        /// Reads the server hello information from the input stream.
        /// </summary>
        /// <param name="inStream">The input stream to read from.</param>
        /// <param name="buffer">The buffer we store the server response in.</param>
        /// <returns>The number of bytes read from the stream.</returns>
        public static int ReadLineRN(Stream inStream, byte[] buffer)
        {
            int pos = 0;
            bool need10 = false;
            int len = 0;
            while (true)
            {
                int c = inStream.ReadByte();
                if (c == -1)
                {
                    throw new IOException("Premature connection close");
                }

                buffer[pos++] = (byte)c;

                if (c == 13)
                {
                    need10 = true;
                    continue;
                }

                if (c == 10)
                {
                    break;
                }

                if (need10 == true)
                {
                    throw new IOException("Malformed line sent by the server, the line does not end correctly.");
                }

                len++;
                if (pos >= buffer.Length)
                {
                    throw new IOException("The server sent a too long line.");
                }
            }

            return len;
        }

        /// <summary>
        /// Gets the client_version.
        /// </summary>
        /// <returns>The client version as a byte array.</returns>
        public byte[] GetClientString()
        {
            return StringEncoder.GetBytes(this.client_line);
        }

        /// <summary>
        /// Gets the server_version.
        /// </summary>
        /// <returns>The server version as a byte array.</returns>
        public byte[] GetServerString()
        {
            return StringEncoder.GetBytes(this.server_line);
        }
    }
}
