﻿// <copyright file="EunomiaBasicTests.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> EunomiaBasicTests
// </summary>
// <author>  David Lawler
// </author>
// <version> 0.5.0, 12/16/2010
// </version>

namespace Eunomia.Tests
{
    using System;
    using System.Collections;
    using System.Configuration;
    using System.Globalization;
    using System.IO;
    using System.Threading;
    
    using Eunomia.SSH2;
    using NUnit.Framework;

    /// <summary>
    /// Description of EunomiaBasicTests.
    /// </summary>
    [TestFixture]
    public static class EunomiaBasicTests
    {
        private static string hostName = ConfigurationManager.AppSettings["hostName"];
        private static string hpstr = ConfigurationManager.AppSettings["hostPort"];
        private static int hostPort = int.Parse(hpstr, NumberStyles.Integer, CultureInfo.InvariantCulture);
        private static string userName = ConfigurationManager.AppSettings["userName"];
        private static string password = ConfigurationManager.AppSettings["password"];
        private static string testFile = "EunomiaTestFile.bin";

        /// <summary>
        /// Test SSH...just logons on and see if it can see a response.
        /// </summary>
        [Test]
        public static void TestSSH()
        {
            Connection conn = new Connection(EunomiaBasicTests.hostName, EunomiaBasicTests.hostPort);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server");
            }

            Session sess = conn.OpenSession();
            sess.ExecCommand("uname -a && date");
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }
            
            if (linecnt == 0)
            {
                Assert.Fail("Nothing received from the ssh server for cammand 'uname'");
            }

            // Close this session
            sess.Close();

            // Close the connection
            conn.Close();
        }
        
        /// <summary>
        /// Tests some SFTP things:
        /// <list type="bullet">
        /// <item>Does a ls on the home directory.</item>
        /// <item>Does a tranfer of a file to the server.</item>
        /// <item>Does a tranfer of the same file from the server.</item>
        /// </list>
        /// Then it compares the original file to the file we get back from the server.
        /// If there are any differences we have an error.
        /// </summary>
        [Test]
        public static void TestSFTP()
        {
            EunomiaBasicTests.InitFileTransfers();
            
            // Set up the connection.
            Connection conn = new Connection(EunomiaBasicTests.hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server");
            }
            
            SFTPv3Client client = new SFTPv3Client(conn);
            ArrayList files = client.Ls(".");
            Console.WriteLine(
                "Command ls returned " + files.Count.ToString(CultureInfo.InvariantCulture) + " entries.");
            SFTPv3DirectoryEntry de = (SFTPv3DirectoryEntry)files[files.Count - 1];
            Console.WriteLine("Last file is: " + de.FileName);
            string fname = EunomiaBasicTests.testFile;
            
            // Read our test file and transfer it to the ssh server.
            Console.WriteLine("Sending file to server with sftp.");
            SFTPv3FileHandle fh = client.CreateFileTruncate(fname);
            FileStream localFile = new FileStream(fname, FileMode.Open);
            byte[] buffer = new byte[32768];
            int offset = 0;
            int count = 0;
            while ((count = localFile.Read(buffer, 0, buffer.Length)) > 0)
            {
                client.Write(fh, offset, buffer, 0, count);
                Console.Write(".");
                offset += count;
            }
            
            localFile.Close();
            Console.WriteLine(string.Empty);
            
            // Read the file back from the ssh server and put it in our current working directory.
            Console.WriteLine("Receiving file from server with sftp.");
            fh = client.OpenFileRO(fname);
            localFile = new FileStream(fname, FileMode.OpenOrCreate);
            offset = 0;
            count = 0;
            while (!fh.Closed)
            {
                count = client.Read(fh, offset, buffer, 0, 32768);
                if (count < 0)
                {
                    break;
                }

                localFile.Write(buffer, 0, count);
                offset = offset + count;
            }
            
            localFile.Close();

            // Close the client and the connection.
            client.Close();
            conn.Close();
            FileAssert.AreEqual("org" + fname, fname, "File received via sftp is not the same!");
        }

        /// <summary>
        /// Tests some SCP things:
        /// <list type="bullet">
        /// <item>Does a tranfer (put) of a file to the server.</item>
        /// <item>Does a tranfer (get) of the same file from the server.</item>
        /// </list>
        /// Then it compares the original file to the file we get back from the server.
        /// If there are any differences we have an error.
        /// </summary>
        [Test]
        public static void TestSCP()
        {
            EunomiaBasicTests.InitFileTransfers();

            // Set up the connection.
            Connection conn = new Connection(EunomiaBasicTests.hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server");
            }

            string fname = EunomiaBasicTests.testFile;
            string cwd = Directory.GetCurrentDirectory();

            Console.WriteLine("Testing scp.");
            SCPClient client = new SCPClient(conn);
            Console.WriteLine("Doing the put.");
            client.Put(fname, ".");
            Console.WriteLine("Doing the get.");
            client.Get(fname, cwd);
            conn.Close();
            FileAssert.AreEqual("org" + fname, fname, "File received via scp is not the same!");
        }
        
        /// <summary>
        /// This will test if a local forward is working.
        /// </summary>
        [Test]
        public static void TestLocalForward()
        {
            Connection conn = new Connection(EunomiaBasicTests.hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server");
            }
            
            conn.CreateLocalPortForwarder(2200, EunomiaBasicTests.hostName, 22);

            // Now see if we can use this connection/forward to get a login prompt
            Connection conn2 = new Connection("localhost", 2200);
            conn2.Connect();
            isAuth = conn2.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server through forwarded port");
            }

            Session sess = conn2.OpenSession();
            sess.ExecCommand("uname -a && date");
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }
            
            if (linecnt == 0)
            {
                Assert.Fail("Nothing received from the ssh server for cammand 'uname'");
            }

            // Close this session
            sess.Close();

            // Close the connections
            conn2.Close();
            conn.Close();
        }

        /// <summary>
        /// This will test if a remote forward is working.
        /// </summary>
        [Test]
        public static void TestRemoteForward()
        {
            Connection conn = new Connection(EunomiaBasicTests.hostName);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server");
            }

            // 1st two are relative to the remote server...what ip and port to bind to.
            // 2nd two are relative to the client (us)...what ipaddress and port to forward to.
            // This particular forward takes network traffic to localhost on the server on port 2222
            // and bounces it from us BACK to the normal ssh port for the server.
            conn.RequestRemotePortForwarding(hostName, 2222, hostName, 22);

            // Now see if we can use this connection/forward to get a login prompt
            // For this to work the remote server needs to have "GatewayPorts yes" in
            // the sshd_config file
            Connection conn2 = new Connection(hostName, 2222);
            conn2.Connect();
            isAuth = conn2.AuthenticateWithPassword(
                EunomiaBasicTests.userName,
                EunomiaBasicTests.password);
            if (isAuth == false)
            {
                Assert.Fail("Could not login to ssh server through forwarded port");
            }

            Session sess = conn2.OpenSession();
            sess.ExecCommand("uname -a && date");
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }
            
            if (linecnt == 0)
            {
                Assert.Fail("Nothing received from the ssh server for cammand 'uname'");
            }

            // Close this session
            sess.Close();

            // Close the connections
            conn2.Close();
            conn.Close();
        }
        
        /// <summary>
        /// Method called by SFTP and SCP tests.  Creates a random 4MB file to
        /// test file transfers to/from an ssh server.
        /// </summary>
        private static void InitFileTransfers()
        {
            if (File.Exists(EunomiaBasicTests.testFile))
            {
                File.Delete(EunomiaBasicTests.testFile);
            }
            
            if (File.Exists("org" + EunomiaBasicTests.testFile))
            {
                File.Delete("org" + EunomiaBasicTests.testFile);
            }
            
            // Create a test file (1MB of random data)
            System.Random rnd = new System.Random();
            FileStream localFile = new FileStream(EunomiaBasicTests.testFile, FileMode.Create, FileAccess.Write);
            byte[] buffer = new byte[1024];
            for (int i = 0; i < 1024; i++)
            {
                for (int j = 0; j < 1; j++)
                {
                    rnd.NextBytes(buffer);
                    localFile.Write(buffer, 0, 1024);
                }
            }
            
            localFile.Close();
            
            // save a copy of the original file.
            File.Copy(EunomiaBasicTests.testFile, "org" + EunomiaBasicTests.testFile);
        }
    }
}