﻿// <copyright file="Basic.cs" company="David Lawler">
//     David Lawler, Copyright (c) 2010 David Lawler. All rights reserved.
// </copyright>
// <summary> Basic
// </summary>
// <author>  David Lawler, translated from Christian Plattner's ganymed 2.51 beta 1
// </author>
// <version> 0.6.0, 12/24/2010
// </version>

using System;
using System.Collections;
using System.Globalization;
using System.IO;

namespace Eunomia.SSH2.Examples
{
    /// <summary>
    /// Program class...just a basic console appliction.
    /// </summary>
    public static class Program
    {
        private static string hostName;
        private static int hostPort;
        private static string userName;
        private static string password;

        /// <summary>
        /// The one and only static Main method.
        /// </summary>
        public static void Main()
        {
            Console.Write("Enter host name: ");
            hostName = Console.ReadLine();

            Console.Write("Enter host port: ");
            string hpstr = Console.ReadLine();
            if (string.IsNullOrEmpty(hpstr))
            {
                hpstr = "22";
            }
            
            hostPort = int.Parse(hpstr, NumberStyles.Integer, CultureInfo.InvariantCulture);
            
            Console.Write("Enter username: ");
            userName = Console.ReadLine();
            
            Console.Write("Enter password: ");
            password = Console.ReadLine();

            TrySSH();
            TrySFTP("BasicTest.bin");
            TrySCP("BasicTest.bin");
            TryLocalForward();
            TryRemoteForward();
            TryX11Forward();

            Console.Write("Hit a key to close this program: ");
            Console.ReadKey();
        }
        
        /// <summary>
        /// Try using ssh.
        /// </summary>
        public static void TrySSH()
        {
            Console.WriteLine("===============> Now testing ssh <====================");
            Connection conn = new Connection(hostName, hostPort);
            conn.Connect();
            bool isAuthenticated = conn.AuthenticateWithPassword(userName, password);
            if (isAuthenticated == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }

            Session sess = conn.OpenSession();
            sess.ExecCommand("uname -a && date");
            Console.WriteLine("Here is some information about the remote host:");

            // This basic example does not handle stderr, which is sometimes dangerous
            // (please read the FAQ).
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            while (true)
            {
                string line = br.ReadLine();
                if (line == null)
                {
                    break;
                }

                Console.WriteLine(line);
            }

            // Show exit status, if available (otherwise "null")
            Console.WriteLine("ExitCode: " + sess.ExitStatus);

            // Close this session
            sess.Close();

            // Close the connection
            conn.Close();
        }

        /// <summary>
        /// Try SFTP.  We test ls, and sending and receiving a file.
        /// </summary>
        /// <param name="fileName">The file name to use for the test file.</param>
        public static void TrySFTP(string fileName)
        {
            InitFileTransfers(fileName);
            Console.WriteLine("===============> Now testing sftp <====================");
            Connection conn = new Connection(hostName, hostPort);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }

            SFTPv3Client client = new SFTPv3Client(conn);
            ArrayList files = client.Ls(".");
            Console.WriteLine(files.Count.ToString(CultureInfo.InvariantCulture));
            SFTPv3DirectoryEntry de = (SFTPv3DirectoryEntry)files[files.Count - 1];
            Console.WriteLine("Last file name: " + de.FileName);

            // Read our test file and transfer it to the ssh server.
            Console.WriteLine("Sending file to server with sftp.");
            SFTPv3FileHandle fh = client.CreateFileTruncate(fileName);
            FileStream localFile = new FileStream(fileName, FileMode.Open);
            byte[] buffer = new byte[32768];
            int offset = 0;
            int count = 0;
            while ((count = localFile.Read(buffer, 0, buffer.Length)) > 0)
            {
                client.Write(fh, offset, buffer, 0, count);
                Console.Write(".");
                offset += count;
            }
            
            localFile.Close();
            Console.WriteLine(string.Empty);
            
            // Read the file back from the ssh server and put it in our current working directory.
            Console.WriteLine("Receiving file from server with sftp.");
            fh = client.OpenFileRO(fileName);
            localFile = new FileStream(fileName, FileMode.OpenOrCreate);
            offset = 0;
            count = 0;
            while (!fh.Closed)
            {
                count = client.Read(fh, offset, buffer, 0, 32768);
                if (count < 0)
                {
                    break;
                }

                localFile.Write(buffer, 0, count);
                Console.Write(".");
                offset = offset + count;
            }
            
            localFile.Close();
            Console.WriteLine(string.Empty);
        }
        
        /// <summary>
        /// Try SCP.  We try putting and getting a file.
        /// </summary>
        /// <param name="fileName">The file name to use for the test file.</param>
        public static void TrySCP(string fileName)
        {
            InitFileTransfers(fileName);
            Console.WriteLine("===============> Now testing scp <====================");
            string cwd = Directory.GetCurrentDirectory();
            Connection conn = new Connection(hostName, hostPort);
            Console.WriteLine("current working directory: " + cwd);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }
            
            SCPClient client = new SCPClient(conn);
            Console.WriteLine("Putting the file to the server");
            client.Put(fileName, ".");
            Console.WriteLine("Getting the file from the server");
            client.Get(fileName, cwd);
            conn.Close();
        }

        /// <summary>
        /// Try a local forward.
        /// </summary>
        public static void TryLocalForward()
        {
            Console.WriteLine("===============> Now testing local forwarding <====================");
            Connection conn = new Connection(hostName, hostPort);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }
            
            conn.CreateLocalPortForwarder(2200, hostName, 22);

            Connection conn2 = new Connection("localhost", 2200);
            conn2.Connect();
            isAuth = conn2.AuthenticateWithPassword(userName, password);

            Session sess = conn2.OpenSession();
            sess.ExecCommand("uname -a && date");
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                Console.WriteLine(line);
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }

            // Close this session
            sess.Close();

            // Close the connections
            conn2.Close();
            conn.Close();

            Console.WriteLine("If you see a login banner above, the forward worked.");
            Console.WriteLine("Hit a key to continue...");
            Console.ReadKey();
            Console.WriteLine();
        }
        
        /// <summary>
        /// Try a remote forward.
        /// </summary>
        public static void TryRemoteForward()
        {
            Console.WriteLine("===============> Now testing remote forwarding <====================");
            Connection conn = new Connection(hostName, hostPort);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }
            
            // 1st two are relative to the remote server...what ip and port to bind to.
            // 2nd two are relative to the client (us)...what ipaddress and port to forward to.
            // This particular forward takes network traffic to localhost on the server on port 2222
            // and bounces it from us BACK to the normal ssh port for the server.
            conn.RequestRemotePortForwarding(hostName, 2222, hostName, 22);

            // Now see if we can use this connection/forward to get a login prompt
            // For this to work the remote server needs to have "GatewayPorts yes" in
            // the sshd_config file
            Connection conn2 = new Connection(hostName, 2222);
            conn2.Connect();
            isAuth = conn2.AuthenticateWithPassword(userName, password);

            Session sess = conn2.OpenSession();
            sess.ExecCommand("uname -a && date");
            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                Console.WriteLine(line);
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }

            // Close this session
            sess.Close();

            // Close the connections
            conn2.Close();
            
            conn.CancelRemotePortForwarding(2222);
            conn.Close();

            Console.WriteLine("If you see a login banner above, the forward worked.");
            Console.WriteLine("Hit a key to continue...");
            Console.ReadKey();
            Console.WriteLine();
        }

        /// <summary>
        /// Try X11 forwarding.  For this to work you must have a X11 server running on your
        /// local machine.  Try Cygwin/X or VcXsrv if you don't have one.
        /// </summary>
        public static void TryX11Forward()
        {
            Console.WriteLine("===============> Now testing X11 forwarding <====================");
            Connection conn = new Connection(hostName, hostPort);
            conn.Connect();
            bool isAuth = conn.AuthenticateWithPassword(userName, password);
            if (isAuth == false)
            {
                Console.WriteLine("Authentication failed.");
                Console.Write("Hit a key to close this program: ");
                Console.ReadKey();
                System.Environment.Exit(1);
            }

            Session sess = conn.OpenSession();
            sess.RequestX11Forwarding("localHost", 6000, null, false);
            sess.ExecCommand("xwininfo -root");

            Stream stdout = new StreamGobbler(sess.StandardOutput);
            StreamReader br = new StreamReader(stdout);

            int linecnt = 0;
            while (true)
            {
                string line = br.ReadLine();
                Console.WriteLine(line);
                if (line == null)
                {
                    break;
                }
                
                linecnt = linecnt + 1;
            }

            // Close this session
            sess.Close();

            // Close the connections
            conn.Close();

            Console.WriteLine("If you see a bunch of information about an X11 window");
            Console.WriteLine("the forward worked.  Hit a key to continue...");
            Console.ReadKey();
            Console.WriteLine();
        }

        /// <summary>
        /// Method called by SFTP and SCP tests.  Creates a random 4MB file to
        /// test file transfers to/from an ssh server.
        /// </summary>
        /// <param name="fileName">
        /// The file name to use for the test file.
        /// </param>
        private static void InitFileTransfers(string fileName)
        {
            if (File.Exists(fileName))
            {
                File.Delete(fileName);
            }
            
            if (File.Exists("org" + fileName))
            {
                File.Delete("org" + fileName);
            }

            // Create a test file (1MB of random data)
            System.Random rnd = new System.Random();
            FileStream localFile = new FileStream(fileName, FileMode.Create, FileAccess.Write);
            byte[] buffer = new byte[1024];
            for (int i = 0; i < 1024; i++)
            {
                for (int j = 0; j < 1; j++)
                {
                    rnd.NextBytes(buffer);
                    localFile.Write(buffer, 0, 1024);
                }
            }
            
            localFile.Close();
            
            // save a copy of the original file.
            File.Copy(fileName, "org" + fileName);
        }
    }
}